#coding: utf-8

from __future__ import with_statement

from logging import INFO
from time import strftime
from time import localtime
from PyQt4.QtCore import QString
from PyQt4.QtCore import Qt
from PyQt4.QtCore import QVariant
from PyQt4.QtCore import SIGNAL
from PyQt4.QtGui import QFileDialog, QLabel, QMessageBox, QPixmap, QWidget, QIcon

from nucentral.common.download import decodeFileContent
from nucentral.common.error import exceptionAsUnicode
from nucentral.common import tr, EDENWALL
from nucentral.common.logger import Logger
from nucentral.common.multisite import MULTISITE_MASTER
from nucentral.common.multisite import MULTISITE_SLAVE
from nucentral.common.odict import odict

from nucentral.client import NuCentralError, CLIENT_SUITE_NAME, CLIENT_SUITE_SHORTNAME

from nucentral.qt.application import load_resource
from nucentral.qt.application_name import APPLICATION_NAME
from nucentral.qt.central_window import CentralWindow
from nucentral.qt.central_window import EMBEDDED
from nucentral.qt.central_window import STANDALONE
from nucentral.qt.i18n import setup_locale
from nucentral.qt.session import showSession
from nucentral.qt.splash import SplashScreen
from nucentral.qt.user_settings import UserSettings

from console_edenwall.ui.console_edenwall_ui import Ui_MainWindow
from console_edenwall.log_viewer import DialogLogViewer
from console_edenwall.statusbar_modified_widget import StatusBarModifiedWidget
from console_edenwall.statusbar_modified_widget import StatusBarTextWidget
#from console_edenwall.audit_dialog import AuditDialog

from console_edenwall.version import EAS_VERSION

STATUS_NOT_SAVED = [ tr("<i>Configuration is not saved</i>"), "orange" ]

# Create applications' information
class ApplicationInfo:
    def __init__(self, name, resource, locale):
        self.name = name
        self.label = APPLICATION_NAME[name]
        self.resource = resource
        self.locale = locale
        self.index = None   # set later

APPLICATIONS = odict()   # name => ApplicationInfo
INDEX_TO_APP = {}        # index => ApplicationInfo

apps = (
    ApplicationInfo('console_edenwall.homepage', 'nulog3.rcc', ':/nulog/locale/nulog'),
    ApplicationInfo('nuconf.client.system', 'system.rcc', ':/locale/system'),
    ApplicationInfo('nuconf.client.services', 'system.rcc', ':/locale/system'),
    ApplicationInfo('nucentral_admin', 'nucentral_admin.rcc', ':/locale/nucentral-admin'),
    ApplicationInfo('nufaceqt', 'nuface3.rcc', ':/nuface/locale/nuface'),
    ApplicationInfo('nulog.client.qt', 'nulog3.rcc', ':/nulog/locale/nulog'),
)
if EDENWALL:
    apps += (ApplicationInfo('nudpi.client.qt', 'nudpi.rcc', ':/nudpi/locale/nudpi'),)
apps += (ApplicationInfo('nupki.qt', "nupki.rcc", ':/nupki/locale/nupki'),)

for index, app in enumerate(apps):
    app.index = index
    APPLICATIONS[app.name] = app
    INDEX_TO_APP[app.index] = app

class MainWindow(CentralWindow, Ui_MainWindow):
    def __init__(self, app, client, parent=None):
        # FIXME: Remove self.app attribute, replaced by self.qt_application
        # to avoid confusion
        self.app = app
        self.qt_application = app
        self._loading_splash_visible = False
        CentralWindow.__init__(self, client, parent, None)
        self.logs_viewer = None
        self.logger = Logger("eas")

        self.setupUi(self)
        self.setWindowTitle(CLIENT_SUITE_NAME)
        self.setupCentralWindow(app, STANDALONE)
        self.setupStatusBar()
        self.settings = UserSettings("EAS")
        # name => object (MainWindow instance)
        self.apps = dict()
        self.app_title = {}
        self.app_initialized = dict()
        self.connectSlots()
        self.read_settings(self.settings)
        self.default_menu_parent = QWidget()

        self._show_loading_splash = False
        self._diag_splash = None
        self.init_apps()

        self.qt_application.splash.showMessage(self.tr("Applications loaded"), Qt.AlignBottom | Qt.AlignRight)
        message = tr("%(CLIENT_SUITE_NAME)s %(EAS_VERSION)s loaded") % {
            "EAS_VERSION": EAS_VERSION,
            "CLIENT_SUITE_NAME": CLIENT_SUITE_NAME}
        self.statusBar().showMessage(message)

        self.EAS_MESSAGES['setTab'] = self.setTab

        self.addTopRightLogo()
        self._show_loading_splash = True

    def isReadOnly(self):
        """
        must be overridden by child
        """
        raise NotImplementedError()

    def setTitle(self, app=None):
        app_title = self.app_title.get(app)
        if app_title:
            title = u"%s - " % app_title
        else:
            title = u''
        title += u"%s@%s - %s" \
            % (self.client.login, self.client.host, CLIENT_SUITE_NAME)
        if EDENWALL:
            if self.multisite_type == MULTISITE_MASTER:
                multisite = tr("master")
            elif self.multisite_type == MULTISITE_SLAVE:
                multisite = tr("firewall")
            else:
                multisite = tr("no multisite")
            title += " (%s)" % multisite
        self.setWindowTitle(title)

    def error(self, message, title=None, dialog=True, escape=True):
        if not title:
            title = self.tr(
                "%(CLIENT_SUITE_SHORTNAME)s error" %
                {"CLIENT_SUITE_SHORTNAME": CLIENT_SUITE_SHORTNAME.upper()}
                )
        CentralWindow.error(self, message, title, dialog=dialog, escape=escape)

    def connectSlots(self):
        self.connect(self.action_About, SIGNAL("triggered()"), self.about)
        self.connect(self.action_Quit, SIGNAL("triggered()"), self.close)
        self.connect(self.tabWidget, SIGNAL("currentChanged(int)"), self.currentChanged)
        self.connect(self.action_show_session, SIGNAL("triggered()"), self.showSession)
        self.connect(self.action_Logs, SIGNAL("triggered()"), self.showLogs)
        #self.connect(self.action_Audit, SIGNAL("triggered()"), self.fillAudit)
        if EDENWALL:
            self.connect(self.action_diagnostic, SIGNAL("triggered()"), self.dl_diag)

    def search_app(self, appinfo, standalone = EMBEDDED):
        # Load module
        module_name = appinfo.name
        module = __import__(module_name)
        for submodule in module_name.split(".")[1:]:
            module = getattr(module, submodule)
        window_cls = module._get_window_class()

        if EDENWALL:
            if not(window_cls.ROLES & self.getRoles()):
                self.logger.error("Disable tab %s: you don't have the required role" % appinfo.label)
                return None

        # Create the application window
        window = window_cls(self.qt_application, self.client, standalone=standalone, eas_window=self)
        return window

    def configure_app(self, window, appinfo):
        window.setObjectName(appinfo.name)
        window.menuBar().hide()
        window.statusBar().hide()

        # Add menu entries on my own menubar
        self.add_menu_actions(appinfo.name, window.menuBar(), self.menuBar())
        self.menu_File.addSeparator()
        self.menu_File.addAction(self.action_Quit)

    def add_menu_actions(self, app_name, dmenu, lmenu):
        """
        Add actions of an hosted application's menu.

        @param app_name  application name
        @param dmenu  hosted menu
        @param lmenu  EAS menu
        """
        if lmenu != self.menuBar():
            lmenu.addSeparator()
        for daction in dmenu.actions():
            laction = None
            for la_tmp in lmenu.actions():
                if la_tmp.text() == daction.text() and not la_tmp.data().isValid():
                    laction = la_tmp
                    break

            daction.setData(QVariant(app_name))
            daction.setVisible(False)
            if not laction:
                lmenu.addAction(daction)
            elif laction.menu() and daction.menu():
                self.add_menu_actions(app_name, daction.menu(), laction.menu())

    def addTopRightLogo(self):
        w = QLabel()
        w.setPixmap(QPixmap(':/images/logo_ban'))
        self.tabWidget.setCornerWidget(w)

    def init_app(self, app):
        name = app.name
        self.logger.warning("Load application %s" % app.label)
        try:
            window = self.search_app(app)
            if not window is None:
                self.configure_app(window, app)
        except SystemExit:
            window = None
        except Exception, err:
            self.writeError(err, tr("Application %s initialization error") % app.label)
            window = None

        index = app.index
        # Tabs entry
        if window:
            self.apps[app.name] = window
            index = self.tabWidget.insertTab(index, window, "")
            self.connect(window, SIGNAL('EAS_Message'), lambda *args, **kwargs: self.gotMessage(app.name, *args, **kwargs))
            if window.ICON:
                icon = QIcon(window.ICON)
                self.tabWidget.setTabIcon(index, icon)
        else:
            index = self.tabWidget.insertTab(index, QWidget(), "")
            self.tabWidget.setTabEnabled(index,False)
        self.tabWidget.setTabText(index, app.label)

    def init_apps(self):
        # Initialize resources and locales
        for index, app in enumerate(APPLICATIONS.itervalues()):
            if app.resource:
                load_resource(app.resource)
            if app.locale:
                setup_locale(self.qt_application, app.locale)

        # Load all applications
        for index, app in enumerate(APPLICATIONS.itervalues()):
            progress = "%s/%s" % (1 + index, len(APPLICATIONS))
            message = tr("Loading application %s (%s)") % (app.label, progress)
            self.qt_application.splash.showMessage(message, Qt.AlignBottom | Qt.AlignRight)
            self.init_app(app)

        # Call app.initialized() for all application having this method
        for appinfo in APPLICATIONS.itervalues():
            app_name = appinfo.name
            try:
                app = self.apps[app_name]
            except KeyError:
                continue
            if not hasattr(app, 'initialized'):
                continue
            func = app.initialized
            if not callable(func):
                continue
            func()

        # Set tab to the first working application
        for appinfo in APPLICATIONS.itervalues():
            try:
                app = self.apps[appinfo.name]
            except KeyError:
                continue
            self.tabWidget.setCurrentIndex(appinfo.index)
            self.setTitle(app)
            break

    def getApp(self, name):
        if name == 'eas':
            return self
        else:
            return self.load_app(name)

    def gotMessage(self, from_app, dest_app, *args, **kwargs):
        window = self.getApp(dest_app)
        if window is None:
            # ignore message to missing applications
            return
        window.EAS_MessageHandler(from_app, *args, **kwargs)

    def setTab(self, from_app, new_tab):
        try:
            index = APPLICATIONS[new_tab].index
            self.tabWidget.setCurrentIndex(index)
        except KeyError:
            pass

    def read_settings(self, settings):
        settings.beginGroup("MainWindow")
        pos = settings.getPoint("pos")
        size = settings.getSize("size")
        settings.endGroup()

        if size:
            self.resize(size)
        if pos:
            self.move(pos)

    def write_settings(self, settings):
        settings.beginGroup("MainWindow")
        settings.setPoint("pos", self.pos())
        settings.setSize("size", self.size())
        settings.endGroup()

    def about(self):
        message = tr(
            "<b>%(CLIENT_SUITE_NAME)s %(EAS_VERSION)s</b>"
            "<p>Copyright &copy;2008-2010 EdenWall Technologies,"
            "All Rights Reserved.") % {
                "EAS_VERSION": EAS_VERSION,
                "CLIENT_SUITE_NAME": CLIENT_SUITE_NAME
            }
        QMessageBox.about(self, self.tr("About"), message)

    def toggle_application(self, app_name):
        index = APPLICATIONS[app_name].index
        enable = not self.tabWidget.isTabEnabled(index)
        self.tabWidget.setTabEnabled(index, enable)

    def setEnabledApplication(self, app_to_modify, is_enabled):
        """
        enable/disable app in eas
        """
        app_name = None
        for name, instance in self.apps.iteritems():
            if app_to_modify is instance:
                app_name = name
                break
        self.tabWidget.setTabEnabled(APPLICATIONS[app_name].index, is_enabled)

    def currentChanged(self, index):
        app_name = INDEX_TO_APP[index].name
        app = self.load_app(app_name)
        if app:
            # Display only EAS menu actions, and actions of the current application
            self.show_app_actions(app_name, self.menuBar().actions())
            self.setTitle(app)

    def showLoadingSplashScreen(self, text):
        if self._loading_splash_visible:
            return False
        self._loading_splash_visible = True
        self.qt_application.setOverrideCursor(Qt.WaitCursor)
        if self._show_loading_splash:
            diag = SplashScreen(text=text)
            diag.show()
            self._diag_splash = diag
        return True

    def getLoadingSplashScreenText(self):
        if self._diag_splash is not None:
            return self._diag_splash.getText()
        else:
            return None

    def hideLoadingSplashScreen(self):
        if not self._loading_splash_visible:
            return False
        self._loading_splash_visible = False
        if self._show_loading_splash:
            self._diag_splash.hide()
            self._diag_splash = None
        self.qt_application.restoreOverrideCursor()
        return True

    def load_app(self, app_name):
        try:
            app = self.apps[app_name]
        except KeyError:
            return None

        if app_name in self.app_initialized:
            return app
        self.app_initialized[app_name] = True

        info = APPLICATIONS[app_name]
        text = tr("Loading application %s...") % info.label
        # else: At startup, we don't use a second splash screen because there
        #       is already the global splash screen
        try:
            self.showLoadingSplashScreen(text)
            try:
                app.load()
            except Exception, err:
                self.exception(err)
                return None
            else:
                return app
        finally:
            self.hideLoadingSplashScreen()

    def show_app_actions(self, app_name, actions):
        """
        This recursive method display only actions of EAS or of the specified
        application.
        """
        for action in actions:
            if action.menu():
                self.show_app_actions(app_name, action.menu().actions())
            if not action.data().isValid() or unicode(action.data().toString()) == app_name:
                action.setVisible(True)
            else:
                action.setVisible(False)

    def showLogs(self):
        if not self.logs_viewer:
            try:
                self.logs_viewer = DialogLogViewer(self.qt_application, self.client)
            except NuCentralError:
                self.action_Logs.setEnabled(False)
                QMessageBox.critical(self,
                    CLIENT_SUITE_SHORTNAME,
                    tr("Logger module is not loaded - Disabling log viewer"))
                return

        self.logs_viewer.show()

#    def fillAudit(self):
#        self.audit_dialog = AuditDialog(self)

    def showSession(self):
        try:
            session = self.client.call('session', 'get')
        except Exception, err:
            self.exception(err)
            return
        showSession(self, session)

    def setupStatusBar(self):
        self._text_widget = w = StatusBarTextWidget(self)
        self.statusBar().addPermanentWidget( w )
        self._modified_widget = w2 = StatusBarModifiedWidget(self)
        self.statusBar().addPermanentWidget( w2 )

    def setModified(self, modified=True, message=None):
        self.setWindowModified(modified)
        self._modified_widget.setModifiedState(modified)
        if modified:
            if not message:
                message = STATUS_NOT_SAVED
            if isinstance(message,(str,unicode)):
                color = None
            else:
                message, color = message
            self._text_widget.setTextWithColor(message, color)
        else:
            self._text_widget.hide()

    def sessionModified(self):
        for name, app in self.apps.iteritems():
            app.sessionModified()

    def ask_save(self):
        msgBox = QMessageBox(self)
        msgBox.setWindowTitle(tr("Exit EAS"))
        msgBox.setText(tr("The configuration has been edited."))
        msgBox.setInformativeText(tr("Do you want to exit anyway (and loose your changes)?"))
        msgBox.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        msgBox.setDefaultButton(QMessageBox.Cancel)
        ret = msgBox.exec_()
        return ret

    def closeEvent(self, closeEvent):
        if self.isWindowModified() and self.ask_save() != QMessageBox.Ok:
            closeEvent.ignore()
            return
        self.logger.error("Exiting")
        self.unloadApps()
        closeEvent.accept()
        self.logger.error("Exit.")

    def unloadApps(self):
        if self.keep_alive:
            self.keep_alive.stop()
            self.keep_alive = None

        if self.logs_viewer:
            self.logs_viewer.close()

        self.write_settings(self.settings)
        for name, app in self.apps.iteritems():
            self.logger.debug("Unload application: %s" % name)
            try:
                app.quit()
            except Exception, err:
                self.writeError(err,
                    "Error on unloading application %s" % name,
                    log_level=INFO)

    def dl_diag(self):
        now = strftime('%Y-%m-%d-%H-%M', localtime())
        default_filename = 'diagnostic_edenwall_%s.tar.gz' % now
        filename = QFileDialog.getSaveFileName(self, tr('Select a destination'),
                                               QString(default_filename))

        if filename.isEmpty():
            return

        filename = unicode(filename)

        async = self.client.async()
        async.call("tools", 'getDiagnosticFile',
            callback = self.successDiag,
            errback = self.errorDiag,
            callbackArgs=(filename,),
            )

        text = tr('Please wait while the diagnostic file is being downloaded')
        self.showLoadingSplashScreen(text)

    def successDiag(self, diagnostic, filename):
        self.hideLoadingSplashScreen()
        with open(filename, 'wb') as fd:
            fd.write(decodeFileContent(diagnostic))
        QMessageBox.information(
            self,
            tr("Download successful"),
            tr("Diagnostic file was downloaded successfully.")
            )

    def errorDiag(self, error):
        self.hideLoadingSplashScreen()
        warning = QMessageBox(self)
        warning.setWindowTitle(tr('Fetching diagnostic failed'))
        warning.setText(tr('An error has occurred while fetching the diagnostic file'))
        warning.setDetailedText(exceptionAsUnicode(error))
        warning.setIcon(QMessageBox.Warning)
        warning.exec_()

