#!/usr/bin/python

"""
A Python API to do requests to an NNDaemon

Copyright 2010 EdenWall Technologies

Written by Feth AREZKI <farezki@edenwall.com>
"""

from functools import wraps
from logging import Formatter, getLogger, StreamHandler, DEBUG
from select import select
from socket import socket, AF_UNIX, SOCK_STREAM, error
from sys import stdout
from threading import Thread
from Queue import Queue

SELECT_FREQ = 0.1 #seconds

RECV_SIZE = 4096

END_OF_RESULT = "\n"

#Some stdout wrappers don't provide isatty!
if hasattr(stdout, 'isatty') and stdout.isatty():
    _FORMATS = {
        'query': "->\033[32m", #green
        'info': "\033[1;30m", #grey
        'success': "<-\033[34m", #blue
        'error': "\033[31m", #red
        'resulterror': "<-\033[31m", #red
    }
    RESET_SEQ = "\033[0m"
else:
    _FORMATS = {
        'query': "->",
        'info': "",
        'success': "<-",
        'error': "",
        'resulterror': "<-",
    }
    RESET_SEQ = ""

SOCKET_FILE = '/var/tmp/nnd.sock'

END_OF_REQUEST = "\n"
RESULT_SEPARATOR = ","

RESULT_OK = "200"
RESULT_NOK = "400"

class ServerError(Exception): pass


class NotAnderstood(Exception): pass


def genloggers():
    for key, color in _FORMATS.iteritems():
        logger = getLogger(".%s" % key)
        stdouthandler = StreamHandler(stdout)
        logger.addHandler(stdouthandler)
        format = "%%(asctime)s %s %%(message)s %s" % (color, RESET_SEQ)
        stdouthandler.setFormatter(Formatter(format))
        logger.setLevel(DEBUG)
        yield key, (logger, stdouthandler)


def _parseresult(result):
    if result is None:
        return
    parsed = result.split(" ", 1)
    if len(parsed) == 2:
        if parsed[0] == RESULT_OK:
            return parsed[1]
        elif parsed[0] == RESULT_NOK:
            raise ServerError(parsed[1])
    raise NotAnderstood(result)


def command(operation):
    @wraps(operation)
    def newoperation(*args, **kwargs):
        result = operation(*args, **kwargs)
        return _parseresult(result)
    return newoperation


def listresult(operation):
    @wraps(operation)
    def newoperation(*args, **kwargs):
        result = operation(*args, **kwargs)
        return result.split(RESULT_SEPARATOR)
    return newoperation


class Client(object):

    PROTO_VERSION = 2

    def __init__(self, blocking=True, logger=None):
        """
        blocking:
            if False: results are not returned, but None.
            if True: waits for request to complete, and returns results
        logger:
            if a logger is supplied, it will be used by this client
        """
        self.blocking = blocking
        self.socket = None
        self._setuploggers(logger=logger)
        self.sockfile = ""
        self._recv_queue = []
        self._processed_results = Queue()

    def connect(self, filename=SOCKET_FILE):
        self.sockfile = filename
        self.infolog("Connecting to socket %s" % self.sockfile)
        try:
            return self._setupsocket()
        except error:
            self.errorlog("Unable to connect. Will try later.")
            return None

    def testgroup(self, groupname):
        if not "@" in groupname:
            domains = self.domains()
            if not domains:
                raise ServerError(
                    "No domains are setup at all, "
                    "cannot infer default domain name."
                    )
            defaultdomain = domains[0]
            groupname = "%s@%s" % (groupname, defaultdomain)
        return groupname in self.searchgroups()

    def testuser(self, username):
        try:
            self.searchuid(username)
        except:
            return False
        return True

    @listresult
    @command
    def searchgroups(self):
        return self._stdcommand(u"groups")

    @listresult
    @command
    def searchuid(self, uid):
        return self._stdcommand(u"usergroups %s" % uid)

    #not @command because this is a delegation to auth2
    def auth(self, username, domain, password):
        return self.auth2("%s@%s" % (username, domain), password)

    @command
    def auth2(self, username, password):
        return self._stdcommand(u"auth %s,%s" % (username, password))

    @listresult
    @command
    def domains(self):
        return self._stdcommand(u"domains")

    @command
    def unbind(self):
        return self._stdcommand(u"unbind")

    def _setuplogger(self, fnname, logger):
        def logfn(message):
            logger[0].warning(message)
        logfn.__name__ = fnname
        setattr(self, fnname, logfn)

    def _setuploggers(self, logger=None):
        loggers = {}
        if not logger:
            loggers.update(dict(genloggers()))
        for word in _FORMATS:
            fnname = "%slog"% word
            self._setuplogger(fnname, loggers.get(word, (logger, None)))

    def _setupsocket(self):
        self.results_expected_nb = 0
        self.socket = socket(AF_UNIX, SOCK_STREAM)
        self.socket.connect(self.sockfile)
        self.infolog("Successfully connected to '%s'" % self.sockfile)
        return self._stdcommand("hello %s" % self.PROTO_VERSION)

    def _stdcommand(self, command):
        self._writequery(command)
        self._sendcommand(command)
        if self.blocking:
            self._writeresult()
            return self._processed_results.get()
        else:
            thread = self._mkthread(self._writeresult)
            thread.start()

    def _mkthread(self, target):
        return Thread(target=target)

    def _sendstring(self, string):
        while True:
            rfd, wfd, efd = select(
                (),
                (self.socket.fileno(),),
                (self.socket.fileno(),),
                0.1
                )
            if wfd:
                if self.socket.send(string.encode("utf-8")) < 0:
                    raise Exception("Socket in error: problem sending data.")
                break
        self.results_expected_nb += 1

    def _writequery(self, command):
        self.querylog("%s" % command)

    def _sendcommand(self, command):
        command = command + END_OF_REQUEST
        try:
            self._sendstring(command)
        except (EnvironmentError, error), err:
            sockerror = isinstance(err, error)
            if sockerror or err.errno == 32:
                self.errorlog("Disconnected? Attempting reconnection")

            if sockerror or err.errno in (32, 107):
                try:
                    self._setupsocket()
                except error:
                    self.errorlog("Socket file does not exist")
                    raise
                else:
                    self._sendstring(command)
            else:
                raise

    def _queue_result(self, data):
        data = unicode(data,"utf-8")
        parsed = [
            item
            for item in data.split(END_OF_RESULT)
            if item
        ]

        if parsed:
            if self._recv_queue:
                #There was something left in req_list: it is by definition an
                #unfinished request
                self._recv_queue[0] += parsed[0]
                del parsed[0]
            self._recv_queue += parsed
        return data.endswith(END_OF_RESULT)

    def _handle_recv_data(self, lastchunk_terminated):
        if not self._recv_queue:
            return 0

        if lastchunk_terminated:
            #Also display the last item in list
            last = len(self._recv_queue)
        else:
            #don't display the last item in list yet,
            #because it is not finished yet
            last = len(self._recv_queue) - 1

        if last == 0:
            #nothing to display
            return 0

        for index in xrange(last):
            #Process any results we might have received
            self._process_result(self._recv_queue[0])
            del self._recv_queue[0]

        return last

    def _process_result(self, text):
        if text.startswith('200'):
            self.successlog(text)
        else:
            self.resulterrorlog(text)
        if self.blocking:
            #self._processed_results is only used in blocking mode
            self._processed_results.put(text)

    def _writeresult(self):
        lastchunk_terminated = False
        while self.results_expected_nb > 0:
            resultsnb = self._handle_recv_data(lastchunk_terminated)
            self.results_expected_nb -= resultsnb
            rfd, wfd, efd = select([self.socket.fileno()], (), (), SELECT_FREQ)

            if not rfd:
                continue
            data = self.socket.recv(RECV_SIZE)

            if not data:
                self.resulterrorlog("Connection reset?")
                break

            lastchunk_terminated = self._queue_result(data)

    # abstract methods
    def successlog(self, text):
        pass

    def resulterrorlog(self, text):
        pass

    def quit(self):
        if not self.socket:
            return
        self.socket.close()
        self.socket = None


if __name__ == "__main__":
    print "======= TEST 1 ========"
    client = Client()
    client.connect()
    groups = client.searchgroups()
    print "Found groups:", groups

    print "======= TEST 2 ========"
    logger = getLogger("coin")
    stdouthandler = StreamHandler(stdout)
    logger.addHandler(stdouthandler)
    format = "[CUSTOM LOGGER] %(message)s"
    stdouthandler.setFormatter(Formatter(format))
    logger.setLevel(DEBUG)
    client = Client(logger=logger)
    client.connect()
    groups = client.searchgroups()

