import ldap
import re
from struct import pack
from .cache import NNDCache
from .error import NNDError
from .server import Server

REALM_REGEX = re.compile(u'CN=Users,(DC=.*(?:,DC=.*)*)$', re.IGNORECASE)

class ADServer(Server):
    def __init__(self, logger, domain, name, parser):
        Server.__init__(self, logger, domain, name, parser)
        self.groupdn_to_name = NNDCache(self, 300)
        self.search_attrs = (
            self.domain.user_member_attr,
            "objectSid",
            "primaryGroupID",
            )

    def _find_user(self, uid, calltime, timeleft):
        user_filter = self.domain.user_filter % uid
        result, timeleft, now = self.timedsyncreq(
            self.domain.user_base_dn,
            ldap.SCOPE_SUBTREE,
            user_filter.encode("utf-8"),
            self.search_attrs,
            calltime,
            timeleft
            )
        return result, timeleft, now

    def _find_primary_group(self, group_list, user):
        groupid = user["primaryGroupID"][0]
        user_sid = user["objectSid"][0]
        group = self.resolve_primarygroupid(groupid, user_sid)
        if group:
            group_list.append(group)

    def _get_group(self, entry):
        dn = unicode(entry[0], 'utf-8')
        match = REALM_REGEX.search(dn)
        if match:
            parts = match.group(1).split(u',')
            parts = [part[3:] for part in parts]
            realm = u'.'.join(parts)
        else:
            self.debug("Unable to get the realm from the DN: %s" % dn)
            realm = self.domain.realm
        name = entry[1][self.domain.group_attr_name][0]
        name = unicode(name, "utf-8")
        return (name, realm)

    def _find_group_dn(self, group_list, group_dn):
        cache_entry = self.groupdn_to_name[group_dn]
        if cache_entry:
            group_list.append(cache_entry)
            return

        res = self.con.search_st(
            group_dn,
            ldap.SCOPE_BASE,
            "objectClass=*",
            (self.domain.group_attr_name, "objectSid"),
            timeout=self.groups_timeout
            )
        if len(res) != 1:
            raise NNDError("Invalid number (%s) of results for the group DN" % len(res))
        group = self._get_group(res[0])
        self.groupdn_to_name[group_dn] = group
        group_list.append(group)

    def _find_other_groups(self, group_list, user):
        try:
            groups = user[self.domain.user_member_attr]
        except KeyError:
            return
        for group_dn in groups:
            self._find_group_dn(group_list, group_dn)

    def _usergroups(self, uid, calltime, timeleft):
        self.debug("AD Searching for uid: %s" % uid)

        # first, find user
        result, timeleft, now = self._find_user(uid, calltime, timeleft)
        if len(result) == 0:
            raise NNDError("Unknown user: %s" % uid)
        if len(result) != 1:
            raise NNDError("More than one user match name: %s" % uid)
        user = result[0][1]

        group_list = []
        self._find_primary_group(group_list, user)
        self._find_other_groups(group_list, user)
        return group_list

    def resolve_primarygroupid(self, groupid, user_sid):
        self.debug("Resolving group ID: %s" % groupid)
        # We have to get the (binary) SID for the user, substitute the RID by
        # the group RID, then query on this RID:
        # http://support.microsoft.com/kb/297951
        new_rid = pack('i', int(groupid))
        new_sid = user_sid[0:24] + new_rid
        filter = "(objectSid=%s)" % (''.join( "\\%02x" % ord(char) for char in new_sid) ,)
        attrs = [self.domain.group_attr_name, ]
        res = self.con.search_st(self.domain.user_base_dn, ldap.SCOPE_SUBTREE, filter, attrs, timeout=self.groups_timeout)
        if res:
            return self._get_group(res[0])
        else:
            return None

    def _initialize(self):
        success = Server._initialize(self)
        if success:
            self.debug("Configuring ldap not to follow referrals as this does not work on windows")
            self.con.set_option(ldap.OPT_REFERRALS,0)
        return success

    def _groups(self, calltime, timeleft):
        self.debug("AD Enumerating groups")
        filter_groups = self.domain.group_enum_filter
        if filter_groups[0] != '(':
            filter_groups = "(" + filter_groups + ")"
        filter = filter_groups
        self.debug("filter: %s" %  filter)
        if self.use_pagination:
            result = self.iter_paginated_request(
                self.domain.group_base_dn,
                ldap.SCOPE_SUBTREE,
                filter,
                timeleft,
                [self.domain.group_attr_name]
            )
        else:
            result = self.con.search_st(
                self.domain.group_base_dn,
                ldap.SCOPE_SUBTREE,
                filter,
                [self.domain.group_attr_name],
                timeout=timeleft
            )
        # we have a list of DN of groups, iterate and query names
        return [self._get_group(obj) for obj in result]

    def auth_user_dn(self, username, timeleft):
        return u'%s@%s' % (username, self.domain.realm), timeleft

