"""
Copyright 2010 EdenWall Technologies

Written by Pierre-Louis Bonicoli <bonicoli@edenwall.com>
Modified by Feth Arezki <farezki@edenwall.com>
"""

from functools import wraps
from ldap import LDAPError, TIMELIMIT_EXCEEDED, TIMEOUT

from .server import AD, LDAP, EDIRECTORY
from .ad_server import ADServer
from .ldap_server import LDAPServer
from .edirectory_server import EDirectoryServer
from .error import UnknownDomainType, NNDError
from .log import setmeupaslogger
from .utils import checkonesubst


def fixfilter(ldap_filter):
    if not ldap_filter:
        return ""

    ldap_filter = ldap_filter.strip()

    if not ldap_filter[0] == '(':
        ldap_filter = u'(%s' % ldap_filter
    if not ldap_filter[-1] == ')':
        ldap_filter = u'%s)' % ldap_filter

    if not ldap_filter.count('(') == ldap_filter.count(')'):
        raise ValueError("'('s != ')'s in %s" % ldap_filter)

    return ldap_filter

_REFERRAL_MSG = "Got unfollowable LDAP REFERRAL. The server %s cannot answer your query. REFERRAL error: %s"

def _nndservererror(server, method, err):
    """
    raises an NNDError if the supplied LDAPError is:
        -REFERRAL
    else, returns a (hopefully) sensible message
        -TIMELIMIT_EXCEEDED
        -TIMEOUT
        -OTHER
    """
    message = ""
    if isinstance(err, (TIMELIMIT_EXCEEDED, TIMEOUT)):
        return "Timeout on server %s while executing '%s'" % (
                server.shortname, method.__name__
                )

    if err.args:
        errdesc = err.args[0].get('desc', '')
        if errdesc != 'REFERRAL':
            #from http://www.python-ldap.org/doc/html/ldap.html
            # The exceptions are accompanied by a dictionary possibly
            # containing an string value for the key desc (giving an
            # English description of the error class) and/or a string
            # value for the key info (giving a string containing more
            # information that the server may have sent).
            info = err.args[0].get('info', '')
            if not info:
                info = errdesc
            return "Got error %s <'%s'> on server %s, method %s" % \
                (type(err), info, server.shortname, method.__name__)
        message = _REFERRAL_MSG % (server.shortname, err)
    else:
        message = "On server %s LDAP error [%s]: %s" % (
            server.shortname, type(err), err
            )

    raise NNDError(message)

def serveroperation(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        gotinitialized = False
        for server in self.iterinitializedservers():
            gotinitialized = True
            try:
                return method(self, server, *args, **kwargs)
            except LDAPError, err:
                self.debug(_nndservererror(server, method, err))

                self.debug("Call on initialized server %s failed:%s" %
                    (server.shortname, unicode(err))
                    )
                server.initialized = False

        if gotinitialized:
            msg = "Could not get satisfaction from any initialized server. "\
            "Reconnecting..."
        else:
            msg = "Connected to 0 server. Reconnecting..."
        self.debug(msg)

        serversnb = len(self.servers)
        for index, server in enumerate(self.iterservers()):
            if not server.initialize():
                self.debug("[%s] Could not reconnect to %s." % (self.realm, server.uri))
                continue
            self.debug("Reconnected to %s." % server.shortname)
            try:
                return method(self, server, *args, **kwargs)
            except LDAPError, err:
                self.debug("Call on initialized server %s failed:%s" %
                    (server.shortname, unicode(err))
                    )
                if index + 1 == serversnb:
                    self.error("Error while interrogating the last server.")
                    raise
                continue
        self.error("[%s] Could not connect to any server!" % self.realm)
        raise NNDError("Could not connect to any server!")
    return wrapper


class Domain(object):
    SERVER_CLASS = None
    COMMERCIAL_TYPE = ""
    FILTERS = {}
    DEFAULT_GROUP_ATTR_NAME = 'cn'

    def __init__(self, logger, conf, section):

        self.conf_section = section
        self.conf = conf

        for option, default in self.FILTERS.iteritems():
            value = self.conf.get(self.conf_section, option, default)
            setattr(self, option, fixfilter(value))

        self.realm = conf.get(section, 'realm')
        self.group_attr_name = conf.get(section, 'group_attr_name', self.DEFAULT_GROUP_ATTR_NAME)
        self.group_base_dn = conf.get(section, 'group_base_dn')

        self.user_base_dn = conf.get(section, 'user_base_dn')

        self.servers = []
        servers = conf.get(section, 'servers')
        servers = servers.split(',')
        setmeupaslogger(self, logger)

        for server_name in servers:
            server = self.SERVER_CLASS(logger, self, server_name, conf)
            self.servers.append(server)

    def initialize(self):
        for server in self.servers:
            if not server.initialize():
                raise NNDError("[%s] Problem initializing server %s." % (self.realm, server.uri))

    def iterinitializedservers(self):
        for server in self.servers:
            if server.initialized:
                yield server

    def iterservers(self):
        # TODO implement fallback here
        return iter(self.servers)

    def testconf(self):
        for server in self.servers:
            server.testconf()

    @serveroperation
    def groups(self, server):
        return server.groups()

    @serveroperation
    def usergroups(self, server, uid):
        return server.usergroups(uid)

    @serveroperation
    def auth(self, server, username, password):
        return server.auth(username, password)

    @serveroperation
    def unbind(self, server):
        server.unbind()


class ADDomain(Domain):
    SERVER_CLASS = ADServer
    COMMERCIAL_TYPE = "Active Directory"
    FILTERS = {
        'group_enum_filter': '(objectClass=group)',
        'group_filter': 'cn=%s',
        'user_filter': 'sAMAccountName=%s',
    }

    def __init__(self, logger, conf, section):
        self.user_member_attr = conf.get(section, 'user_member_attr', 'memberOf')

        Domain.__init__(self, logger, conf, section)
        checkonesubst(self.group_filter, 'domain.group_filter %s' % section)

        checkonesubst(self.user_filter, 'domain.user_filter %s' % section)

class LDAPDomain(Domain):
    SERVER_CLASS = LDAPServer
    COMMERCIAL_TYPE = "LDAP"
    FILTERS = {
        'group_filter': 'objectClass=posixGroup',
        'user_filter': '(&(uid=%s)(objectClass=posixAccount))'
        }
    DEFAULT_GROUP_MEMBER_ATTR = 'memberUid'

    def __init__(self, logger, conf, section):
        self.group_member_attr = conf.get(section, 'group_member_attr', self.DEFAULT_GROUP_MEMBER_ATTR)

        Domain.__init__(self, logger, conf, section)

        if self.COMMERCIAL_TYPE == LDAP:
            checkonesubst(self.user_filter, 'domain.user_filter %s' % section)

class EDirectoryDomain(LDAPDomain):
    SERVER_CLASS = EDirectoryServer
    COMMERCIAL_TYPE = "eDirectory"
    FILTERS = {
        'group_enum_filter': 'objectClass=groupOfNames',
        'group_filter': 'objectClass=groupOfNames',
        'user_filter': '(&(uid=%s)(objectClass=Person))',
        }
    DEFAULT_GROUP_ATTR_NAME = 'dn'
    DEFAULT_GROUP_MEMBER_ATTR = 'securityEquals'

def createDomain(logger, conf, section):
    domaintype = conf.get(section, 'type', default=LDAP).lower()
    if domaintype == AD:
        return ADDomain(logger, conf, section)
    elif domaintype == LDAP:
        # TODO use a factory
        return LDAPDomain(logger, conf, section)
    elif domaintype == EDIRECTORY:
        return EDirectoryDomain(logger, conf, section)
    else:
        raise UnknownDomainType()

