"""
Copyright 2010 EdenWall Technologies

Written by Pierre Chifflier <chifflier@edenwall.com>
"""

import os
import socket
from threading import Thread
from weakref import ref

from .config import load_config
from .workers import NNDWorkerThread
from .log import setmeupaslogger
from .utils import socketinfo, formatsockerror, shortsockinfo
from .error import NNDError

class NNDMainThread(Thread):

    def __init__(self, logger, config, domains):
        """
            config: dict with configuration
            domains: list of domains
        """
        Thread.__init__(self)

        setmeupaslogger(self, logger)

        self.logger = logger
        self.domains = {}
        for domain in domains:
            realm = domain.realm.lower()
            if realm in self.domains:
                raise NNDError("Duplicate domain realm: %s" % realm)
            self.domains[realm] = domain
        self.conf = config
        defaut_domain_section = self.conf.get('global', 'default_domain')
        self.default_domain = self.conf.get(defaut_domain_section, 'realm')
        self.workers = set()

    def _cleanuprefs(self):
        dead = set(
            workerref
            for workerref in self.workers
            if not workerref()
        )
        self.workers -= dead

    def run(self):
        socket_path = self.conf.get("global",  "socket")

        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            os.remove(socket_path)
        except OSError:
            pass
        os.umask(0177)
        s.bind(socket_path)
        s.listen(5)

        self.info("Waiting for clients on %s." % shortsockinfo(s))
        while 1:

            self._cleanuprefs()

            try:
                conn, addr = s.accept()

                self.debug("New client %s" % socketinfo(conn))

                worker = NNDWorkerThread(
                    self.logger,
                    self.domains,
                    conn, addr,
                    self.default_domain
                    )
                worker.start()
                self.workers.add(ref(worker))
            except socket.error, e:
                try:
                    (errnum, errmsg) = e
                    if errnum == 4:
                        # interrupted system call, retry
                        continue
                except Exception:
                    pass
                self.warning(formatsockerror(conn, e))
                break

        self.info("Exiting main thread")

    def close_all(self):
        for workerref in self.workers:
            worker = workerref()
            if worker:
                worker.stop()

def reload_program_config(*args, **kwargs):
    raise NotImplementedError()
    #self.critical("reload (currently does nothing)")
    return load_config()

