"""
Copyright 2010 EdenWall Technologies

Written by Pierre Chifflier <chifflier@edenwall.com>
Modified by Pierre-Louis Bonicoli <bonicoli@edenwall.com>
Modified by Feth Arezki <farezki@edenwall.com>
"""

from itertools import imap, ifilter
import ldap
from ldap.controls import SimplePagedResultsControl
from time import time

from .cache import NNDCache
from .formatter import format_item
from .log import setmeupaslogger, Filewrapper
from .error import TLSError, NNDError
from .utils import accept_groupname

ALLGROUPS = object()

REC_INT_OPTIONS = (
#all default values: ldap version. timeouts in seconds
    ('version', 3),
    ('auth_timeout', 10),
    ('groups_timeout', 90),
    ('usergroups_timeout', 15),
)

REC_STR_OPTIONS = (
    ('bind_dn', ''),
    ('bind_pw', ''),
)

REC_BOOL_OPTIONS = (
    ('tls', False),
    ('checkcert', False),
)

AD = 'ad'
LDAP = 'ldap'
EDIRECTORY = 'edirectory'

def updatetimes(calltime, timeleft):
    now = time()
    timeleft -= now - calltime
    return timeleft, now

def _set_check_cert(check_cert, ldap_obj):
    if check_cert :
        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_HARD)
    else:
        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)


def _ldapconnect_basic(logger, ldap_version, uri, check_cert):
    ldap.protocol_version = ldap_version

    #Set at ldap module level.
    _set_check_cert(check_cert, ldap)

    trace_file = Filewrapper(logger, prefix="[LDAP]")

    return  ldap.initialize(
        uri,
        trace_level=logger.ldap_level,
        trace_file=trace_file
        )


def _ldap_starttls(logger, connection, check_cert):
    #Set at connection level.
    _set_check_cert(check_cert, connection)

    try:
        connection.start_tls_s()
    except ldap.SERVER_DOWN, e:
        logger.error("""TLS negociation: python-ldap says: "%s".""" % e.args[0]['desc'])
        raise TLSError(e)
    except ldap.UNAVAILABLE, e:
        # CRITICAL, but we continue
        logger.error("*** LDAP server cannot perform TLS operation ***" \
              "If you are connecting to an AD server, " \
              "look if the Active Directory Service has access to " \
              "certificates (certificates MMC snap-in)." \
              "Continuing")


class Server(object):
    def __init__(self, logger, domain, name, conf):
        """
        Abstract base class:
            abstract method: search_group_list_s

        mandatory parameter in conf:
            ldapuri

        optionnal:
            bind_dn (None), bind_pw (None), version ('3')
        """
        self.domain = domain
        self.name = name
        self.con = None

        setmeupaslogger(self, logger)
        self.logger = logger
        self.readopts(conf)

        self.shortname = "[domain %s server:%s uri:%s]" % (domain.realm, name, self.uri)

        self.grouplist_cache = NNDCache(self)
        self.page_size = 50
        self.use_pagination = 1
        if self.use_pagination:
            self.lc = SimplePagedResultsControl(ldap.LDAP_CONTROL_PAGE_OID, True, (self.page_size, ''))
        # TODO else self.lc = ???
        ldap.set_option(ldap.OPT_REFERRALS, 0)

    def readopts(self, conf):
        #uri required
        self.uri = conf.get(self.name, 'ldapuri')

        #other values have defaults
        rsections = (self.name, self.domain.conf_section, "global")

        for option, default in REC_INT_OPTIONS:
            setattr(self, option, conf.rgetint(rsections, option, default))

        for option, default in REC_STR_OPTIONS:
            setattr(self, option, conf.rget(rsections, option, default))

        for option, default in REC_BOOL_OPTIONS:
            setattr(self, option, conf.rgetboolean(rsections, option, default))

    def testconf(self):
        self._ssl_tls()

    def _binderror(self, err):
        if isinstance(err, ldap.SERVER_DOWN):
            self.critical(
                "[%s] Bind failed: server down" % self.shortname
                )
            return
        self.critical(
            "[%s] Bind failed (ldap error): %s" % (self.shortname, err)
            )

    def initialize(self):
        self.initialized = self._initialize()
        return self.initialized

    def unbind(self):
        self.initialized = False
        if not self.con:
            return False
        try:
            self.con.unbind()
        except Exception:
            return False
        else:
            self.info("Deconnected from %s" % self.shortname)
            return True

    def _initialize(self):
        self.unbind()

        self.con = self.init_connection()

        self.info("[%s] Binding..." % self.shortname)

        try:
            self.con.simple_bind_s(self.bind_dn, self.bind_pw)
        except (ldap.LDAPError), err:
            self._binderror(err)
            return False
        else:
            self.info("[%s] Bind succeeded \o/" % self.shortname)
            return True

    def _ldapversion(self):
        if self.tls:
            self.debug("Forcing LDAPv3 because of TLS")
            return 3
        return self.version

    def _ssl_tls(self):
        ssl = self.uri.lower().startswith('ldaps://')

        if ssl and self.tls:
            raise ValueError("Choose ssl (ldaps://) or tls, not both")

        clear = not (ssl or self.tls)
        return clear, ssl, self.tls

    def init_connection(self):
        clear, ssl, tls = self._ssl_tls()

        if not clear and self.checkcert:
            check_cert_option = ldap.OPT_X_TLS_HARD
        else:
            check_cert_option = ldap.OPT_X_TLS_NEVER

        check_cert = ((not clear) and (self.checkcert))

        con = _ldapconnect_basic(
            self.logger,
            self._ldapversion(),
            self.uri,
            check_cert
            )

        if tls:
            _ldap_starttls(self, con, check_cert)

        return con

    def auth(self, username, password):
        user_dn, timeleft = self.auth_user_dn(username, self.auth_timeout)

        self.info("[%s] Authenticate user %s: %s" % (self.shortname, username, user_dn))

        #make connection
        calltime = time()
        con = self.init_connection()

        #recompute timeleft
        timeleft, now = updatetimes(calltime, timeleft)
        if timeleft <= 0:
            raise NNDError("Time out while connecting to %s" % self.uri)

        con.network_timeout = timeleft
        try:
            con.simple_bind_s(user_dn, password)
        except (ldap.LDAPError), err:
            self._binderror(err)
            return False

        con.unbind()
        return True

    def _process_groups(self, uid, groups):
        groups = [
            '%s@%s' % (format_item(name), domain)
            for name, domain in groups
            if ("\n" not in name) and ("@" not in name)]
        self.grouplist_cache[uid] = groups
        return groups

    def groups(self):
        """
        Unsorted, maybe with dupes
        """
        uid = ALLGROUPS

        # try cache
        result = self.grouplist_cache[uid]
        if result is not None:
            return result

        calltime = time()
        groups = self._groups(calltime, self.groups_timeout)
        return self._process_groups(uid, groups)

    def usergroups(self, uid):
        """
        Unsorted, maybe with dupes
        """
        # try cache
        result = self.grouplist_cache[uid]
        if result is not None:
            return result

        calltime = time()
        groups = self._usergroups(uid, calltime, self.usergroups_timeout)
        return self._process_groups(uid, groups)

    def timedasyncreq(self, basedn, scope, ldapfilter, attrs, calltime, timeleft):
        """
        Returns (msgid, timeleft, now)
        """
        if timeleft <= 0:
            raise NNDError("Time allowed for request ran out")

        msgid = self.con.search_ext(
            basedn,
            scope,
            ldapfilter,
            attrs,
            serverctrls=[self.lc],
            timeout=timeleft
        )
        timeleft, now = updatetimes(calltime, timeleft)
        return msgid, timeleft, now

    def timedsyncreq(self, basedn, scope, ldapfilter, attrs, calltime, timeleft):
        """
        Returns (result, timeleft, now)
        """
        if timeleft <= 0:
            raise NNDError("Time allowed for request ran out")

        result = self.con.search_st(
            basedn,
            scope,
            ldapfilter,
            attrs,
            timeout=timeleft
            )
        timeleft, now = updatetimes(calltime, timeleft)
        return result, timeleft, now

    def timedpagedreq(self, basedn, scope, ldapfilter, attrs, calltime, timeleft):
        result = list(self.iter_paginated_request(basedn, scope, ldapfilter, timeleft, attrs=attrs))
        timeleft, now = updatetimes(calltime, timeleft)
        return result, timeleft, now


    def iter_paginated_request(self, basedn, scope, ldapfilter, timeleft, attrs=None):
        # Send search request
        msgid, timeleft, now = self.timedasyncreq(
            basedn,
            scope,
            ldapfilter,
            attrs,
            time(),
            timeleft
            )
        pages = 0
        while True:
            pages += 1
            self.debug("Getting page %d" % (pages,))
            item = self.con.result3(msgid, timeout=timeleft) # rtype, rdata, rmsgid, serverctrls
            rdata = item[1]
            serverctrls = item[3]
            self.debug('%d results' % len(rdata))
            for d in rdata:
                yield d
            pctrls = [
              c
              for c in serverctrls
              if c.controlType == ldap.LDAP_CONTROL_PAGE_OID
            ]
            if pctrls:
                item = pctrls[0].controlValue # est, cookie
                cookie = item[1]
                if cookie:
                    self.lc.controlValue = (self.page_size, cookie)
                    msgid, timeleft, now = self.timedasyncreq(
                        basedn,
                        scope,
                        ldapfilter,
                        attrs,
                        time(),
                        timeleft
                    )
                else:
                    break
            else:
                self.warning("Server ignores RFC 2696 control.")
                break

    # --- Abstract methods --------------------

    def _groups(self, calltime, timeleft):
        """
        Return a list of (name, domain). It can be a generator.
        """
        raise NotImplementedError()

    def _usergroups(self, calltime, uid, timeleft):
        """
        Return a list of (name, domain). It can be a generator.
        """
        raise NotImplementedError()

    def auth_user_dn(self, username, timeleft):
        raise NotImplementedError()

