"""
Copyright 2010 EdenWall Technologies

Written by Pierre Chifflier <chifflier@edenwall.com>
"""

from select import select
from threading import Thread
import ldap
import socket
from nucentral.common.error import writeError

from .log import setmeupaslogger
from .utils import socketinfo, formatsockerror, shortsockinfo, accept_username
from .error import NNDError
from .formatter import END_OF_REQUEST, OK, ERROR, list_response

SELECT_FREQ = 0.5 #seconds

RECV_SIZE = 4096

def _pop_requests(req_list, full):
    """
    req_list: a list of received requests
    full: is the last chunk in the list a full request?
    """
    last = len(req_list) - 1 if full else len(req_list) - 2
    for index in xrange(last):
        yield req_list[0]
        del req_list[0]

def _queue_requests(req_list, data):
    """
    Puts received data in supplied requests list.
    If the list is not empty, the first detected request is appended
    to the last item of the list, assuming the list only contains 1 item.

    Returns: was the last received chunk a finished request?
    """
    # decode string assuming it is utf-8,
    #which is RFC enforced for ldap v3
    data = unicode(data,"utf-8")
    parsed = data.split(END_OF_REQUEST)
    if parsed:
        if req_list:
            #There was something left in req_list: it is by definition an
            #unfinished request
            req_list[0] += parsed[0]
            del parsed[0]
        req_list += parsed

    return data.endswith(END_OF_REQUEST)

class NNDWorkerThread(Thread):
    """Thread object which execute received commands"""

    VERSION_STRING = "2"

    def __init__(self, logger, domains, sock, addr, default_domain):
        """
            domains: dict of domains (realm => domain)
            sock: socket object
            addr: the address bound to the socket on the other end of the connection
        """
        Thread.__init__(self)
        setmeupaslogger(self, logger)
        self.domains = domains
        self.conn = sock
        sock.setblocking(0)
        self.addr = addr
        self.default_domain = default_domain
        self.shortsockinfo = shortsockinfo(self.conn)
        self.client_version = None
        self._stop = False

    def _iter_requests(self):
        request_queue = []
        full = True
        while not self._stop:
            for request in _pop_requests(request_queue, full):
                yield request

            rfd, wfd, efd = select([self.conn.fileno()], (), (), SELECT_FREQ)
            if not rfd:
                continue
            data = self.conn.recv(RECV_SIZE)
            if not data:
                break
            full = _queue_requests(request_queue, data)

    def run(self):
        """execute commands from client while connection not ended"""

        try:
            for data in self._iter_requests():
                self.execute(data)

        except socket.error, e:
            self.warning(formatsockerror(self.conn, e))
        finally:
            self.debug("Closing connection with client %s" % socketinfo(self.conn))
            self.conn.close()

    def stop(self):
        self._stop = True

    def _implicit_hello(self):
        #first ever client version integrated in EdenWall: 2
        self.client_version = "2"

    def _runcommand(self, command, command_handler, *parameter):
        err = None
        try:
            code, data = command_handler(*parameter)
        except NNDError, err:
            code = ERROR
            data = str(err)
        except ldap.LDAPError, err:
            code = ERROR
            data = "Search failed (internal error, look at server logs) %s" % err
        except ldap.SERVER_DOWN,  err:
            code = ERROR
            data = "Search failed (server down) %s" % err
        except TypeError, err:
            code = ERROR
            #The replacement occurs to prevent doc of containing END_OF_REQUEST

            data = "Error in service, or improper use thereof.\n" \
                "Documentation for service '%s':\n%s" % (
                    command, command_handler.__doc__
                    )
        except Exception, err:
            code = ERROR
            data = "Internal error"
        finally:
            if err:
                writeError(err, "Error in command %s:" % command)
        return code, data


    def execute(self, data):
        """execute one command specified in data"""
        if not len(data):
            self.send(ERROR, "no query")
            return

        commands = data.split(' ', 1)
        command = commands[0].lower()
        parameter = commands[1:]

        if not self.client_version and command not in ("hello", "help"):
            self._implicit_hello()

        func_name = 'command_%s' % command

        if not hasattr(self, func_name):
            code = ERROR
            data = "unknown command: '%s'" % command
        else:
            command_handler = getattr(self, func_name)
            code, data = self._runcommand(command, command_handler, *parameter)
        self.send(code, data)
        return

        try:
            code, data = command_handler(*parameter)
            err = None
        except NNDError, err:
            code = ERROR
            data = str(err)
        except ldap.LDAPError, e:
            code = ERROR
            data = "Search failed (internal error, look at server logs) %s" % e
        except ldap.SERVER_DOWN,  e:
            code = ERROR
            data = "Search failed (server down) %s" % e
        except TypeError, err:
            code = ERROR
            #The replacement occurs to prevent doc of containing END_OF_REQUEST

            data = "Error in service, or improper use thereof.\n" \
                "Documentation for service '%s':\n%s" % (
                    command, command_handler.__doc__
                    )
            raise
        except Exception, err:
            code = ERROR
            data = "Internal error"
        finally:
            if err:
                writeError(err, "Error in command %s:" % command)

    def send(self, code, result):
        """send result to client"""
        # decode string from directory, assuming it is utf-8
        data = u"%s %s\n" % (
            unicode(code), unicode(result).replace(END_OF_REQUEST, " ")
            )
        # always format output as utf-8
        data = data.encode("utf-8")
        self.debug("[%s] answer: %r" % (self.shortsockinfo, data))
        self.conn.send(data)

    # Handler for commands :
    def command_help(self, arg):
        return OK, ("Allowed commands: "
            "help, hello <id>, usergroups <user>[@domain], "
            "groups [domain], domains")

    def command_hello(self, arg):
        """
        We only support version 2
        """
        self.client_version = arg.strip()
        if not arg:
            # Require at least one arg but tolerate any string we might
            # want the client to send later.
            return ERROR, "command 'hello' takes one argument."
        return OK, "%s" % self.VERSION_STRING

    def command_usergroups(self, user):
        """
usage:
    usergroups $user

returns groups for $user

If $user expression doesn't match 'user@domain', default domain used
        """
        if not user:
            return ERROR, "command usergroups take one argument: user[@domain]"

        if '@' in user:
            user, domain_name = user.split('@', 1)
        else:
            domain_name = self.default_domain

        domain = self.getDomain(domain_name)
        group_list = domain.usergroups(user)
        return list_response(group_list)

    def command_groups(self):
        """return all groups for all domains"""
        all_groups = []
        for domain in self.domains.itervalues():
            domain_groups = domain.groups()
            all_groups.extend(domain_groups)
        return list_response(all_groups)

    def command_domains(self):
        """return all available domains"""
        # TODO return the list of domains defined in file configuration

        ordered_list = [self.default_domain,]
        for domain in self.domains:
            if domain != self.default_domain.lower():
                ordered_list.append(domain)
        return list_response(ordered_list)

    def getDomain(self, name):
        if not name:
            try:
                return self.domains[self.default_domain.lower()]
            except Exception:
                raise
        name = name.lower()
        try:
            return self.domains[name]
        except KeyError:
            raise NNDError("Unknown domain: %s" % name)

    def _auth_usage(self, message=""):
        if message:
            message = "usage: auth username[@domain],password (%s)" % message
        else:
            message = "usage: auth username[@domain],password"
        return ERROR, message

    def command_auth(self, arg):
        """check user password

        Syntax:
            auth username[@domain],password
        First ',' separates username[@domain] & password.
        Last '@' in username[@domain] separates username and domain
        """
        arg = arg.split(',', 1)
        if len(arg) != 2:
            return self._auth_usage()
        password = arg[1]

        if not password:
            return self._auth_usage("password empty?")

        #Split user from domain
        user_domain = arg[0].rsplit('@', 1)
        if not user_domain:
            return self._auth_usage()
        if len(user_domain) != 2 or not user_domain[1]:
            #If there was no such thing as a domain, we use default
            user, domain_name = user_domain[0], self.default_domain
        else:
            user, domain_name = user_domain

        if not user:
            return self._auth_usage("user empty?")

        if not accept_username(user):
            return self._auth_usage("invalid username: '%s'" % user)

        domain = self.getDomain(domain_name)

        valid = domain.auth(user, password)

        if valid:
            return OK, "%s@%s" % (user, domain_name)
        else:
            return ERROR, "Invalid login or password"

    def command_unbind(self):
        """Disconnect the daemon from all servers.

        Syntax:
            unbind
        """
        for domain in self.domains.itervalues():
            domain.unbind()
        return OK, "OK"

