# -*- coding: utf-8 -*-
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from PyQt4.QtGui import QTableWidgetItem, QMessageBox
from PyQt4.QtCore import Qt, SIGNAL, QVariant

from nucentral_admin.acl_ui import Ui_AclTab
from nucentral.common import tr
from nucentral.qt.roles_name import display_name

from .base_tab import BaseTab

class Application(object):
    def __init__(self, name):
        self.name = name
        self.roles = {}

    def add_role(self, role_name):
        role = Role(self, role_name)
        self.roles[role.perm] = role
        return role

    def get_perm_role(self, perm_name):
        return self.roles[perm_name]

    def __repr__(self):
        return '<Application name=%s>' % self.name

    def __eq__(self, cmp):
        if isinstance(cmp, (str,unicode)):
            return cmp == self.name
        if isinstance(cmp, Application):
            return cmp.name == self.name
        return False

class Role(object):
    def __init__(self, app, name):
        self.app = app
        self.name = name
        self.perm = Role.split_name(name)[-1]
        self.deps = []

    def __repr__(self):
        return '<Role name=%s>' % self.name

    @staticmethod
    def split_name(role_name):
        return role_name.rsplit('_', 1)

class AclTab(BaseTab):
    ICON = ':/icons/Acl.png'
    # Regit can be gruik too. Add necessary for translation.
    PERMS_LIST = [
        tr('Write'),
        tr('Read'),
        tr('Anonymous'),
        tr('Debug'),
        ]

    def __init__(self, client, window):
        BaseTab.__init__(self, window)

        self.window = window
        self.ui = Ui_AclTab()
        self.ui.setupUi(self)

        # Hide the host list as its only used on multisite permissions
        self.ui.hostLabel.hide()
        self.ui.hostList.hide()

        self.client = client

        self.connect(self.ui.applyButton, SIGNAL('clicked()'), self.applyAcls)
        self.connect(self.ui.cancelButton, SIGNAL('clicked()'), self.resetAcls)
        self.connect(self.ui.removeGroupButton, SIGNAL('clicked()'), self.removeGroup)
        self.connect(self.ui.addGroupButton, SIGNAL('clicked()'), self.addGroup)
        self.connect(self.ui.tablePermissions, SIGNAL("itemChanged(QTableWidgetItem *)"), self._item_changed)

        roles = self.listAllRoles()
        self.group = ''
        self.apps = []
        self.perms = []
        for role_name, deps in roles.iteritems():
            app_name, perm_name = Role.split_name(role_name)

            try:
                app_i = self.apps.index(app_name)
                app = self.apps[app_i]
            except ValueError:
                app = Application(app_name)
                self.apps.append(app)

            role = app.add_role(role_name)
            role.deps = deps

            try:
                self.perms.index(perm_name)
            except ValueError:
                self.perms.append(perm_name)

        self.ui.tablePermissions.setRowCount(len(self.apps))
        self.ui.tablePermissions.setColumnCount(len(self.perms))
        self.ui.tablePermissions.setHorizontalHeaderLabels([self.tr(s.title()) for s in self.perms])
        self.ui.tablePermissions.setVerticalHeaderLabels([ display_name(s.name) for s in self.apps])
        for row, app in enumerate(self.apps):
            for column, perm in enumerate(self.perms):
                try:
                    role = app.get_perm_role(perm)
                except KeyError:
                    continue

                item = QTableWidgetItem()
                item.setFlags(Qt.ItemIsUserCheckable|Qt.ItemIsEnabled|Qt.ItemIsSelectable)
                item.setCheckState(Qt.Unchecked)
                item.setData(Qt.UserRole, QVariant(role))
                self.ui.tablePermissions.setItem(row, column, item)

        self.updateGroups()

    def updateGroups(self):
        self.disconnect(self.ui.groupList, SIGNAL('currentIndexChanged(int)'), self.groupListChanged)

        self.ui.groupList.clear()
        groups = self.client.call('auth', 'listGroups')
        for group in groups:
            self.ui.groupList.addItem(group)

        acls = self.getAcl()
        for id, group, role in acls:
            if self.ui.groupList.findText(group) < 0:
                self.ui.groupList.addItem(group)

        if self.group:
            i = self.ui.groupList.findText(self.group)
            if i < 0:
                i = self.ui.groupList.count() - 1
                self.group = unicode(self.ui.groupList.itemText(i))
                self.resetAcls()

            self.ui.groupList.setCurrentIndex(i)
        else:
            i = self.ui.groupList.currentIndex()
            if i >= 0:
                self.group = unicode(self.ui.groupList.itemText(i))
            self.resetAcls()

        self.connect(self.ui.groupList, SIGNAL('currentIndexChanged(int)'), self.groupListChanged)

    def groupListChanged(self, index):
        if self.ui.applyButton.isEnabled() and \
            QMessageBox.warning(self, tr("Save your changes"),
                                      tr("Do you want to save changes made to the '%s' group?") % self.group,
                                      QMessageBox.Yes|QMessageBox.No) == QMessageBox.Yes:
            self.applyAcls()

        if index >= 0:
            self.group = unicode(self.ui.groupList.itemText(index))
            self.ui.removeGroupButton.setEnabled(True)
        else:
            self.group = ''
            self.ui.removeGroupButton.setEnabled(False)

        self.resetAcls()

    def removeGroup(self):
        if not self.group:
            return

        if QMessageBox.warning(self, tr("Delete a group"),
                                     tr("Are you sure you want to delete the '%s' group?") % self.group,
                                     QMessageBox.Yes|QMessageBox.No) != QMessageBox.Yes:
            return

        self.client.call('auth', 'delGroup', self.group)
        self.client.call('acl', 'deleteAcl', self.group)
        self.updateGroups()
        self.emit(SIGNAL('groups_updated'))
        self.setModified(True)

    def addGroup(self):
        name = unicode(self.ui.groupEdit.text())
        if not name:
            return

        if self.ui.groupList.findText(name, Qt.MatchExactly|Qt.MatchCaseSensitive) >= 0:
            raise Exception(unicode(self.tr('The %s group already exists.')) % name)

        self.client.call('auth', 'addGroup', name)
        self.ui.groupList.addItem(name)
        self.ui.groupList.setCurrentIndex(self.ui.groupList.count() - 1)
        self.ui.groupEdit.clear()
        self.emit(SIGNAL('groups_updated'))
        self.setModified(True)

    # Overriden in acls_multisite
    def resetAcls(self):
        self.disconnect(self.ui.tablePermissions, SIGNAL("itemChanged(QTableWidgetItem *)"), self._item_changed)

        if self.group:
            permissions = self.getAcl(self.group)
            roles = [role for id, g, role in permissions]
            self.ui.removeGroupButton.setEnabled(True)
        else:
            self.ui.removeGroupButton.setEnabled(False)

        for row in xrange(self.ui.tablePermissions.rowCount()):
            for column in xrange(self.ui.tablePermissions.columnCount()):
                item = self.ui.tablePermissions.item(row, column)
                # perm doesn't exist for this application
                if not item:
                    continue

                if not self.group:
                    item.setFlags(Qt.ItemIsUserCheckable|Qt.ItemIsSelectable)
                    item.setCheckState(Qt.Unchecked)
                else:
                    item.setFlags(Qt.ItemIsUserCheckable|Qt.ItemIsEnabled|Qt.ItemIsSelectable)
                    role = item.data(Qt.UserRole).toPyObject()
                    if role.name in roles:
                        item.setCheckState(Qt.Checked)
                    else:
                        item.setCheckState(Qt.Unchecked)

        self.connect(self.ui.tablePermissions, SIGNAL("itemChanged(QTableWidgetItem *)"), self._item_changed)
        self.setModified(False)

    def _item_changed(self, item):
        role = item.data(Qt.UserRole).toPyObject()
        if not self.group:
            return

        self.setModified(True)

        if item.checkState() == Qt.Checked:
            # Set all dependances.
            for dep in role.deps:
                dep_app, dep_perm = Role.split_name(dep)
                i = self.ui.tablePermissions.item(self.apps.index(dep_app), self.perms.index(dep_perm))
                if not i.checkState():
                    i.setCheckState(Qt.Checked)

        else:
            for row, app in enumerate(self.apps):
                for r in app.roles.itervalues():
                    if role.name in r.deps:
                        i = self.ui.tablePermissions.item(row, self.perms.index(r.perm))
                        if i.checkState():
                            i.setCheckState(Qt.Unchecked)

    def setModified(self, modified):
        self.ui.applyButton.setEnabled(modified)
        self.ui.cancelButton.setEnabled(modified)
        self.window.setCentralModified(modified)

    # Overriden in acls_multisite
    def applyAcls(self):
        if not self.group:
            return

        for row, app in enumerate(self.apps):
            for r in app.roles.itervalues():
                i = self.ui.tablePermissions.item(row, self.perms.index(r.perm))
                if i.checkState():
                    self.setAcl(self.group, r.name)
                else:
                    self.deleteAcl(self.group, r.name)
        self.setModified(False)

    # Overriden in acls_multisite
    def getAcl(self, group='', role=''):
        return self.client.call('acl', 'getAcl', group, role)

    # Overriden in acls_multisite
    def setAcl(self, group, name):
        self.client.async().call('acl', 'setAcl', group, name, callback=lambda r: True)

    # Overriden in acls_multisite
    def deleteAcl(self, group, name):
        self.client.async().call('acl', 'deleteAcl', group, name, callback=lambda r: True)

    # Overriden in acls_multisite
    def listAllRoles(self):
        return self.client.call('acl', 'listAllRoles')

