# -*- coding: utf-8 -*-
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from PyQt4.QtCore import Qt, SIGNAL, QVariant
from PyQt4.QtGui import QHeaderView, QMessageBox, QTableWidgetItem

from nucentral_admin.users_ui import Ui_UsersTab
from nucentral.common import tr

from .edit_user import EditUserDialog
from .base_tab import BaseTab

class UsersTab(BaseTab):
    ICON = ':/icons/users.png'

    key2label = {'secret':       tr('Shared secret'),
                 'hostname':     tr('Hostname'),
                 'uri':          tr('URI'),
                 'login':        tr('Login'),
                 'basedn':       tr('Base DN'),
                 'group_attribute': tr('Group attribute'),
                 'binddn':       tr('Bind DN'),
                 'bindpw':       tr('Bind PW'),
                 'additional_filter': tr('Additional filter'),
                 'authfile':     tr('Users file'),
                 'groupfile':    tr('Groups file'),
                }

    def __init__(self, client, window, parent=None):
        BaseTab.__init__(self, parent)

        self.ui = Ui_UsersTab()
        self.ui.setupUi(self)
        self.ui.groupBox_2.hide()
        self.ui.tableUsers.horizontalHeader().setResizeMode(QHeaderView.Stretch)
        self.ui.ldap.horizontalHeader().setResizeMode(QHeaderView.Stretch)
        self.ui.ldap.horizontalHeader().setResizeMode(QHeaderView.Stretch)

        self.connect(self.ui.userAddButton, SIGNAL("clicked()"), self._cb_add_user)
        self.connect(self.ui.ldap, SIGNAL("cellChanged(int, int)"), self._cb_auth_parameters_changed)
        self.connect(self.ui.cancelButton, SIGNAL("clicked()"), self.refreshSecondaryList)
        self.connect(self.ui.applyButton, SIGNAL("clicked()"), self._cb_apply_auth)

        self.connect(self.ui.userEditButton, SIGNAL("clicked()"), self.edit)
        self.connect(self.ui.tableUsers, SIGNAL("cellDoubleClicked(int, int)"), self.edit)
        self.connect(self.ui.userDelButton, SIGNAL("clicked()"), self.remove)

        self.client = client
        self.window = window

        types = self.client.call('auth', 'getAuthTypes')
        self.ui.typeEdit.addItem('-- None --', QVariant(''))
        for auth_type in types:
            if auth_type == 'file': continue
            self.ui.typeEdit.addItem(auth_type.title(), QVariant(auth_type))

        self.refreshList()
        self.refreshSecondaryList()

    def edit(self, *args):
        username = self.getSelectedUser()
        if username is None:
            return

        dialog = EditUserDialog(self.client, self)
        if dialog.editUser(username):
            self.refreshList()

    def remove(self):
        username = self.getSelectedUser()
        if username is None:
            return

        if QMessageBox.question(self, tr("Delete a user"),
                                tr("Are you sure you want to delete the user '%s'?") % username,
                                QMessageBox.Yes|QMessageBox.No) != QMessageBox.Yes:
            return


        if username == self.client.login and \
           QMessageBox.warning(self, tr("Warning"),
                               tr("You are logged in on the account you are trying to delete! Continue?"),
                               QMessageBox.Yes|QMessageBox.No) != QMessageBox.Yes:
            return
        self.client.call('auth', 'delUser', username)
        self.refreshList()

    def getSelectedUser(self):
        """
        return None if there is no selection
        """
        row = self.ui.tableUsers.currentRow()
        if -1 == row:
            return
        return unicode(self.ui.tableUsers.item(row, 0).text())

    def refreshList(self):
        users = self.client.call('auth', 'listUsers')
        self.ui.tableUsers.setRowCount(len(users))
        for row, (username, method, groups) in enumerate(users):
            item = QTableWidgetItem(username)
            item.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
            self.ui.tableUsers.setItem(row, 0, item)

            item = QTableWidgetItem(method)
            item.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
            self.ui.tableUsers.setItem(row, 1, item)

            item = QTableWidgetItem(', '.join(groups))
            item.setFlags(Qt.ItemIsEnabled|Qt.ItemIsSelectable)
            self.ui.tableUsers.setItem(row, 2, item)

        self.emit(SIGNAL('groups_updated'))

    def _cb_add_user(self):
        dialog = EditUserDialog(self.client, self)
        if dialog.addUser():
            self.refreshList()

    def refreshSecondaryList(self):
        self.disconnect(self.ui.typeEdit, SIGNAL('currentIndexChanged(int)'), self._cb_auth_type_changed)

        auth_type, params = self.client.call('auth', 'getSecondaryAuth')
        if auth_type:
            i = self.ui.typeEdit.findData(QVariant(auth_type))
            if i < 0:
                self.ui.typeEdit.addItem(auth_type.title(), QVariant(auth_type))
                i = self.ui.typeEdit.count() - 1
            self.ui.typeEdit.setCurrentIndex(i)

        self.setParameters(auth_type, params)

        self.connect(self.ui.typeEdit, SIGNAL('currentIndexChanged(int)'), self._cb_auth_type_changed)
        self.setModified(False)

    def setParameters(self, auth_type, params):
        self.ui.ldap.setHidden(True)
        self.ui.radius.setHidden(True)
        if 'ldap' == auth_type:
            self.setLdap(params)
        if 'radius' == auth_type:
            self.setRadius(params)

    def setLdap(self, params):
        self.ui.ldap.setVisible(True) # hide / show ?
        self.ui.ldap.setRowCount(len(params))
        for row, (key, value) in enumerate(params.iteritems()):
            item = QTableWidgetItem(self.key2label.get(key, key))
            item.setData(Qt.UserRole, QVariant(key))
            item.setFlags(Qt.ItemIsEnabled)
            self.ui.ldap.setItem(row, 0, item)

            item = QTableWidgetItem(value)
            self.ui.ldap.setItem(row, 1, item)

    def setRadius(self, params):
        self.ui.radius.setVisible(True)

    def _cb_auth_parameters_changed(self, row, col):
        self.setModified(True)

    def _cb_auth_type_changed(self, i):
        self.setModified(True)

        auth_type = unicode(self.ui.typeEdit.itemData(i).toString())
        if not auth_type:
            self.setParameters('', {})
            return

        keys = self.client.call('auth', 'getAuthParameters', auth_type)
        params = dict([(key, '') for key in keys])

        self.setParameters(auth_type, params)

    def _cb_apply_auth(self):
        index = self.ui.typeEdit.currentIndex()
        auth_type = unicode(self.ui.typeEdit.itemData(index).toString())
        if 'ldap' == auth_type:
            conf = self.ldapConf()
        elif 'radius' == auth_type:
            conf = self.radiusConf()
        else:
            # deconfigure
            auth_type = ''
            conf = {}

        self.client.call('auth', 'setSecondaryAuth', auth_type, conf)
        self.setModified(False)

    def ldapConf(self):
        params = {}
        for row in xrange(self.ui.ldap.rowCount()):
            key = self.ui.ldap.item(row, 0)
            value = self.ui.ldap.item(row, 1)
            params[unicode(key.data(Qt.UserRole).toString())] = unicode(value.text())
        return params

    def radiusConf(self):
        params = {}
        # TODO
        return params

    def setModified(self, modified):
        self.ui.applyButton.setEnabled(modified)
        self.ui.cancelButton.setEnabled(modified)
        self.window.setCentralModified(modified)

