# -*- coding: utf-8 -*-
"""
Copyright (C) 2010-2011 EdenWall Technologies
Written by François Toussenel <ftoussenel AT edenwall.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

$Id$

"""

from __future__ import with_statement

from crypt import crypt
import os
import random
import re

from nucentral.common.process import formatCommand
from nucentral.core import events
from nucentral.backend import Component
from nucentral.backend import tr
from nucentral.backend.exceptions import NuCentralError
from nucentral.backend.process import runCommand
from nucentral.backend.variables_store import VariablesStore

DEFAULT_PASSWORD = "admin"
PASSWORD_MINCHAR = 4
RE_ADMIN_SHADOW = re.compile(r"^admin:(?P<password>[$]5[$](?P<salt>[^$]+)[^:]+)")

def create_salt(n=8):
    salt = ""
    for i in range(n):
        salt += random.choice(
            "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")
    return salt

class ConsoleAccessComponent(Component):
    """
    Control the console access.
    """
    NAME = "console_access"
    VERSION = "1.0"
    API_VERSION = 1
    ROLES = {
        'nucentral_admin': set((
                'applyConsoleAccessConfig',
                'consoleAccessHasDefaultPassword',
                'getConsoleAccessConfig',
                'setConsoleAccessConfig')),
    }

    def __init__(self):
        Component.__init__(self)

    def init(self, core):
        self.core = core
        self.notify = core.notify
        Component.init(self, core)
        self.console_access_conf_filename = os.path.join(
            self.core.config.get("CORE", "vardir"),
            "console_access.xml")
        self.read_config()

    def create_default_config(self):
        self.console_access_cfg = VariablesStore()
        self.console_access_cfg.set("enabled", False)
        self.console_access_cfg.set("password",
                                    self.crypt_password(DEFAULT_PASSWORD))
        self.warning(
            tr('Console access not configured, default values loaded.'))

    def read_config(self, *args, **kwargs):
        if os.path.exists(self.console_access_conf_filename):
            self.console_access_cfg = VariablesStore()
            self.console_access_cfg.load(self.console_access_conf_filename)
        else:
            self.create_default_config()

    def apply(self):
        self.read_config()
        password = self.console_access_cfg.get("password")
        command = ["usermod", "-p", password, "admin"]
        cmdstr = formatCommand(("usermod", "-p", "***", "admin"))
        runCommand(self, command, cmdstr=cmdstr)
        if self.console_access_cfg.get("enabled"):
            runCommand(self, "chsh -s /bin/rbash admin".split())
        else:
            runCommand(self, "chsh -s /bin/false admin".split())
        self.generateTemplates(
            {"enabled": self.console_access_cfg.get("enabled")},
            {"/etc/motd": ("root:root", "0644", "/etc/motd")})

    def save_config(self):
        self.console_access_cfg.save(self.console_access_conf_filename)

    def get_ports(self):
        return [{'proto': 'tcp',
                 'port': 22},]

    def crypt_password(self, clear_text_password, salt=None):
        if salt is None:
            salt = create_salt()
        encrypted = crypt(clear_text_password, "$5$%s" % salt)
        return encrypted

    def compare_password_in_shadow(self, tested_password):
        password_in_shadow = ""
        with open("/etc/shadow") as fd:
            for line in fd.readlines():
                m = RE_ADMIN_SHADOW.search(line)
                if m:
                    salt = m.group("salt")
                    password_in_shadow = m.group("password")
                    if self.crypt_password(tested_password, salt) == \
                            password_in_shadow:
                        return True
        return False

    # Services:
    def service_getConsoleAccessConfig(self, context):
        self.read_config()
        return {"enabled": self.console_access_cfg.get("enabled"),
                "password": self.console_access_cfg.get("password")}

    def service_setConsoleAccessConfig(self, context, config):
        try:
            config.get("password", "admin").decode()
        except UnicodeEncodeError:
            raise NuCentralError(
                "Only ASCII characters are allowed for admin's password for console access.")
        try:
            self.console_access_cfg.set("enabled", config.get("enabled", True))
            # If the password is empty in the config sent by the frontend, it
            # must be left alone:
            if config.get("password", ""):
                if len(config["password"]) < PASSWORD_MINCHAR:
                    raise NuCentralError(tr(
                        "The password for console access must have at least %d characters.")
                                         % PASSWORD_MINCHAR)
                if "confirm password" in config and \
                        config["password"] !=  config["confirm password"]:
                    raise NuCentralError(tr(
                        "The password and its confirmation are different."))
                self.console_access_cfg.set(
                    "password", self.crypt_password(config["password"]))
            else:
                raise NuCentralError(tr(
                        "You must (re)set the admin password."))
        except KeyError, err:
            raise NuCentralError(tr("Your EAS did not send the %s element of the Console Access configuration.") % err)

        self.save_config()

        self.notify.emit(self.NAME, 'configModified')

    def service_applyConsoleAccessConfig(self, context):
        # Need this service since the frontend is not in NuConf but in
        # nucentral-admin.
        self.apply()

    def service_consoleAccessHasDefaultPassword(self, context):
        """ Return True iff system user "admin" has password "admin". """
        return self.compare_password_in_shadow(DEFAULT_PASSWORD)

    def service_runtimeFiles(self, context):
        return {
            'added': [(self.console_access_conf_filename, 'text')],
            'deleted': (self.console_access_conf_filename,)
        }

    def service_runtimeFilesModified(self, context):
        self.read_config()
        self.apply()

    def formatServiceArguments(self, service, arguments):
        if service == 'setConsoleAccessConfig':
            arguments = list(arguments)
            # hide the new configuration because it contains a password in
            # clear text
            arguments[0] = '***'
        return Component.formatServiceArguments(self, service, arguments)
