"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Romain Bignon <romain AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id$
"""

from twisted.internet import reactor
#from twisted.internet.threads import deferToThread
from twisted.web.xmlrpc import Proxy
from twisted.web import xmlrpc

from nucentral.common.transport import checkComplexType
from nucentral.core.ssl_strict import StrictOpenSSLClientContextFactory
from nucentral.core.getter import getUnicode
from nucentral.backend.error import NuCentralError
from nucentral.backend import tr
from nucentral.util.roles import ALL_MULTISITE_ROLES
import new

class ClientXMLRPC(Proxy):

    def __init__(self, core, name, acls, host, port, use_ssl, ssl_cert = '', ssl_key = '', ssl_ca = ''):
        Proxy.__init__(self, "https://%s:%d/RPC2" % (str(host), int(port)))

        self.core = core
        self.name = name
        self.use_ssl = use_ssl
        self.acls = acls

        if use_ssl:
            self.context = StrictOpenSSLClientContextFactory()
            self.context.certificateFileName = ssl_cert
            self.context.privateKeyFileName = ssl_key
            self.context.certificateChainFile = ssl_ca

    def callService(self, ctx, component, service, *args):
        # Make sure that arguments are serializables
        checkComplexType(args)

        if ctx.component:
            # set objects are not serializable
            acls = {}
            for name, services in ctx.component.acls.iteritems():
                acls[name] = list(services)

            context_desc = {'component': ctx.component.name,
                            'acls':      acls}
            return self.call('callService', context_desc, component, service, *args)

        if not ctx.user:
            raise NuCentralError(tr("Invalid context (not component nor user): %s"), repr(ctx))

        context_desc = {'host': ctx.user.host,
                        'port': ctx.user.port,
                       }

        if not ('ServerCookie:%s' % self.name) in ctx.getSession():
            # User hasn't any session on the remote server.
            defer = self.call('createSession', context_desc, ctx.user.login,
                                               ctx.user.groups, ctx.user.client_name,
                                               ctx.user.protocol_version, self.getRoles(ctx))
            defer.addCallback(self._cbCreateSession, ctx, context_desc, component, service, args)
            defer.addErrback(self.err_handle)
            return defer
        else:
            # Use the previous cookie.
            context_desc['cookie'] = ctx.getSession()['ServerCookie:%s' % self.name]
            d = self.call('callService', context_desc, component, service, *args)
            d.addErrback(self.err_invalidSession, ctx, context_desc, component, service, *args)
            return d

    def err_handle(self, err):
        err.raiseException()

    def err_invalidSession(self, err, ctx, context_desc, component, service, *args):
        if err.type.__name__ != 'SessionError':
            err.raiseException()

        del ctx.getSession()['ServerCookie:%s' % self.name]
        del context_desc['cookie']
        defer = self.call('createSession', context_desc, ctx.user.login,
                                           ctx.user.groups, ctx.user.client_name,
                                           ctx.user.protocol_version, self.getRoles(ctx))
        defer.addCallback(self._cbCreateSession, ctx, context_desc, component, service, args)
        return defer

    def _cbCreateSession(self, output, ctx, context_desc, component, service, args):
        """
        Callback for the service_callRemote session creation.
        """

        version, cookie = output
        ctx.getSession()['ServerCookie:%s' % self.name] = cookie
        context_desc['cookie'] = cookie
        return self.call('callService', context_desc, component, service, *args)

    # What is it? -romain
    #def call(self, method, *args):
    #    return deferToThread(self.callInThread, method, *args)

    def call(self, method, *args):

        factory = xmlrpc._QueryFactory(
            self.path, self.host, method, self.user,
            self.password, self.allowNone, args)

        if self.use_ssl:
            reactor.connectSSL(self.host, self.port or 54321,
                factory, self.context)
        else:
            reactor.connectTCP(self.host, self.port or 54321,
                factory)

        factory.deferred.addCallback(self._result)
        return factory.deferred

    def _result(self, r):

        value = r[1]
        if r[0]:
            err_type = value['type']
            klass = new.classobj(err_type, (NuCentralError,), {})
            err = None
            if 'format' in value:
                format = value['format']
                format = format
                if 'arguments' in value:
                    arguments = value['arguments']
                    if isinstance(arguments, (tuple, list)):
                        err = klass(format, *arguments)
                    else:
                        err = klass(format, **arguments)
                else:
                    err = klass(format)
            else:
                message = value['message']
                message = unicode(message)
                err = klass(message)

            if 'additional' in value:
                err.additional = value['additional']

            raise err
        else:
            return value

    def getRoles(self, ctx):
        roles = set()

        if 'multisite_admin' in ctx.user.roles:
            return list(ALL_MULTISITE_ROLES)

        session = ctx.getSession()
        if not 'multisite_transport' in session:
            session['multisite_transport'] = {}

        if not 'roles' in session['multisite_transport']:
            session['multisite_transport']['roles'] = {}

        if not self.name in session['multisite_transport']['roles']:
            for group in ctx.user.groups:
                group = getUnicode("group", group, 0, 100)
                role = [a[2] for a in self.acls.get_acl(group, '', self.name)]
                roles |= set(role)
            session['multisite_transport']['roles'][self.name] = list(roles)
            session.save()

        return session['multisite_transport']['roles'][self.name]

