"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Romain Bignon <romain AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from nucentral.core.publish_xmlrpc import RPCPublisher
from nucentral.core.version import PROTOCOL_VERSION, VERSION as nucentral_version
from nucentral.core.context import Context, ComponentContext, UserContext
from nucentral.backend import NuCentralError
from nucentral.backend.session_storage import SessionStorage
from nucentral.backend import tr, AclError
from nucentral.core.error import ACL_PERMISSION_DENIED

class SessionExpiredError(NuCentralError):
    pass

class ServerXMLRPC(RPCPublisher):
    """
    The XMLRPC server.
    """
    _multisite_session = None

    def __init__(self, core, remote_servers, allow_remote_calls):
        RPCPublisher.__init__(self, core)

        self.remote_servers = remote_servers
        self.allow_remote_calls = allow_remote_calls

    def getFunctionArguments(self, args, request):
        ctx = self._getContext(request, args[0])
        return (ctx,) + args[1:]

    def _getContext(self, request, context_desc):
        """
        From a context_desc dictionary, get a Context object.

        @param context_desc [dict]  contains some parameters usefull to
                                    create the Context object:
                                    - 'host': the client hostname.
                                    - 'port': tho client port.
                                    - 'cookie': optionnal cookie.
                                    - 'component': optionnal, if caller is a component.
                                    - 'acls': optionnal, component acls
        @return [Context]  a Context object built from arguments
        """
        if 'component' in context_desc:
            if self._multisite_session is None:
                self._multisite_session = SessionStorage()
            context = Context.fromClient(request)
            # deserialize to use set objects
            acls = context_desc.get('acls', {})
            for name, services in acls.iteritems():
                acls[name] = set(services)

            component_context = ComponentContext(context_desc['component'], acls)
            context = Context(component=component_context)
            context.setSession(self._multisite_session)
        else:
            user = UserContext(context_desc['host'], context_desc['port'])
            context = Context(user=user)
            if 'cookie' in context_desc:
                context.readSession(self.core, context_desc['cookie'])


        other_host = request.getClientIP()
        try:
            cn = dict(request.channel.transport.getPeerCertificate().get_subject().get_components())['CN']
            firewall = self.remote_servers[cn]
            firewall.host = other_host
            context.firewall = firewall.name
        except (KeyError,AttributeError), err:
            context.firewall = other_host

        # TODO What can we do when we don't find the sender firewall?
        return context

    def xmlrpc_version(self):
        return nucentral_version

    def xmlrpc_createSession(self, context, login, groups, client_name, protocol_version, roles=[]):
        """
        This xmlrpc service is used to create a Session on this server for the
        remote client.

        @param context [Context]  context object
        @param login [str]  user's login.
        @param groups [list]  list of groups.
        @param client_name [str]  client name
        @param protocol_version [str]  version of client. It must match with the
                                       server's client.
        @param roles [list]  list of all user roles
        @return  [(str, str)] a tuple with firstly the protocol version, and the
                              cookie string.
        """
        if not self.allow_remote_calls:
            return self.serviceError(AclError(ACL_PERMISSION_DENIED, tr('User calls are denied')), context, 'multisite_transport', 'createSession')

        try:
            assert isinstance(login, (str,unicode))
            assert isinstance(groups, list)
            assert isinstance(client_name, (str, unicode))
            assert isinstance(protocol_version, (str,unicode))
            assert isinstance(roles, list)

            client_data = {
                'client_name' : client_name,
                'protocol_version' : protocol_version,
            }
            cookie = self.core.session_manager.create(context, client_data)

            session = context.getSession()
            session.user.login = login
            session.user.groups = groups
            session.user.roles = set(roles)
            session.save()

            return self.formatResult((PROTOCOL_VERSION, cookie), 'createSession')
        except Exception, e:
            return self.serviceError(e, context, 'multisite_transport', 'createSession')

    def xmlrpc_callService(self, context, component, service, *args):
        """
        A remote client ask for calling a local service.

        @param context [Context]  context object
        @param component [str]  component name.
        @param service [str]  service name.
        @param args [...]  arguments passed to service.
        @return [Deferred]  a deferred object for the service result.
        """

        if not self.allow_remote_calls:
            comp = self.core.getComponent(context, component)

            if not hasattr(comp, 'MULTISITE_SERVICES'):
                return self.serviceError(AclError(ACL_PERMISSION_DENIED, tr('Access to the %s component is denied') % component), context, component, service)

            if not service in comp.MULTISITE_SERVICES:
                return self.serviceError(AclError(ACL_PERMISSION_DENIED, tr('Access to the service %s of component %s is denied') % (service, component)), context, component, service)

        if context.isUserContext():
            self.core.info(tr('User %s on host %s calls %s.%s through host %s') % (context.user.login, context.user.host, component, service, context.firewall))
        else:
            self.core.info(tr('Component %s on host %s calls %s.%s') % (context.component.name, context.firewall, component, service))

        try:
            d = self.core.callService(context, component, service, *args)
            d.addCallback(self.formatResult, service='%s.%s' % (component, service))
            d.addErrback(self.serviceError, context, component, service)
            return d
        except Exception, e:
            return self.serviceError(e, context, component, service)


