"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Romain Bignon <romain AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""
from twisted.application.internet import SSLServer, TCPServer
from twisted.web.resource import Resource
from twisted.web.static import Data
from twisted.web.client import getPage
from twisted.internet.error import BindError
from twisted.internet.defer import inlineCallbacks, returnValue, succeed

from copy import copy
import re
import os
import random

from nucentral.core.service_twisted import NuCentralService, NuCentralSite
from nucentral.core.ssl_strict import StrictOpenSSLContextFactory
from nucentral.backend.exceptions import ConfigError
from nucentral.common.error import writeError
from nucentral.common.tools import mkdirNew
from nucentral.backend import Component, NuCentralError, tr
from nucentral.backend.variables_store import VariablesStore
from nucentral.core.acl import formatSqliteItem
from nucentral.core.getter import getUnicode
from nucentral.core.acl_sqlite import SQLiteAclStorage as AclStorage

from .client import ClientXMLRPC
from .server import ServerXMLRPC

ACL_STORAGE_NAME='multisite/multisite_acls.db'

class MultiSiteTransport(Component):
    NAME = "multisite_transport"
    VERSION = "1.0"
    API_VERSION = 1

    default_conf = {'address':         '127.0.0.1',
                    'port':            54321,
                    'ssl_cert':        '',
                    'ssl_key':         '',
                    'ssl_ca':          '',
                    'use_ssl':         False,
                    'allow_remote_calls': False,
                    'client_ssl_cert': '',
                    'client_ssl_key':  '',
                    'client_ssl_ca':   '',
                    'client_use_ssl':  False,
                   }
    CONFIG_PATH = "multisite_transport.xml"

    ACLS = { 'acl': set(('listAllRoles',)) }

    ROLES = {
        'multisite_read' : set(("callRemote", "getAcl", "getRoles")),
        'multisite_write' : set(("@multisite_read", )),
        'multisite_admin' : set(("setAcl", "deleteAcl", "removeRemote"))
        }
    def init(self, core):
        self.core = core

        self.remote_servers = {}
        self.servers = {}
        self.download = Resource()
        self.config = VariablesStore()
        self.config_path = os.path.join(core.config.get('CORE', 'vardir'), self.CONFIG_PATH)
        self.acls = None
        return self.reloadConf()

    def reloadConf(self):
        self.debug("Reloading conf")
        self.acls = AclStorage(self.core, ACL_STORAGE_NAME)

        try:
            self.config.load(self.config_path)
        except ConfigError:
            pass
        self.root_path = os.path.join(self.core.config.get('CORE', 'vardir'), 'multisite')

        try:
            mkdirNew(self.root_path)
        except OSError:
            raise NuCentralError('Unable to create directory: %s', self.root_path)

        self.all_roles = None
        d = succeed('done')
        d.addCallback(lambda x:self.loadConfig())
        #d.addCallback(self.startServers)
        d.addErrback(writeError)
        return d

    #def startServers(self, result):
    #    if not result:
    #        return
    #    try:
    #        for name in self.servers.iterkeys():
    #            self.startServer(name)
    #    except Exception, e:
    #        writeError(e)

    def loadConfig(self):
        """ Load configuration """

        # Get configuration for me.
        self.conf = {}
        try:
            self.conf.update(self.config.get('self'))
        except ConfigError, e:
            self.warning(unicode(e))
            return False

        self.writeCertificates()

        # Load configuration of each remote servers connections.
        try:
            sections = self.config.get('remotes')
            for name in sections.iterkeys():
                if name in self.remote_servers:
                    self.remote_servers[name].clear()
                else:
                    self.remote_servers[name] = {}

                for server_name, items in sections[name].iteritems():
                    if not items.has_key('address') or not items.has_key('port'):
                        self.warning("Trying to load a remote server which hasn't enough config parameters.")
                        continue

                    self.remote_servers[name][server_name] = ClientXMLRPC(self.core,
                                                                    server_name,
                                                                    self.acls,
                                                                    items['address'],
                                                                    int(items['port']),
                                                                    items['client_use_ssl'],
                                                                    os.path.join(self.root_path, name, 'cert.pem'),
                                                                    os.path.join(self.root_path, name, 'key.pem'),
                                                                    os.path.join(self.root_path, name, 'cert_chain.pem')
                                                                    )
        except ConfigError, e:
            self.warning(unicode(e))
        return True

    def writeCertificates(self):
        for name in self.conf.iterkeys():
            files = {}

            if self.conf[name]['use_ssl']:
                files.update({
                    'server-cert.pem':       self.conf[name]['ssl_cert'],
                    'server-key.pem':        self.conf[name]['ssl_key'],
                    'server-cert_chain.pem': self.conf[name]['ssl_ca'] + self.conf[name]['ssl_cert']
                })

            if self.conf[name]['client_use_ssl']:
                files.update({
                    'cert.pem':       self.conf[name]['client_ssl_cert'],
                    'key.pem':        self.conf[name]['client_ssl_key'],
                    'cert_chain.pem': self.conf[name]['client_ssl_ca'] + self.conf[name]['client_ssl_cert']
                })

            if len(files.keys()):
                try:
                    mkdirNew(os.path.join(self.root_path, name))
                except OSError:
                    raise NuCentralError('Unable to create directory: %s', self.root_path)

                for filename, pem in files.iteritems():

                    f = file(os.path.join(self.root_path, name, filename), 'w')
                    f.write(pem)
                    f.close()

    def saveConfig(self):
        """return a deferred"""
        try:
            self.config.delete('self')
        except ConfigError:
            pass # it doesn't exit yet, it isn't a problem.

        for name in self.conf.iterkeys():
            for key, val in self.conf[name].iteritems():
                self.config.set('self', name, key, val)

        for name in self.remote_servers.iterkeys():
            for server_name, client in self.remote_servers[name].iteritems():
                self.config.set('remotes', name, server_name, 'address', client.host)
                self.config.set('remotes', name, server_name, 'port', client.port)
                self.config.set('remotes', name, server_name, 'use_ssl', client.use_ssl)
                self.config.set('remotes', name, server_name, 'client_use_ssl', client.use_ssl)

        self.config.save(self.config_path)

        self.writeCertificates()
        return self.core.notify.emit(self.NAME, 'configModified')

    def createService(self, name):
        if name not in self.remote_servers:
            self.remote_servers[name] = {}
        childs = {}

        childs['RPC2'] = ServerXMLRPC(self.core, self.remote_servers[name], self.conf[name]['allow_remote_calls'])
        childs['Download'] = self.download

        return NuCentralService(**childs)

    def startServer(self, name):

        self.stopServer(name)

        try:
            if self.conf[name]['use_ssl']:
                strict_ssl_context = StrictOpenSSLContextFactory(os.path.join(self.root_path, name, 'server-key.pem'),
                                                                 os.path.join(self.root_path, name, 'server-cert.pem'),
                                            certificateChainFile=os.path.join(self.root_path, name, 'server-cert_chain.pem'))


                self.servers[name] = SSLServer(int(self.conf[name]['port']),
                                        NuCentralSite( self.createService(name) ),
                                        strict_ssl_context,
                                        interface=self.conf[name]['address'])
            else:
                self.servers[name] = TCPServer(int(self.conf[name]['port']),
                                        NuCentralSite( self.createService(name) ),
                                        interface=self.conf[name]['address'])

            self.servers[name].setServiceParent(self.core.service_collection)
        except BindError, err:
            self.warning(tr('Unable to start server: %s') % str(err))
            return False
        return True

    def stopServer(self, name):
        if name in self.servers and not self.servers[name] is None:
            self.servers[name].stopService()
            self.servers[name].disownServiceParent()
            del self.servers[name]

    def service_runtimeFiles(self, ctx):
        acl_path = os.path.join(self.core.config.get('CORE', 'vardir'), ACL_STORAGE_NAME)
        return {
            'deleted' : (self.config_path, acl_path),
            'added' : ((self.config_path, 'xml'), (acl_path, 'sqlite'),),
        }

    def service_runtimeFilesModified(self, ctx):
        return self.reloadConf()

    def service_getPort(self, ctx, name):
        return self.conf[name]['port']

    def service_setSelf(self, ctx, name, address, port, allow_remote_calls, use_ssl, ssl_cert='', ssl_key='', ssl_ca=''):
        """
        Set my own server settings.

        @param address  address used to listen on (str)
        @param port  port listened (int)
        @param use_ssl use an https connection
        @param ssl_cert  certificate used for server (str)
        @param ssl_key  key of certificate (str)
        @param ssl_ca  certificate authority (str)
        """
        self.conf[name] = copy(self.default_conf)
        self.conf[name]['address'] = address
        self.conf[name]['port'] = port
        self.conf[name]['ssl_cert'] = ssl_cert
        self.conf[name]['ssl_key'] = ssl_key
        self.conf[name]['ssl_ca'] = ssl_ca
        self.conf[name]['use_ssl'] = use_ssl
        self.conf[name]['allow_remote_calls'] = allow_remote_calls

        if name not in self.remote_servers:
            self.remote_servers[name] = {}

        return self.saveConfig()

    def service_setClientSelf(self, ctx, name, use_ssl, ssl_cert='', ssl_key='', ssl_ca=''):
        """
        Set my own client settings.

        @param use_ssl use an https connection
        @param ssl_cert  certificate used for server (str)
        @param ssl_key  key of certificate (str)
        @param ssl_ca  certificate authority (str)
        """

        if name not in self.conf.keys():
            raise NuCentralError(tr("No such transport configured: %s") % name)

        self.conf[name]['client_ssl_cert'] = ssl_cert
        self.conf[name]['client_ssl_key'] = ssl_key
        self.conf[name]['client_ssl_ca'] = ssl_ca
        self.conf[name]['client_use_ssl'] = use_ssl

        return self.saveConfig()

    def service_addRemote(self, ctx, name, host, address, port):
        """
        Add a remote server.

        If certificate settings aren't specified, it uses the default
        certificate settings.

        \note if a remote server with same name already exists, new
              erases it.
        """

        if name not in self.remote_servers:
            self.remote_servers[name] = {}

        use_ssl = self.conf[name]['use_ssl']

        self.remote_servers[name][host] = ClientXMLRPC(self.core,
                                                 host,
                                                 self.acls,
                                                 address,
                                                 int(port),
                                                 use_ssl,
                                                 os.path.join(self.root_path, name, 'cert.pem'),
                                                 os.path.join(self.root_path, name, 'key.pem'),
                                                 os.path.join(self.root_path, name, 'cert_chain.pem'))

        return self.saveConfig()

    def service_removeRemote(self, ctx, name, host):
        """
        Remove a remote server.
        """
        try:
            del self.remote_servers[name][host]
        except KeyError:
            raise NuCentralError(tr("No such remote server: %s"), host)

    def service_start(self, ctx, name):
        """
        Rehash configuration, and (re)start the SSL server.
        """
        if not self.loadConfig():
            return False

        if name in self.servers and self.servers[name] is not None:
            return True

        return self.startServer(name)

    def service_stop(self, ctx, name):
        return self.stopServer(name)

    def service_getRemoteServers(self, ctx, name):
        return self.remote_servers.keys(name)

    def service_hostFile(self, ctx, data):
        filename = ''.join([random.choice('abcdefghijklmnopqrstuvwxyz1234567890') for i in xrange(16)])
        self.download.putChild(filename, Data(data, 'text/plain'))
        return 'Download/%s' % filename

    def service_removeFile(self, ctx, path):
        m = re.match('Download/(.*)', path)
        if m:
            path = m.group(1)

        del self.download.children[path]
        return True

    def service_getFile(self, ctx, fw_name, path):
        for name in self.remote_servers.iterkeys():
            for server_host, client in self.remote_servers[name].iteritems():
                if client.name == fw_name:
                    if self.conf[name]['use_ssl']:
                        return getPage('https://%s:%d/%s' % (client.host, client.port, path), contextFactory=client.context)
                    else:
                        return getPage('http://%s:%d/%s' % (client.host, client.port, path))
        raise NuCentralError(tr('Unknown firewall: %s') % fw_name)

    def service_callRemote(self, ctx, name, server, component, service, *args, **kwargs):
        if not self.remote_servers.has_key(name) or not self.remote_servers[name].has_key(server):
            raise NuCentralError(tr("%s isn't an available remote server."), server)

        client = self.remote_servers[name][server]

        return client.callService(ctx, component, service, *args, **kwargs)

    def service_setAcl(self, ctx, group='', role='', host=''):
        """
        Add the specified role into group.
        return a deferred
        """
        self.acls.set_acl(group, role, host)
        self.info(ctx, "Group %s is now in %s role on host %s" % (group, role, host))
        return self.core.notify.emit(self.NAME, 'configModified')

    @inlineCallbacks
    def service_deleteAcl(self, ctx, group='', role='', host=''):
        """
        Delete the ACLs for the specified arguments.
        Return the number of deleted ACLs.
        """
        info_str = ""
        if group:
            if role:
                info_str += tr("Group %s is removed from role %s") % (group, role)
            else:
                role = None
                info_str += tr("Group %s is removed from all roles") % group
        else:
            group = None
            if role:
                info_str += tr("Every groups are removed from role %s") % role
            else:
                role = None
                info_str += tr("Every groups are removed from every roles")

        if host:
            info_str += tr(" on host %s") % host
        else:
            host = None
            info_str += tr(" on all distant host")

        self.info(ctx, info_str)
        r = self.acls.delete_acl(group, role, host)
        yield self.core.notify.emit(self.NAME, 'configModified')
        returnValue(r)

    def service_getAcl(self, ctx, group='', role='', host=''):
        """
        Get the ACLs (group, role): all arguments are optional.
        """
        group = getUnicode("group", group, 0, 100)
        if not group:
            group = None
        role = getUnicode("role", role, 0, 100)
        if not role:
            role = None
        host = getUnicode("host", host, 0, 100)
        if not host:
            host = None
        acls = self.acls.get_acl(group, role, host)
        return ( map(formatSqliteItem, acl) for acl in acls )

    def service_getRoles(self, ctx, name, host):
        client = self.remote_servers[name][host]
        return client.getRoles(ctx)


