# -*- coding: utf-8 -*-

# $Id$

"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from nucentral.common import tr
from nucentral.common.tools import abstractmethod
from nucentral.backend import Component
from nucentral.core.context import Context
from nucentral.core.lock import LockError

class AbstractNuConfComponent(Component):
    """
    base class for all nuconf components
    """
    API_VERSION = 2

    ACLS = {
        'nuconf': set(('takeWriteRole', 'endUsingWriteRole',)),
    }

    ROLES = {
        'nuconf_read': set(),
        'nuconf_write': set(('@nuconf_read',)),
    }

    def __init__(self):
        self.CONFIG_FILES = {}
        self.core = None
        self.TEMPLATE_PATH = None
        Component.__init__(self)

    def init(self, core):
        Component.init(self, core)
        self.core = core
        self.createSubscriptions()
        self.read_config(None)

    def checkServiceCall(self, context, service_name):
        """
        if called by a user and service modify configuration then
        check user have acquired 'nuconf_write' lock. See session.acquire(...).
        Lock is acquired with nuconf.takeWriteRole().
        """
        component_ctx = Context.fromComponent(self)
        if 'nuconf' not in self.core.getComponentList(component_ctx):
            # Lock must be acquired by calling nuconf.takeWriteRole(). If nuconf
            # component is missing user will not be able to acquire the lock,
            # therefore lock checking is disabled if nuconf component is not
            # loaded
            self.warning("Could not find nuconf component: "
                "sessions are able to modify configuration concurrently")
            return

        if (service_name in self.roles['nuconf_read']) \
        or (service_name not in self.roles['nuconf_write']):
            # Only check the lock for editing services
            return

        # Check if the user has the lock
        if not self.core.lock_manager.contextHasLock(context, 'nuconf_write'):
            service = "%s.%s()" % (self.name, service_name)
            raise LockError(
                tr("You need nuconf lock to call the %s service: call nuconf.takeWriteRole()"),
                service)

    @classmethod
    def getConfigDepends(cls):
        """return names of required modules for cls
        Current module will be reconfigured when any of module in CONFIG_DEPENDS
        is (re)configured. (Re)configuration is done by calling the apply
        callback.
        """
        return cls.getTag('CONFIG_DEPENDS', AbstractNuConfComponent)

    def createSubscriptions(self):
        """
        This default implementation requires a class attribute CONFIG_DEPENDS.
        """
        if not hasattr(self.__class__, 'CONFIG_DEPENDS'):
            raise NotImplementedError(
                "This component class needs "
                "a 'CONFIG_DEPENDS' attribute (iterable of components names)."
                "See %s/%s" % (self.__module__, self.__class__.__name__)
                )

        if hasattr(self, 'rollback_config'):
            rollback = self.rollback_config
        else:
            rollback = self.apply_config

        self.core.config_manager.registerConfigComponent(self, self.read_config,
            self.apply_config, rollback)

    @abstractmethod
    def read_config(self, responsible, *args, **kwargs):
        pass

    @abstractmethod
    def apply_config(self, responsible, *unused):
        pass

    @abstractmethod
    def apply_success(self, *unused):
        pass

    def addConfFile(self, template_path, owner, mode, dest=None):
        """
        templates are located in: /usr/share/nucentral/templates
        You must specify dest path if it is different from template path.
        """
        if dest is None:
            dest = template_path
        self.CONFIG_FILES[dest] = (owner, mode, template_path)

    #FIXME: rename to generate_templates and fix children classes
    #FIXME: instead of calling self.generateTemplates, call parent class'
    def generate_configfile(self, template_variables, confFiles=None, prefix=''):
        """
        default all config files are generated
        """
        if confFiles is None:
            templates = self.CONFIG_FILES
        else:
            templates = {}
            for template in confFiles:
                templates[template] = self.CONFIG_FILES[template]
        self.generateTemplates(template_variables, templates, prefix)

