#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright(C) 2006,2007 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
conntracking.py is a part of pyctd which manages data extraction from
Netfilter Conntrack in a thread.
"""
import logging
import time
import os
import sys
import traceback
import cStringIO
import pynetfilter_conntrack as nfct
from IPy import IP

# Regit is the ultimate gruik
uidcache = {}

class DataCollector(object):
    """
    Thread to store Netfilter conntrack table up to date and compute
    byte rates. The constructor use thread.start_new_thread() to start
    a new thread, use the method stop() to stop the thread.

    Actions on the table:
     - kill(): kill one or more connections
     - change(): change a field of a connection (mark or timeout)

    Attributes:
     - update_second (float): time between refreshes
     - log (logger): object to write log messages
     - last_log (str): last writen log message (used to not write the
                       same message several times)
     - nb_same_log (int): number of times where last_log has not been
                          writen again
    """

    def __init__(self, log_appname, uid_mask, filter=None, update_second=1.0):
        self.update_second = update_second
        self.log = logging.getLogger(log_appname)
        self.last_log = None
        self.nb_same_log = 0
        self.__running = True
#        self.__running_lock = Lock()
        self.__netfilter = nfct.Conntrack()
        self.current = None
     #   self.__stopping_lock = Lock()
        self.filter = []
        if filter:
            for net in filter:
                 self.filter.append(IP(net))
        ConnectionSet.uid_mask = uid_mask

    def __del__(self):
        self.stop()

    def stop(self):
        """
        Cleanly stop the thread.
        """
        self.current = None

    def formatException(self, exc_info):
        """
        Format and return the specified exception information as a string.

        This default implementation just uses
        traceback.print_exception()
        """
        sio = cStringIO.StringIO()
        traceback.print_exception(exc_info[0],
                                  exc_info[1],
                                  exc_info[2],
                                  None,
                                  sio)
        s = sio.getvalue()
        sio.close()
        return s

    def kill(self, conn_ids):
        """
        Kill one connection or more, conn_ids is a list of connection
        identifiers (type 'int').
        """
        killed = 0
        connset = self.current
        for id in conn_ids:
            conn = connset.get(int(id))
            if conn is None:
                self.log.warning("Connection #%s not found" % (id))
                continue
            killed += 1
            self.log.info("Deleting conntrack: %s" %(conn))
            try:
                self.__netfilter.destroy_conntrack(conn.conntrack)
            except Exception, e:
                self.log.info("Unable to delete conntrack entry; %s" %(e))
        self.refresh()
        return killed

    def change(self, conn_id, field, value):
        """
        Change a field on a connection:
        - field: field name, "mark" or "timeout"
        - conn_id: connection identifier (type 'int')
        - value: new field value (int)

        Raise a ValueError if field in unknown.
        Raise a TypeError if value type is invalid.
        """
        connset = self.current
        conn = connset.get(int(conn_id))
        if not conn:
            self.log.warning("Unable to find connection #%s" %(conn_id,))
            return False

        old_val = None
        try:
            if field == "mark":
                old_val = conn.conntrack.mark
                conn.conntrack.mark = long(value)
            elif field == "timeout":
                old_val = conn.conntrack.timeout
                conn.conntrack.timeout = long(value)
                conn.conntrack.status |= nfct.IPS_FIXED_TIMEOUT
            else:
                raise ValueError("%s fields cannot be changed.")
        except TypeError:
            raise TypeError("Type of the new value unmatched (%s is a %s ; %s is a %s)" %(field, type(old_val), value, type(value)))
        conn.conntrack.update()
        self.log.info("Replacing %s from '%s' to '%s' in %s" %(field, old_val, value, conn))
        self.refresh()
        return True

    def refresh(self):
        """
        Get current conntrack table from Netfilter, and then update our
        local copy of the conntrack table (recompute byte rates and usernames).
        """
        dumped_table = self.__netfilter.dump_table()

        # suppress unwanted entries
        for entry in dumped_table[:]:
            if not hasattr(entry, 'orig_ipv4_src'):
                continue
            if IP(entry.orig_ipv4_src) not in self.filter:
                continue
            dumped_table.remove(entry)

        # end suppression
        new_connset = ConnectionSet(dumped_table, time.time())
        old_connset = self.current
        if old_connset:
            new_connset.calculate(old_connset)
        self.current = new_connset

    def __str__(self):
        current = self.current
        message = '<DataCollector %s>' % current
        return message

class ConnectionSet(dict):
    """
    ConnectionSet instances are created by DataCollector. To get the current
    ConnectionSet instance, use the read-only property DataCollector.current

    Instance attributes:
     - hit (int): date in EPOCH format where the conntrack table has
                  been dumped

    Class attributes:
     - next_id (int): identifier which will be affected to the next new
                      Connection instance
     - id_manager (dict): identifier (int) => Connection instance
     - uid_mask (int): bitmask representing bits used to define UID in the
                       netfilter connection mark
    """
    next_id = 1
    id_manager = {}
    uid_mask = None

    def __init__(self, table, hit):
        self.hit = hit

        # affect an ID to each connection #

        # 1. look for identified connections
        for tupl, id in self.id_manager.items():
            tuple_found = False
            for conn in table[:]:
                if conn.hashtuple == tupl:
                    tuple_found = True
                    self[id] = Connection(id, conn, hit)
                    table.remove(conn)
                    break
            if not tuple_found:
                # delete a dead connection
                del self.id_manager[tupl]

        # 2. new IDs for new connections ordered by descending timeout
        table_length = len(table)
        while table_length > 0:
            max = 0
            vmax = vprevious = table[0]
            for i in xrange(1, table_length):
                value_i = table[i]
                if value_i.timeout > vprevious.timeout:
                    table[i-1] = value_i
                    table[i] = vprevious
                    if value_i.timeout > vmax.timeout:
                        max = i-1
                        vmax = value_i
                vprevious = value_i
            del table[max]
            new_conn = Connection(self.next_id, vmax, hit)
            assert(self.next_id not in self)
            self[self.next_id] = new_conn
            self.id_manager[new_conn.conntrack.hashtuple] = self.next_id
            ConnectionSet.next_id = self.next_id + 1
            table_length = len(table)

    def find(self, l3proto, src_ip, l4proto, src_port):
        """
        Find a connection defined by its layer3 protocol, source ip
        layer4 protocol and source port.
        """
        if l3proto == nfct.PF_INET:
            attr_ip_src = "orig_ipv4_src"
        elif l3proto == nfct.PF_INET6:
            attr_ip_src = "orig_ipv6_src"
        else:
            raise NotImplementedError('The layer3 protocol "%s" is not yet managed.' %(l3proto))
        if l4proto not in (nfct.IPPROTO_TCP, nfct.IPPROTO_UDP):
            raise NotImplementedError('The layer4 protocol "%s" is not yet managed.' %(l4proto))
        for conn in self.itervalues():
            conn_ip_src = getattr(conn.conntrack, attr_ip_src)
            if conn.conntrack.orig_l3proto == l3proto \
                                and conn_ip_src == src_ip \
                                and conn.conntrack.orig_l4proto == l4proto \
                                and conn.conntrack.orig_port_src == src_port:
                return conn
        return None

    def calculate(self, ref):
        """
        Compute the byte rates and the username of each connections.

        Parameter ``ref`` is the previous dumped table (ConnectionSet
        instance), used as reference to compute byte rates.
        """
        for id, conn in self.iteritems():
            ref_conn = ref.get(id)
            if ref_conn is not None:
                conn.setByteRate(ref_conn)
            if self.uid_mask is not None:
                conn.findUsername(self.uid_mask)

    def __str__(self):
        s = ["Connection Set at %s:" %(self.hit)]
        for id, conn in self.iteritems():
            s.append("\t[%s] %s" %(id, conn))
        return "\n".join(s)


class Connection:
    """
    Connection instances are created by ConnectionSet. To get connections,
    just iterate on the ConnectionSet instance.

    Attributes:
     - id (int): identifier
     - conntrack: pynetfilter_conntrack.ConntrackEntry instance
     - authenticated (bool): True if the connection is authenticated
     - username (str): username who the authenticated connection belongs to,
                       None if authenticated attribute is False
     - user_id (int): user_id
     - orig_byterate (float): outbound byte rates
     - repl_byterate (float): inbound byte rates
     - hit (int): date in EPOCH format where the conntrack table has
                  been dumped (redonding the ConnectionSet attribute)
    """
    def __init__(self, id, conntrack, hit):
        self.id = id
        self.conntrack = conntrack
        self.username = ''
        self.user_id = 0
        self.authenticated = False
        self.orig_byterate = 0.0
        self.repl_byterate = 0.0
        self.hit = hit

    def setByteRate(self, ref_conn):
        ref_hit = ref_conn.hit

        current_bytes = self.conntrack.orig_counter_bytes
        ref_bytes = ref_conn.conntrack.orig_counter_bytes
        self.orig_byterate = float(current_bytes - ref_bytes) / float(self.hit - ref_hit)
        if self.orig_byterate < 0:
            curtup = self.conntrack.hashtuple
            reftup = ref_conn.conntrack.hashtuple

        current_bytes = self.conntrack.repl_counter_bytes
        ref_bytes = ref_conn.conntrack.repl_counter_bytes
        self.repl_byterate = float(current_bytes - ref_bytes) / float(self.hit - ref_hit)

    def findUsername(self, mask):
        """
        Retrieve the username from the UID included in the netfilter
        connection mark. Use ``getent`` to do that. If it fails, the
        username attribute is set to the UID value.
        """
        global uidcache
        uid = self.conntrack.mark & mask
        if uid:
            self.authenticated = True
            self.user_id = uid
            if uidcache.has_key(uid):
                self.username = uidcache[uid]
                return
            line = os.popen("getent passwd %d" %(uid)).read()
            if line:
                self.username = line.split(':')[0]
                uidcache[uid] = self.username
            else:
                self.username = uid

    def __str__(self):
        return "<Connection %(conntrack)s in:%(orig_byterate)s out:%(repl_byterate)s>" % {
            'conntrack': self.conntrack,
            'orig_byterate': self.orig_byterate,
            'repl_byterate': self.repl_byterate}

