from os.path import join as path_join
from twisted.internet.threads import deferToThread
from twisted.internet.defer import inlineCallbacks, returnValue

from nucentral.backend.variables_store import VariablesStore
from nucentral.backend import tr, Component, NuCentralError
from nucentral.core.context import Context

from nuconf.common.netcfg import deserialize
from nuconf.common.openvpn_cfg import OpenVpnConf

from .database import DataBaseManager
from .classification import Classification
from .configuration import DPIEngineConfiguration
import table

class DPIManagerComponent(Component):
    API_VERSION = 1
    NAME = 'nudpi'
    VERSION = '1.0'
    REQUIRES = set()
    ROLES = {
        'nudpi_read': frozenset(('table', 'table_filters', 'getClassification', 'setupClient', 'configuration', 'getDPIStatus',)),
        'nudpi_write': frozenset(('@nudpi_read', 'setConfiguration','setDPIStatus',)),
    }
    ACLS = {
            'network': frozenset(('getNetconfig',)),
            'openvpn': frozenset(('getOpenVpnConfig',)),
            'ha': frozenset(('getHAMode',)),
    }

    tables = {
            # TrafficTable
            'TrafficTable' : table.TrafficTable,
            'ProtoTimeTable' : table.ProtoTimeTable,

            # Users
            'UserIDTable' : table.UserIDTable,

            # DportTable
            'DportTable_most' : table.DportTable_most,
            'DportTable_less' : table.DportTable_less,
            'DportTable_most_encrypted' : table.DportTable_most_encrypted,
            'DportTable_most_unencrypted' : table.DportTable_most_unencrypted,
            #'DportTable_unknown' : table.DportTable_most,

            # UserProtoTable
            'HostProtoTable' : table.HostProtoTable,
            'HostProtoTable_http' : table.HostProtoTable_http,
            'HostProtoTable_ssl' : table.HostProtoTable_ssl,
            'HostProtoTable_encrypted' : table.HostProtoTable_encrypted,
            'HostProtoTable_unencrypted' : table.HostProtoTable_unencrypted,

            # MostUsedProtoTable
            'MostUsedProtoTable' : table.MostUsedProtoTable,
            'MostUsedProtoTable_http': table.MostUsedProtoTable_http,
            'MostUsedProtoTable_ssl' : table.MostUsedProtoTable_ssl,
            'MostUsedProtoTable_encrypted' : table.MostUsedProtoTable_encrypted,
            'MostUsedProtoTable_unencrypted' : table.MostUsedProtoTable_unencrypted,
            'MostUsedProtoTable_unknown' : table.MostUsedProtoTable_unknown,

            # MostDroppedUserTable
#            'MostDroppedUserTable' : table.MostDroppedUserTable,

            # MostDroppedSourceTable
            'MostDroppedSourceTable' : table.MostDroppedSourceTable,

            # MostUsedSourcesTable
            'MostUsedSourcesTable' : table.MostUsedSourcesTable,
            'MostUsedSourcesTable_http' : table.MostUsedSourcesTable_http,
            'MostUsedSourcesTable_ssl' : table.MostUsedSourcesTable_ssl,
            'MostUsedSourcesTable_encrypted' : table.MostUsedSourcesTable_encrypted,
            'MostUsedSourcesTable_unencrypted' : table.MostUsedSourcesTable_unencrypted,
            'MostUsedSourcesTable_unknown' : table.MostUsedSourcesTable_unknown,

            # MostUsedDestinationTable
            'MostUsedDestinationTable' : table.MostUsedDestinationTable,
            'MostUsedDestinationTable_http' : table.MostUsedDestinationTable_http,
            'MostUsedDestinationTable_ssl' : table.MostUsedDestinationTable_ssl,
            'MostUsedDestinationTable_encrypted' : table.MostUsedDestinationTable_encrypted,
            'MostUsedDestinationTable_unencrypted' : table.MostUsedDestinationTable_unencrypted,
            'MostUsedDestinationTable_unknown' : table.MostUsedDestinationTable_unknown,

            # ProtoOnDportTable
            'ProtoOnDportTable' : table.ProtoOnDportTable,

            # HostsOnDportTable
            'HostsOnDportTable' : table.HostsOnDportTable,
            'HostsOnDportTable_http' : table.HostsOnDportTable_http,
            'HostsOnDportTable_ssl' : table.HostsOnDportTable_ssl,
            'HostsOnDportTable_dropped' : table.HostsOnDportTable_ssl,

            # Protocols Classification
            'AllProtoTable' : table.AllProtoTable,

            # Last Dropped Connections
            'LastDroppedConnections' : table.LastDroppedConnections,

            #### USERS SECTION ###
            # MostUsedSourcesTable
#            'MostUsedUsersTable' : table.MostUsedUsersTable,
#            'MostUsedUsersTable_http' : table.MostUsedUsersTable_http,
#            'MostUsedUsersTable_ssl' : table.MostUsedUsersTable_ssl,
#            'MostUsedUsersTable_encrypted' : table.MostUsedUsersTable_encrypted,
#            'MostUsedUsersTable_unencrypted' : table.MostUsedUsersTable_unencrypted,

            # UserProtoTable
#            'UserProtoTable' : table.UserProtoTable,
#            'UserProtoTable_http' : table.UserProtoTable_http,
#            'UserProtoTable_ssl' : table.UserProtoTable_ssl,
#            'UserProtoTable_encrypted' : table.UserProtoTable_encrypted,
#            'UserProtoTable_unencrypted' : table.UserProtoTable_unencrypted,

            # UsersOnDportTable
#            'UsersOnDportTable' : table.UsersOnDportTable,
#            'UsersOnDportTable_http' : table.UsersOnDportTable_http,
#            'UsersOnDportTable_ssl' : table.UsersOnDportTable_ssl,

            }

    def init(self, core):
        Component.init(self, core)
        self.core = core
        self.conn = None
        self.cur = None
        self.conf_filename = path_join(core.var_dir, 'nudpi.xml')
        self.conf = VariablesStore()
        self.database = DataBaseManager()

        if self.conf_filename:
            self.conf.load(self.conf_filename)
            self.database.load(self.conf)
        else:
            print "No database conf file...."

        config = self.core.config_manager
        config.subscribe(self._apply_callback, self.NAME, (), 'apply', 'network')
        self.dpi_engine_conf = DPIEngineConfiguration(self)

    def service_table(self, context, name, args):
        """ This is the 'table' distant access function.
            @param name [string] function name;
            @param args [dictionnary]
        """

        typ = self.tables.get(name) # We get class type

        if typ is None:
            raise NuCentralError(tr('Table not found: %s'), name)

        try:
            my_object = typ(self.database) # We instance class
            defer = my_object(args) # <=> p.__call__(..)
            return defer

        except Exception, err:
            self.writeError(err, "table() error")
            raise

    def service_setupClient(self, context, attr):
        """
        Setup the client: attr = {
            'version': client_version,
        }.

        Return {'version': server_version}.
        """
        self.client_version = attr.get('version')
        component = attr.get('component')
        session = context.getSession()
        session[component] = self.client_version
        return {'version' : self.VERSION}


    def service_getClassification(self, context):
        """Get classification as a dictionary."""
        classification = Classification(self.core.var_dir)
        return classification('GetClassification', None)

    def service_configuration(self, context):
        """ Get configuration for dpi engine
        """
        return deferToThread(self.dpi_engine_conf.getConfiguration)

    @inlineCallbacks
    def service_setConfiguration(self, context, args):
        """ Set configuration for dpi engine
            @param args [dictionnary]
        """
        yield self.generateConfiguration(args)
        yield self.core.notify.emit(self.NAME, 'configModified')

    def service_getDPIStatus(self, context):
        return deferToThread(self.dpi_engine_conf.getDPIStatus)

    @inlineCallbacks
    def service_setDPIStatus(self, context, args):
        started = True
        nudpi_conf = self.dpi_engine_conf.getConfiguration()
        if not nudpi_conf['start_engine']:
            if not nudpi_conf['filters'] and not nudpi_conf['from'] and not nudpi_conf['ips_edw'] and not nudpi_conf['rule']:
                started = False
        if started and not self.dpi_engine_conf.isConfExpired() or args == 'stop':
            context = Context.fromComponent(self)
            hamode = yield self.core.callService(context, 'ha', 'getHAMode')
            yield deferToThread(self.dpi_engine_conf.setDPIStatus, args, hamode)
            yield self.core.notify.emit(self.NAME, 'configModified')

    def _apply_callback(self, responsible, *args, **kwargs):
        self.dpi_engine_conf.setConfExpired(True)
        #return self.generateConfiguration()

    @inlineCallbacks
    def generateConfiguration(self, args={}):
        context = Context.fromComponent(self)
        cfg = yield self.core.callService(context, 'network', 'getNetconfig')
        netcfg  = deserialize(cfg)

        cfg = yield self.core.callService(context, 'openvpn', 'getOpenVpnConfig')
        vpncfg = OpenVpnConf.deserialize(cfg)

        hamode = yield self.core.callService(context, 'ha', 'getHAMode')
        yield deferToThread(self.dpi_engine_conf.setConfiguration, args, netcfg, vpncfg, hamode)

    def service_table_filters(self, context, name):
        typ = self.tables.get(name) # We get class type

        if typ is None:
            raise NuCentralError(tr('Table not found: %s'), name)

        try:
            my_object = typ(self.database) # We instance class
            return my_object.getFilters()

        except Exception, err:
            self.writeError(err, "table() error")
            raise

    @classmethod
    def getConfigDepends(cls):
        return frozenset(('network', 'openvpn', 'ha'))

    def service_runtimeFiles(self, context):
        xml_file = path_join(self.core.var_dir, DPIEngineConfiguration.XML_FILENAME)
        return  {
                 'deleted' : (xml_file,),
                 'added'   : ((xml_file, 'xml',),),
             }

    def service_runtimeFilesModified(self, context):
        # Reinstantiate configuration object to reload from the file
        self.dpi_engine_conf = DPIEngineConfiguration(self)
        return self.generateConfiguration()

