from __future__ import with_statement
import os, signal
from IPy import IP
from ConfigParser import ConfigParser

from nucentral.backend.exceptions import ConfigError
from nucentral.backend.process import runCommandAsRoot, set_enabled_in_runit
from nucentral.backend.variables_store import VariablesStore
from nucentral.backend import tr, NuCentralError
from nucentral.common.error import NUDPI
from nucentral.common.logger import LoggerChild

from nuconf.common.ha_statuses import PRIMARY, SECONDARY

from nudpi.backend.errors import NUDPI_UNKNOWN_STATUS

# filter:
# - only process TCP and UDP packets
# - ignore packets from the loopback (127.0.0.0/8)
# - ignore packets from the high availability (5.0.0.0/8)
# - ignore packets from the VPN support (2.0.0.0/8)
BASE_FILTER = '(tcp or udp) and not (net 127) and not (net 5) and not (net 2)'

def dict2list(d):
    return [d[str(i)] for i in xrange(len(d.keys()))]

def list2dict(l):
    d = {}
    for no, val in enumerate(l):
        d[str(no)] = val
    return d

class StatusError(NuCentralError):
    def __init__(self, *args, **kwargs):
        NuCentralError.__init__(self, NUDPI, NUDPI_UNKNOWN_STATUS, *args, **kwargs)

class DPIEngineConfiguration(ConfigParser, LoggerChild):
    XML_FILENAME = 'nudpi_filters.xml'
    FILEPATH = "/etc/nudpi/nudpi_filters.conf"
    PID_FILE = "/etc/sv/nudpi-engine/supervise/pid"
    HEARTBEAT_SCRIPT = "/etc/ha.d/resource.d/ha_nudpi"
    HEARTBEAT_TEMPLATE_SCRIPT = None
    DEFAULT_SECTION = "config"
    DEFAULT_CONFIG = {
                    "filters"   : "",
                    "from"      : "",
                    "ips_edw"   : "",
                    "rule"      : {},
                    "start_engine" : False,
            }
    STATUS = ('start',
              'stop',
             )
    INTERNET = "0.0.0.0/0"
    RUNSV_BIN = "/usr/bin/sv"

    def __init__(self, dpi_component):
        LoggerChild.__init__(self, dpi_component)
        self.core = dpi_component.core
        self.dpi_component = dpi_component
        self.xml_conf_path = os.path.join(dpi_component.core.var_dir, self.XML_FILENAME)
        self.xml_conf = VariablesStore()
        # Becomes true when Nuconf network conf has changed
        self.confExpired = False
        try:
            self.xml_conf.load(self.xml_conf_path)
        except ConfigError, e:
            # Exception if file does not exist
            self.xml_conf.set(self.DEFAULT_SECTION, self.DEFAULT_CONFIG)
            self.debug(str(e))
            pass

        self.HEARTBEAT_TEMPLATE_SCRIPT = os.path.join(dpi_component.TEMPLATE_PATH, self.HEARTBEAT_SCRIPT)

# *** Used to know if Nuconf network configuration has changed ***
    def isConfExpired(self):
        return self.confExpired

    def setConfExpired(self, expired=True):
        self.confExpired = True
# ****************************************************************

    def getConfiguration(self):
        conf = self.xml_conf.get(self.DEFAULT_SECTION)
        conf['rule'] = dict2list(conf['rule'])
        return conf

    def setConfiguration(self, args, netcfg, vpncfg, hamode):
        old_conf = self.xml_conf.get(self.DEFAULT_SECTION)

        if 'rule' in args:
            args['rule'] = list2dict(args['rule'])
            args['eths_rule'] = list2dict(args['eths_rule'])
            # We neeed to delete "rule" because it's a dictionary, otherwise its values
            # would only be update instead overriden
            old_conf.pop(self.DEFAULT_SECTION, "rule")

        old_conf.update(args)
        self.xml_conf.delete(self.DEFAULT_SECTION)
        self.xml_conf.set(self.DEFAULT_SECTION, old_conf)
        self.xml_conf.save(self.xml_conf_path)

        is_vlan = True if 'is_vlan' in args and args['is_vlan'] else False

        self.applyConfiguration(netcfg, vpncfg, hamode, is_vlan)
        self.confExpired = False

    def applyConfiguration(self, netcfg, vpncfg, hamode, is_vlan):
        conf = ConfigParser()
        conf.add_section(self.DEFAULT_SECTION)

        self.xml_conf.get(self.DEFAULT_SECTION)
        for opt in ["filters", "from"]:
            conf.set(self.DEFAULT_SECTION, opt, self.xml_conf.get(self.DEFAULT_SECTION, opt))

        # Set the edenawll's ip
        ips_edw = []
        for net in netcfg.iterAddresses():
            ips_edw.append(str(net))
        if (vpncfg is not None) and vpncfg.enabled:
            # take the first ip of the net as the edw's ip
            vpn_ip = IP(vpncfg.client_network)[1]
            ips_edw.append(str(vpn_ip))
        conf.set(self.DEFAULT_SECTION, "ips_edw", ' '.join(ips_edw))

        # Generate pacp filter:
        rule = self.buildPcapFilter(self.xml_conf.get(self.DEFAULT_SECTION, "from"),
                                    dict2list(self.xml_conf.get(self.DEFAULT_SECTION, "rule")), netcfg, vpncfg, is_vlan)

        conf.set(self.DEFAULT_SECTION, "rule", rule)
        conf.set(self.DEFAULT_SECTION, "is_vlan", is_vlan)

        # Write nudpi_filters.conf file
        with open(self.FILEPATH, 'w') as fd:
            conf.write(fd)

        return self.setDPIStatus(None, hamode)

    def generateHeartbeatScript(self, start_nudpi):
        # The templates take a 'start_nudpi' with a value of '0' or '1':
        start_nudpi = str(int(start_nudpi))

        template_variables = { 'start_nudpi' : start_nudpi }
        self.dpi_component.generateTemplates(template_variables, {self.HEARTBEAT_SCRIPT : ('root:root', '755', self.HEARTBEAT_TEMPLATE_SCRIPT,)})

    def reloadDPIConfig(self):
        if os.path.exists(self.PID_FILE):
            with open(self.PID_FILE, 'r') as file:
                try:
                    pid = int(file.read())
                    os.kill(pid, signal.SIGUSR1)
                    return
                except ValueError, e:
                    # When nudpi-engine has an invalid configuration,
                    # Its pid file may be empty when runit is restarting it
                    # In this case sending the signal is unnecessary
                    self.warning("sending signal to nudpi failed : %s" % e)
        else:
            self.warning("nudpi-engine pid file doesn't exist")

        self.warning("nudpi-engine doesn't appear to be running while it should")
        return True

    def setDPIStatus(self, status, hamode):
        was_started = self.getDPIStatus()['started']

        if status is None:
            start_engine = self.xml_conf.get(self.DEFAULT_SECTION, "start_engine")
        else:
            if status not in self.STATUS:
                raise StatusError(tr('Unknown status'))

            start_engine = (status == "start")
            self.xml_conf.set(self.DEFAULT_SECTION, "start_engine", start_engine)
            self.xml_conf.save(self.xml_conf_path)

        # Always generate this file to correctly handle the case where
        # nudpi is first configured, then the HA is enabled
        self.generateHeartbeatScript(start_engine)

        # Start or stop the daemon as necessary
        if start_engine != was_started:
            if hamode not in [PRIMARY, SECONDARY]:
                # When no ha is configured set runit conf
                if start_engine:
                    set_enabled_in_runit(self, True, 'nudpi-engine')
                else:
                    runCommandAsRoot(self, (self.RUNSV_BIN, 'stop', 'nudpi-engine'))
                    # If we don't stop "runit nudpi-engine" process, reenable
                    # quickly nudpi-engine in runit (create the symlink) may fail:
                    # sometimes runit starts nudpi-engine and then quickly stops it.
                    # Stop runit ensures this weird behaviour.
                    runCommandAsRoot(self, (self.RUNSV_BIN, 'exit', 'nudpi-engine'))
                    set_enabled_in_runit(self, False, 'nudpi-engine')
            else:
                # When ha is configured start/stop using heartbeat's script
                if start_engine:
                    runCommandAsRoot(self, (self.HEARTBEAT_SCRIPT, 'start'))
                else:
                    runCommandAsRoot(self, (self.HEARTBEAT_SCRIPT, 'stop'))

        # or reload the conf
        elif start_engine:
            self.reloadDPIConfig()
        return True

    #def isDPIStarted(self):
    def getDPIStatus(self):
        results = {}

        nudpi_conf = self.getConfiguration()
        started = nudpi_conf['start_engine']

        file_exists = True
        if not started:
            if not nudpi_conf['filters'] and not nudpi_conf['from']:
                file_exists = False

        results['started'] = started
        results['isConfExpired'] = self.confExpired
        results['file_exists'] = file_exists

        return results

    def buildPcapFilter(self, from_if, networks, netcfg, vpncfg, is_vlan):
        if from_if == "any":
            return BASE_FILTER

        # Checks if the source interface is a vlan
        vlans_activated = False
        ip_from = []
        for ethernet in netcfg.ethernets:
            if ethernet.hard_label == from_if and ethernet.vlans:
                vlans_activated = True
            if ethernet.hard_label == from_if:
                for net in ethernet.iterNets():
                    ip_from.append(str(net.net))

        all_networks = [self.INTERNET]
        for net in netcfg.iterNetworks():
            if str(net.net) not in ip_from:
                all_networks.append(str(net.net))
        if vpncfg is not None:
            all_networks.append(vpncfg.client_network)
        filters = []
        var = ''

        if self.INTERNET in networks:
            var = ' and '
            for net in all_networks:
                if net not in networks:
                    if net:
                        exp = 'not src net ' + net + ' and not dst net ' + net
                        filters.append(exp)
        else:
            for net in networks:
                if net:
                    exp = 'src net ' + net + ' or dst net ' + net
                    filters.append(exp)
            var = ' or '

        # Filter on tcp and udp
        if vlans_activated or is_vlan:
            filter = 'vlan and '
        else:
            filter = ''

        filter += '(%s)' % BASE_FILTER

        if len(filters) != 0:
            filter += ' and (%s)' % var.join(filters)

        return filter


