#coding: utf-8
"""
Modified by Pierre-Louis Bonicoli <bonicoli AT edenwall.com>
"""
from tablebase import TableBase

ENCRYPTED_PROTO = [85, 92, 198, 199, 330, 336]

##################### TrafficTable ############################
class TrafficTable(TableBase):
    """Traffic on a range time (static view)
    """
    COLUMNS = ['oob_time_sec', 'bytes']
    FILTERS = ['ip_saddr', 'ip_daddr']
    DEFAULT_ARGS = {'sortby' : 't_start', 'sort' : 'ASC'}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']

        self.query = """SELECT t_start, (sum(ct_bytes)/(t_end-t_start))::integer
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_end < (%d)::bigint
                    """ % (start_time, end_time)

        #self._arg_string()
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')

        self.query += """ GROUP BY t_start, t_end"""
        self._arg_order_by(args)

##################### ProtoTimeTable ##########################
class ProtoTimeTable(TableBase):
    """Protocols Traffic on a range time
    """

    COLUMNS = []
    FILTERS = []
    DEFAULT_ARGS = {}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']

        # 1st query: get 5 most used protocols
        self.query = """SELECT proto_path, sum(ct_bytes) as size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_end < (%d)::bigint
                    """ % (start_time, end_time)

        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')
        self._arg_int(args, 'tcp_dport')
        self.query += """ GROUP BY proto_path
                         ORDER BY size DESC
                         LIMIT 5"""

        # self.database comes from TableBase
        data = self.database.execute(self.query)
        self.COLUMNS = []
        self.COLUMNS.append('oob_time_sec')
        for dat in data:
            if not dat[0]:
                dat = ([3, 81, 205], dat[1])
            self.COLUMNS.append(dat[0])


        # 2nd query
        self.query = """SELECT t_start, (t_end-t_start) as delta, proto_path, sum(ct_bytes) as size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_end < (%d)::bigint
                    """ % (start_time, end_time)

        #self._arg_string()
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')
        self._arg_int(args, 'tcp_dport')
        self._arg_int(args, 'proto')

        self.query += """ GROUP BY t_start, delta, proto_path
                     ORDER BY t_start ASC"""


    def _print_results(self, columns, data, args):
        results = {}
        results['args'] = args

        if data and data[0] and data[0][0]:
            data = self._getData(data)
        else:
            data = []
        columns = self.COLUMNS

        results['columns'] = columns
        results['rowcount'] = len(data)
        results['filters'] = ""
        results['table'] = data
        results['args']['limit'] =  5
        if 'start' in args:
            results['args']['start'] = args['start']
        else:
            results['args']['start'] = 0

        return results


    def _getData(self, data):
        columns = self.COLUMNS
        results = {}
        # ...
        for dat in data:
            t_start = dat[0]
            delta = dat[1]
            proto_path = dat[2]
            bytes = float(dat[3])

            if proto_path in columns:
                if t_start not in results:
                    results[t_start] = self._new_tab(len(columns))
                    results[t_start][0] = t_start
                    results[t_start][columns.index(proto_path)] = float(bytes)
                else:
                    results[t_start][columns.index(proto_path)] += float(bytes)

        return self._sortedDict(results)

    def _sortedDict(self, adict):
        keys = adict.keys()
        keys.sort()
        return map(adict.get, keys)

    def _new_tab(self, columns):
        tab = []
        for index in xrange(columns):
           tab.append(0)

        return tab



###################### DportTable ###############################
class DportTable(TableBase):
    """Most/Less used destination port, depending on 'sort'
    """
    COLUMNS = ['dport', 'size', 't_start', 't_end']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10}
    FILTERS = ['ip_saddr', 'proto']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        self.query = """SELECT tcp_dport, sum(ct_bytes) AS size, min(t_start) as t_start, max(t_end) as t_end
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_int(args, 'tcp_dport')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')

        self.query += """ GROUP BY tcp_dport"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d""" % (limit, start)


class DportTable_most(DportTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10}

    def __init__(self, database):
        DportTable.__init__(self, database)

class DportTable_less(DportTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'ASC', 'limit' : 10}

    def __init__(self, database):
        DportTable.__init__(self, database)

class DportTable_most_encrypted(DportTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10, 'proto' : ENCRYPTED_PROTO, 'include' : True}

    def __init__(self, database):
        DportTable.__init__(self, database)

class DportTable_most_unencrypted(DportTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10, 'proto' : ENCRYPTED_PROTO, 'include' : False}

    def __init__(self, database):
        DportTable.__init__(self, database)

###################### HostProtoTable ###############################
class HostProtoTable(TableBase):
    """Used protocols for an user
       State is used to know if instance of class needs a filter to be construct
    """
    COLUMNS = ['proto', 'usage', 'size']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'include' : True, 'limit' : 10}
    FILTERS = ['ip_saddr', 'tcp_dport', 'proto']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        self.query = """SELECT proto_path, sum(ct_bytes) AS size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')
        self._arg_int(args, 'tcp_dport')

        self.query += """GROUP BY proto_path"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d """ % (limit, start)


class HostProtoTable_http(HostProtoTable):
    """Used protocols whith http protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 67, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        HostProtoTable.__init__(self, database)

class HostProtoTable_ssl(HostProtoTable):
    """Used protocols with ssl protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 199, 'include' : True,'limit' : 10}

    def __init__(self, database):
        HostProtoTable.__init__(self, database)

class HostProtoTable_encrypted(HostProtoTable):
    """Used protocols with encrypted protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : True,'limit' : 10}

    def __init__(self, database):
        HostProtoTable.__init__(self, database)

class HostProtoTable_unencrypted(HostProtoTable):
    """Used protocols with unencrypted protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : False,'limit' : 10}

    def __init__(self, database):
        HostProtoTable.__init__(self, database)

###################### MostUsedProtoTable ###############################
class MostUsedProtoTable(TableBase):
    """
    """
    COLUMNS = ['proto', 'usage', 'size']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'include' : True, 'is_base' : True,'limit' : 10}
    FILTERS = ['proto', 'tcp_dport', 'ip_saddr']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        self.query = """SELECT proto_path, sum(ct_bytes) AS size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """  % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')
        self._arg_int(args, 'tcp_dport')

        self.query += """ GROUP BY proto_path"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d """ % (limit, start)

class MostUsedProtoTable_http(MostUsedProtoTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 67, 'include' : True, 'is_base' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedProtoTable.__init__(self, database)

class MostUsedProtoTable_ssl(MostUsedProtoTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 199, 'include' : True, 'is_base' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedProtoTable.__init__(self, database)

class MostUsedProtoTable_encrypted(MostUsedProtoTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : True, 'is_base' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedProtoTable.__init__(self, database)

class MostUsedProtoTable_unencrypted(MostUsedProtoTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : False, 'is_base' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedProtoTable.__init__(self, database)

class MostUsedProtoTable_unknown(MostUsedProtoTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 4, 'include' : True, 'is_base' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedProtoTable.__init__(self, database)

###################### MostUsedDestinationTable ###############################
class MostUsedDestinationTable(TableBase):
    COLUMNS = ['ip_daddr',  'size', 't_start', 't_end']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'include' : True, 'limit' : 10}
    FILTERS = ['ip_daddr', 'tcp_dport', 'proto']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        if 'count' in args and args['count']:
            self.query = """SELECT sum(ct_bytes) AS size
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                        """ % (start_time, end_time)

            self._arg_int(args, 'tcp_dport')
            self._arg_int(args, 'proto')
            self._arg_ip(args, 'ip_daddr_str')
            self._arg_ip(args, 'ip_saddr_str')

        else:
            self.query = """SELECT host(ip_daddr_str) as ip_daddr, sum(ct_bytes) AS size, min(t_start) as t_start, max(t_end) as t_end
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

            self._arg_int(args, 'tcp_dport')
            self._arg_int(args, 'proto')
            self._arg_ip(args, 'ip_daddr_str')
            self._arg_ip(args, 'ip_saddr_str')

            self.query += """ GROUP BY ip_daddr"""
            self._arg_order_by(args)
            self.query += """ LIMIT %d OFFSET %d """ %  (limit, start)

class MostUsedDestinationTable_http(MostUsedDestinationTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 67, 'include' : True,'limit' : 10}

    def __init__(self, database):
        MostUsedDestinationTable.__init__(self, database)

class MostUsedDestinationTable_ssl(MostUsedDestinationTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 199, 'include' : True,'limit' : 10}

    def __init__(self, database):
        MostUsedDestinationTable.__init__(self, database)

class MostUsedDestinationTable_encrypted(MostUsedDestinationTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedDestinationTable.__init__(self, database)

class MostUsedDestinationTable_unencrypted(MostUsedDestinationTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : False, 'limit' : 10}

    def __init__(self, database):
        MostUsedDestinationTable.__init__(self, database)

class MostUsedDestinationTable_unknown(MostUsedDestinationTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 4, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedDestinationTable.__init__(self, database)


###################### MostUsedSourcesTable ###############################
class MostUsedSourcesTable(TableBase):
    COLUMNS = ['ip_saddr',  'size', 't_start', 't_end']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'include' : True, 'limit' : 10}
    FILTERS = ['proto', 'tcp_dport', 'ip_saddr']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        if 'count' in args and args['count']:
            self.query = """SELECT sum(ct_bytes) AS size
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                        """ % (start_time, end_time)

            self._arg_int(args, 'tcp_dport')
            self._arg_int(args, 'proto')
            self._arg_ip(args, 'ip_saddr_str')
            self._arg_ip(args, 'ip_daddr_str')

        else:
            self.query = """SELECT host(ip_saddr_str) as ip_saddr, sum(ct_bytes) AS size, min(t_start) as t_start, max(t_end) as t_end
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

            self._arg_int(args, 'tcp_dport')
            self._arg_int(args, 'proto')
            self._arg_ip(args, 'ip_saddr_str')
            self._arg_ip(args, 'ip_daddr_str')

            self.query += """ GROUP BY ip_saddr"""
            self._arg_order_by(args)
            self.query += """ LIMIT %d OFFSET %d """ %  (limit, start)

class MostUsedSourcesTable_http(MostUsedSourcesTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 67, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedSourcesTable.__init__(self, database)

class MostUsedSourcesTable_ssl(MostUsedSourcesTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 199, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedSourcesTable.__init__(self, database)

class MostUsedSourcesTable_unencrypted(MostUsedSourcesTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : False, 'limit' : 10}

    def __init__(self, database):
        MostUsedSourcesTable.__init__(self, database)

class MostUsedSourcesTable_encrypted(MostUsedSourcesTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : ENCRYPTED_PROTO, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedSourcesTable.__init__(self, database)

class MostUsedSourcesTable_unknown(MostUsedSourcesTable):
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'proto' : 4, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedSourcesTable.__init__(self, database)

###################### ProtoOnDportTable ###############################
class ProtoOnDportTable(TableBase):
    COLUMNS = ['proto', 'usage', 'size']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10}
    FILTERS =  ['proto', 'ip_saddr', 'tcp_dport']

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'start' in args else 0

        self.query = """SELECT proto_path, sum(ct_bytes) AS size
                   FROM dpi
                   WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_int(args, 'tcp_dport')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')

        self.query += """ GROUP BY proto_path"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d""" % (limit, start)


###################### HostsOnDportTable ###############################
class HostsOnDportTable(TableBase):
    COLUMNS = ['ip_saddr', 'size', 'dport']
    FILTERS = ['ip_saddr', 'tcp_dport', 'proto']
    DEFAULT_ARGS = {'sortby' : 'size', 'sort' : 'DESC', 'limit' : 10}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.DEFAULT_ARGS['limit']
        start = args['start'] if 'args' in args else 0

        if 'count' in args:
            self.query = """SELECT sum(ct_bytes) AS size
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

            self._arg_int(args, 'proto')
            self._arg_int(args, 'tcp_dport')
            self._arg_ip(args, 'ip_saddr_str')
            self._arg_ip(args, 'ip_daddr_str')

        else:
            self.query = """ SELECT host(ip_saddr_str) as ip_saddr, sum(ct_bytes) AS size, tcp_dport
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

            self._arg_int(args, 'proto')
            self._arg_int(args, 'tcp_dport')
            self._arg_ip(args, 'ip_saddr_str')
            self._arg_ip(args, 'ip_daddr_str')

            self.query += """ GROUP BY ip_saddr_str, tcp_dport"""
            self._arg_order_by(args)
            self.query += """ LIMIT %d OFFSET %d""" % (limit, start)

class HostsOnDportTable_http(HostsOnDportTable):
    DEFAULT_ARGS = {'limit' : 10, 'tcp_dport' : 80}

    def __init__(self, database):
        HostsOnDportTable.__init__(self, database)

class HostsOnDportTable_ssl(HostsOnDportTable):
    DEFAULT_ARGS = {'limit' : 10, 'tcp_dport' : 443}

    def __init__(self, database):
        HostsOnDportTable.__init__(self, database)

class HostsOnDportTable_dropped(TableBase):
    DEFAULT_ARGS = {'limit' : 10}
    COLUMNS = ['ip_saddr', 'dport', 'nb_dropp', 'oob_time_sec']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.DEFAULT_ARGS['limit']

        start = args['start'] if 'start' in args else 0

        self.query = """SELECT ip_saddr_str as ip_saddr, tcp_dport, count(ip_saddr_str) as nb_dropp, min(oob_time_sec) as oob_time_sec
                    FROM dropped_users
                    WHERE oob_time_sec > (%d)::bigint AND oob_time_sec < (%d)::bigint """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_int(args, 'tcp_sport')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')

        self.query += """GROUP BY ip_saddr_str, tcp_dport"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d""" % (limit, start)




###################### USERS SECTION ###################################
###################### MostUsedUsersTable ###############################
class MostUsedUsersTable(TableBase):
    DEFAULT_ARGS = {'include' : True, 'limit' : 10}
    COLUMNS = ['username',  'size']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.FILTERS['limit']
        start = args['start'] if 'start' in args else 0

        if 'count' in args and args['count']:
            self.query = """SELECT sum(ct_bytes) AS size
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                        """ % (start_time, end_time)

            self._arg_int(args, 'proto')

        else:
            self.query = """SELECT user_, sum(ct_bytes) AS size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                    """ % (start_time, end_time)

            self._arg_int(args, 'proto')

            self.query += """ GROUP BY user_
                        ORDER BY size DESC
                        LIMIT %d OFFSET %d """ %  (limit, start)

class MostUsedUsersTable_http(MostUsedUsersTable):
    DEFAULT_ARGS = {'proto' : 67, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedUsersTable.__init__(self, database)

class MostUsedUsersTable_ssl(MostUsedUsersTable):
    DEFAULT_ARGS = {'proto' : 199, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedUsersTable.__init__(self, database)

class MostUsedUsersTable_unencrypted(MostUsedUsersTable):
    DEFAULT_ARGS = {'proto' : ENCRYPTED_PROTO, 'include' : False, 'limit' : 10}

    def __init__(self, database):
        MostUsedUsersTable.__init__(self, database)

class MostUsedUsersTable_encrypted(MostUsedUsersTable):
    DEFAULT_ARGS = {'proto' : ENCRYPTED_PROTO, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedUsersTable.__init__(self, database)

class MostUsedUsersTable_unknown(MostUsedUsersTable):
    DEFAULT_ARGS = {'proto' : 4, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        MostUsedUsersTable.__init__(self, database)

####################### UserProtoTable ###############################
class UserProtoTable(TableBase):
    """Used protocols for an user
       State is used to know if instance of class needs a filter to be construct
    """

    ################################
    ################# REDO

    DEFAULT_ARGS = {'include' : True, 'limit' : 10}
    COLUMNS = ['proto', 'usage', 'size']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        username = args['username']
        end_time = args['end_time']
        start_time = args['start_time']
        limit = self.FILTERS['limit']
        start = args['start'] if 'start' in args else 0

        self.query = """SELECT proto_path, sum(ct_bytes) AS size
                    FROM dpi
                    WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        #self._arg_string(args, 'username')

        self.query += """ AND user_ = '%s'
                    GROUP BY proto_path
                    ORDER BY size DESC
                    LIMIT %d OFFSET %d """ % (username, limit, start)


class UserProtoTable_http(UserProtoTable):
    """Used protocols whith http protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'proto' : 67, 'include' : True, 'limit' : 10}

    def __init__(self, database):
        UserProtoTable.__init__(self, database)

class UserProtoTable_ssl(UserProtoTable):
    """Used protocols with ssl protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'proto' : 199, 'include' : True,'limit' : 10}

    def __init__(self, database):
        UserProtoTable.__init__(self, database)

class UserProtoTable_encrypted(UserProtoTable):
    """Used protocols with encrypted protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'proto' : ENCRYPTED_PROTO, 'include' : True,'limit' : 10}

    def __init__(self, database):
        UserProtoTable.__init__(self, database)

class UserProtoTable_unencrypted(UserProtoTable):
    """Used protocols with unencrypted protocol
       State is used to know if instance of class needs a filter to be construct
    """
    DEFAULT_ARGS = {'proto' : ENCRYPTED_PROTO, 'include' : False,'limit' : 10}

    def __init__(self, database):
        UserProtoTable.__init__(self, database)

###################### UsersOnDportTable ###############################
class UsersOnDportTable(TableBase):


###########################
############# REDO

#    columns = ['proto', 'usage', 'size']
    COLUMNS = ['proto', 'size']
    FILTERS = []
    DEFAULT_ARGS = {'limit' : 10}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.FILTERS['limit']

        start = args['start'] if 'args' in args else 0

        if 'tcp_dport' in args:
            dport = args['tcp_dport']
        elif 'tcp_dport' in self.filters:
            dport = self.filters['tcp_dport']
        else:
            return ''

        if 'count' in args:
            self.query = """SELECT sum(ct_bytes) AS size
                        FROM dpi
                        WHERE t_start > (%d)::bigint AND t_start < (%d)::bigint
                        AND tcp_dport = %d""" % (start_time, end_time, dport)
        else:
            self.query = """SELECT user_, sum(ct_bytes) AS size
                        FROM dpi
                        WHERE tcp_dport = %d
                        AND t_start > (%d)::bigint AND t_start < (%d)::bigint
                        GROUP BY user_
                        ORDER BY size DESC
                        LIMIT %d OFFSET %d""" % (dport, start_time, end_time, limit, start)


class UsersOnDportTable_http(UsersOnDportTable):
    DEFAULT_ARGS = {'limit' : 10, 'tcp_dport' : 80}

    def __init__(self, database):
        UsersOnDportTable.__init__(self, database)

class UsersOnDportTable_ssl(UsersOnDportTable):
    DEFAULT_ARGS = {'limit' : 10, 'tcp_dport' : 443}

    def __init__(self, database):
        UsersOnDportTable.__init__(self, database)

###################################### AllProtoTable ##################################
class AllProtoTable(TableBase):
    DEFAULT_ARGS = {}
    COLUMNS = ['proto', 'usage', 'size']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']

        self.query = """SELECT proto_path, sum(ct_bytes) AS size
                    FROM dpi
                    WHERE t_start > (%d)::bigint
                    AND t_start < (%d)::bigint
                    GROUP BY proto_path""" % (start_time, end_time)


######################## UserTable_most_drop_dpi ####################################
class MostDroppedUserTable(TableBase):
    DEFAULT_ARGS = {'limit' : 10}
    COLUMNS = ['username', 'nb_dropp']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.FILTERS['limit']
        start = args['start'] if 'args' in args else 0

        self.query = """SELECT user_, count(user_)
                    FROM dropped_users
                    WHERE oob_time_sec > (%d)::bigint AND oob_time_sec < (%d)::bigint
                    GROUP BY user_
                    LIMIT %d OFFSET %d""" % (start_time, end_time, limit, start)

######################## SourceTable_most_drop_dpi ####################################
class MostDroppedSourceTable(TableBase):
    DEFAULT_ARGS = {'limit' : 10}
    COLUMNS = ['ip_saddr', 'nb_dropp', 'oob_time_sec']
    FILTERS = []

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.DEFAULT_ARGS['limit']

        start = args['start'] if 'start' in args else 0

        self.query = """SELECT ip_saddr_str as ip_saddr, count(ip_saddr_str) as nb_dropp, min(oob_time_sec) as oob_time_sec
                    FROM dropped_users
                    WHERE oob_time_sec > (%d)::bigint AND oob_time_sec < (%d)::bigint """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_int(args, 'tcp_dport')
        self._arg_int(args, 'tcp_sport')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')


        self.query += """GROUP BY ip_saddr_str"""
        self._arg_order_by(args)
        self.query += """ LIMIT %d OFFSET %d""" % (limit, start)




############################ Users ####################
class UserIDTable(TableBase):
    COLUMNS = ['username']
    FILTERS = []
    DEFAULT_ARGS = {}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        username = args['username']
        self.query = """SELECT DISTINCT(user_)
                   FROM dpi
                   WHERE user_
                   LIKE '%%%s%%' ORDER BY user_ DESC LIMIT 20 OFFSET 0""" % username

############################## Last Dropped Connections #####################"
class LastDroppedConnections(TableBase):
    COLUMNS = ['ip_saddr', 'ip_daddr', 'oob_time_sec', 'dport', 'sport', 'oob_prefix']
    FILTERS = []
    DEFAULT_ARGS = {'limit' : 40, 'sortby' : 'oob_time_sec', 'sort' : 'DESC'}

    def __init__(self, database):
        TableBase.__init__(self, database)

    def _make_query(self, args):
        start_time = args['start_time']
        end_time = args['end_time']
        limit = self.DEFAULT_ARGS['limit']

        start = args['start'] if 'args' in args else 0

        self.query = """SELECT ip_saddr_str as ip_saddr, ip_daddr_str as ip_daddr, oob_time_sec, tcp_dport, tcp_sport, oob_prefix
                        FROM dropped_users
                        WHERE oob_time_sec > (%d)::bigint AND oob_time_sec < (%d)::bigint
                        """ % (start_time, end_time)

        self._arg_int(args, 'proto')
        self._arg_int(args, 'tcp_dport')
        self._arg_int(args, 'tcp_sport')
        self._arg_ip(args, 'ip_saddr_str')
        self._arg_ip(args, 'ip_daddr_str')

        self.query += """ GROUP BY ip_saddr_str, ip_daddr_str, oob_time_sec, tcp_dport, tcp_sport, oob_prefix """
        self._arg_order_by(args)
        self.query += """LIMIT %d OFFSET %d """ % (limit, start)

