# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2011 EdenWall Technologies
Written by Romain Bignon <romain AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id: user_settings.py 12897 2009-09-30 15:12:20Z bonicoli $
"""

import random

from copy import deepcopy

from nucentral.client import NuCentralError
from pages_index import PagesIndex
from page import PagesList, Page
from links import Links
from printer import PrinterSettings
from reports import ReportsList
from nudpi.client.qt.user_settings.compatibility import Compatibility
from nudpi.client.qt.args import *
from default import default_settings, default_params
from fragment import Fragment, FragmentsList
from nudpi.client.qt.user_settings.db_association import db_association

class UserSettings:

    """
    User settings class.

    Nulog settings tree is:

    romain              :settings of a user
    |- pages_index      :pages index
    |  |- section1      :section
    |  |  |- title      :section title
    |  |  `- pages      :pages list (IDs)
    |  '
    |- pages            :list of Page objects
    |  |- Page1         :object which describes a page
    |  |- Page2         :likewise
    |  '
    |- frags            :list of Fragment objects
    |  |- Fragment1     :object which describes a fragment
    |  |- Fragment2     :likewise
    |  '
    |- links            :links for all arguments (Links)
    |- reports          :reports types
    |  |- Report1       :object which describes a report
    |  |- Report2       :likewise
    |  '
    `- printer          :printer settings (PrinterSettings)
    """

    sections = {'pagesindex':   PagesIndex,
                'pages':        PagesList,
                'frags':        FragmentsList,
                'links':        Links,
                'printer':      PrinterSettings,
                'reports':      ReportsList,
               }

    def __init__(self, main_window, client):
        self.main_window = main_window
        self.client = client
        self._data = dict()
        self.old_saved = None
        self.compatibility = Compatibility(main_window, client)

    def __contains__(self, key):
        return (key in self._data)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        if key not in self._data:
            return
        del self._data[key]

    def isOrphan(self, obj):
        if isinstance(obj, Fragment):
            for page in self['pages'].pages.values():
                for frame in page.frames:
                    if dict(frame.frags).has_key(obj.name):
                        return False

            return True
        elif isinstance(obj, Page):
            for section in self['pagesindex'].sections:
                for name, page in section.pages:
                    if name == obj.name:
                        return False
            for arg in arg_types.values():
                if arg.pagelinks.count(obj.name) > 0:
                    return False

            return True

        else:
            return False

    def __createObj(self, _class, storage, namebase):
        # First we try to create the object with the namebase, and only
        # if it exists yet, we try with randomly integers prefixes
        name = namebase
        while storage.has_key(name):
            name = '%s_%d' % (namebase, random.randint(0,999999))

        obj = _class(name)
        storage[name] = obj

        return obj

    def createPage(self, field, section=None):
        page = self.__createObj(Page, self['pages'].pages, field)
        if section:
            self['pagesindex'].addPage(page, section)

        return page

    def createFragment(self, type):
        frag = self.__createObj(Fragment, self['frags'].frags, type)
        frag.type = type
        return frag

    def removeFragment(self, frag):
        if not frag:
            return

        try:
            self['frags'].frags.pop(frag.name)
        except KeyError:
            return

        for page in self['pages'].pages.values():
            page.removeFragment(frag.name)

    def removePage(self, page):
        if not page:
            return

        try:
            self['pages'].pages.pop(page.name)
        except KeyError:
            return

        for section in self['pagesindex'].sections:
            section.removePage(page.name)

    def loadParams(self, params):
        self.old_saved = deepcopy(params)
        for key, value in params.iteritems():

            try:
                if not key in self._data:
                    self[key] = self.sections[key](key)
            except KeyError:
                print 'Warning: %s isn\'t a correct key' % key

            self[key].load(value)

        for section in self['pagesindex'].sections:
            for i, (page_name, page) in enumerate(section.pages):
                try:
                    section.pages[i] = (page_name, self['pages'][page_name])
                except KeyError:
                    del section.pages[i]
        for page in self['pages'].pages.values():
            for frame in page.frames:
                for j, (frag_name, frag) in enumerate(frame.frags):
                    try:
                        frame.frags[j] = (frag_name, self['frags'][frag_name])
                    except KeyError:
                        del frame.frags[j]

    def reset(self):
        self._data = {}
        try:
            self.client.call('users_config', 'delete', 'nudpi-qt')
            self.client.call('users_config', 'save')
        except NuCentralError, e:
            pass
        self.load()

    def loadDefaultPage(self, page_name):
        for section in self['pagesindex'].sections:
            for page in section.pages:
                if page[0] == page_name:
                    del self._data['pages'].pages[page_name]
                    self.save()
                    self.load()
                    break

    def load(self):
        try:
            self.loadParams(default_settings)
            for obj in self._data.itervalues():
                obj.setDefault(True)

            config = self.client.call('users_config', 'get', 'nudpi-qt')
            self.loadParams(config)
        except NuCentralError, err:
            # No config available, I don't care about
            return

    def save(self):
        params = {}
        for key, value in self._data.iteritems():
            params[key] = value.properties()

        # self.check_params()
        if 'pages' in params:
            for page in params['pages'].itervalues():
                if 'frames' in page:
                    new_frames = {}
                    for index, key in enumerate(sorted(page['frames'].keys())):
                        new_frames[str(index)] = page['frames'][key]
                        if 'pos' in page['frames'][key]:
                            frame = page['frames'][key]
                            frame['pos'] = str(index)

                    if len(new_frames) > 0:
                        page['frames'] = new_frames

        if self.old_saved == params:
            return
        else:
            self.old_saved = deepcopy(params)

        try:
            try:
                self.client.call('users_config', 'delete', 'nudpi-qt')
            except NuCentralError:
                # there isn't any configuration yet
                pass

            self.client.call('users_config', 'set', 'nudpi-qt', params)
            self.client.call('users_config', 'save')
        except NuCentralError, err:
            self.main_window.writeError(err)
            try:
                self.client.call('users_config', 'close')
            except NuCentralError:
                pass



# ****************** Used to generate xml file **************
    def saveClassification(self):
        try:
            try:
                pass
                self.client.call('users_config', 'deleteClassification', 'classification')
            except NuCentralError:
                # there isn't any configuration yet
                pass



            d = {}
            for k, v in db_association.items():
                d[v] = k if k != None else ""
            default_params['db_association'] = d

            self.client.call('users_config', 'setClassification', 'classification', default_params)
            self.client.call('users_config', 'saveClassification')
        except NuCentralError, err:
            self.main_window.writeError(err)
            try:
                self.client.call('users_config', 'close')
            except NuCentralError:
                pass


    def loadClassification(self):
        try:
#            self.classification = self.client.call('users_config', 'getClassification', 'classification')
            classification = self.client.call('users_config', 'getClassification', 'classification')
            return classification
        except NuCentralError:
            pass
