# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2011 EdenWall Technologies
Written by Romain Bignon <romain AT inl.fr>
           Laurent Defert <lds AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id: histogram.py 13277 2009-10-09 10:28:24Z haypo $
"""
from math import fabs, cos
from PyQt4.QtGui import QGraphicsView, QPainter, QPen, QGraphicsRectItem, QGraphicsEllipseItem, \
                        QFontMetrics, QGraphicsLineItem, \
                        QFont, QGraphicsSimpleTextItem, \
                        QColor
from PyQt4.QtCore import QTimer, SIGNAL, QVariant, Qt, QRectF, QLineF, QPointF

from nucentral.common.human import humanFilesize
from nulog.client.qt.info_area import InfoArea
from nulog.client.qt.views.nulog_base import BARCHART, PIECHART, C_PIECHART

from nudpi.client.qt.args import arg_types
from nudpi.client.qt.views.graphics_scene import GraphicsScene
from nudpi.client.qt.views.baritem import BarItem
from nudpi.client.qt.views.pieitem import PieItem
from nudpi.client.qt.views.graph import GraphFragmentView

TITLE_AREA = 0.2 # Ratio of the area where the title of the column is printed
COLUMN_SPACE_RATIO = 0.66 # Ratio of the size of one column and the size of the space between 2 columns
RESIZE_INCREMENT = 50


class GraphicsView(GraphFragmentView, QGraphicsView): #QGraphicsView

    INITIAL_SCALE = 0.05

    IN_ANIMATION = False
    ANIMATION_ON = True
    UPDATE_ALL = False

    # Custom Item Data
    NO_ROTATION = 0
    ROTATION = 1

    def __init__(self, fetcher, parent=None): #, type=BARCHART):

        QGraphicsView.__init__(self, parent)
        GraphFragmentView.__init__(self, fetcher)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)


        self.my_model = None
        self.scene = GraphicsScene(self)
        self.setScene(self.scene)
        self.centerOn(self.sceneRect().center())

        self.values = []
        self.labels = []
        self.formated_data = []
        self.current_max = 0
        self.sorted = True
        self.max_slices = 10
        self.min_slice_percent = 0

        # used for concentric pie
        self.total_depth = None
        self.radius_max = 0

        self.fetcher = fetcher

        self.fm = QFontMetrics(QFont())
        self.setResizeAnchor(QGraphicsView.AnchorViewCenter)
        self.setTransformationAnchor(QGraphicsView.AnchorViewCenter)

        self.timer = QTimer(self)
        self.connect(self.timer, SIGNAL("timeout()"), self.updateScene)
        self.setRenderHint(QPainter.Antialiasing)

        self.connect(self, SIGNAL("updateData_bis"), self._updateAll)
        self.info_area = InfoArea(self.parent())
        self.max_size_label = 0

        self.setResizeAnchor(QGraphicsView.AnchorViewCenter)

        self.legend = False
        self.show_unknown = False
        self.other = False

        self.in_animation = False

    def setLegend(self, state):
        self.legend = state

    def resizeEvent(self, event):
        # only update size of items
        if self.my_model and not self.is_closed:
            GraphicsView.UPDATE_ALL = True

            if not GraphicsView.IN_ANIMATION:
                self.emit(SIGNAL('resized(QString)'), self.uuid)
                self.requestData()
        QGraphicsView.resizeEvent(self, event)

    def setModel(self, model):
        self.my_model = model

    def redraw(self, simultaneous=False):
        if not self.my_model:
            return

        self.scene.reset()
        self.setUpdatesEnabled(False)

        # build objects
        center = self.sceneRect().center()

        if self.chart_type == BARCHART or self.chart_type == PIECHART:
            if self.current_max > 0 and self.values:
                if self.chart_type == BARCHART:
                    x = self.sceneRect().center().x()
                    y = self.sceneRect().center().y() - self.height() / 4.
                    center = QPointF(x, y)
                    self.build_objects_barchart(self.formated_data, self.order)
                    self.build_object_axes(self.formated_data)
                elif self.chart_type == PIECHART:
                    self.build_objects_piechart(self.formated_data, self.order)
                    if self.legend:
                        center = QPointF(self.sceneRect().right() / 4, self.sceneRect().center().y())
                        self.build_pie_legend()
        else:
            if self.legend:
                center = QPointF(self.sceneRect().right() / 4, self.sceneRect().center().y())
                self.centerOn(center)
            self.draw_concentric_pie()

        if not GraphicsView.IN_ANIMATION or simultaneous:
            if GraphicsView.ANIMATION_ON:
                GraphicsView.IN_ANIMATION = True
                self.timerStart()
            else:
                self.setUpdatesEnabled(True)
        else:
            self.emit(SIGNAL('replayAnimation'), self.uuid)

        self.emit(SIGNAL("showButtons"))
        self.centerOn(center)

    def _updateAll(self):
        labels = [ ]
        values = [ ]
        self.order = []

        if not self.my_model:
            return

        for row in xrange(self.my_model.rowCount()):
            label = self.my_model.data(self.my_model.index(row, 0))
            label = label.toString()
            labels.append(label)
            value = self.my_model.data(self.my_model.index(row, 1))
            value = value.toDouble()[0]
            values.append(value)
            self.order.append(label)

        if not len(values) > 0:
            self.emit(SIGNAL("showButtons"))
            return

        # build sorted list of (label,value)
        data = []
        for idx in range(0, len(values)):
            data.append([labels[idx], values[idx]])

        # create a palette of brushes for n objects
        self.create_brushes(len(values))

        self.values = values
        self.labels = labels
        self.formated_data = data
        self.current_max = max(values)
        self.current_sum = sum(values)

        # build objects
        if self.ready:
            self.redraw()

    def show(self):
        if GraphicsView.ANIMATION_ON:
            self.timerStart()
        else:
            self.setUpdatesEnabled(True)

    def timerStart(self):
        self.count = 0
        self.in_animation = True
        if self.chart_type == PIECHART or self.chart_type == C_PIECHART:
            self.timer.start(50)
        elif self.chart_type == BARCHART:
            self.timer.start(30)


    def updateData(self, result):
        if self.is_closed:
            return

        self.my_model = None
        self.scene.reset()

        if GraphFragmentView.updateData(self, result):
            if self.chart_type != C_PIECHART:
                self._updateAll()
            else:
                self.redraw()
        else:
            self.emit(SIGNAL("showButtons"))


    def draw_concentric_pie(self):
        # Only IP and TCP
        if self.total_depth < 3:
            return

        rootnode = self.my_model.rootItem
        current_depth = 1
        parent = rootnode

        if len(rootnode.childItems) == 0:
            return

        total_height = self.height()
        ellipse_radius = (total_height * 0.8) / self.total_depth

        parent.span = 360.0
        parent.z_value = 25.0
        parent.start_angle = 0.0
        parent.radius = ellipse_radius
        parent.depth = 0

        first_item = True
        items_legend = []
        self.draw_child(parent, current_depth, first_item, items_legend)


    def draw_child(self, parent, current_depth, first_item, items_legend):
        total_height = self.height()
        ellipse_radius = (total_height * 0.8) * current_depth / self.total_depth

        if self.total_depth == current_depth:
            return

        sum = 0.0
        cpt = 0

        for child in parent.childItems:
            if not self.show_unknown:
                if ("unknown" in child.data(child.ITEM_NAME)):
                    continue

            child.span = child.value * parent.span / parent.value
            child.z_value = parent.z_value - current_depth
            child.start_angle = parent.start_angle + sum
            child.radius = ellipse_radius
            child.depth = current_depth

            # keep data to make legend
            if self.legend:
                item_data = []
                item_data.append(child.data(child.ITEM_NAME))
                item_data.append(child.value)
                item_data.append(child.color)

                items_legend.append(item_data)

            rect = QRectF(0, -ellipse_radius, ellipse_radius, ellipse_radius)
            rect.moveCenter(self.scene.sceneRect().center())
            ellipse = PieItem(rect, C_PIECHART)
            ellipse.setStartAngle(child.start_angle)
            ellipse.setSpanAngle(child.span)
            ellipse.setZValue(child.z_value)
            ellipse.setColor(child.color)
            ellipse.label = child.data(child.ITEM_NAME)
            ellipse.value = child.value

            ellipse.setData(child.ITEM_NAME, QVariant(child.data(child.ITEM_NAME)))
            ellipse.setData(child.ITEM, QVariant(child))

            self.scene.addObject(ellipse)

            # *** Draw text of rootitem
            if first_item:
                first_item = False
                font = QFont()
                font.setBold(True)
                font.setPointSize(15)
                text = self.scene.addText(ellipse.label)
                text.setZValue(26)
                text.setFont(font)
                text.setPos(rect.center().x() - text.boundingRect().width() / 2,
                            rect.center().y() - text.boundingRect().height() / 2)
            # *****

            self.draw_child(child, current_depth + 1, first_item, items_legend)

            if self.radius_max < fabs(ellipse.rect().height()) / 2.0:
                self.radius_max = fabs(ellipse.rect().height()) / 2.0

            sum += child.span
            cpt += 1

        if self.legend:
            if items_legend and current_depth == 2:
                self.build_cpie_legend(items_legend, parent.value)

    def build_cpie_legend(self, items_legend, total_val):
        ellipse_radius = self.height() * 0.6

        rect = QRectF(ellipse_radius, -ellipse_radius, ellipse_radius, ellipse_radius)
        rect.moveCenter(self.scene.sceneRect().center())
        rect.moveRight(1.75 * ellipse_radius)

        x_start = rect.left()
        y_offset = rect.y() + 10

        cpt = 0

        for line in items_legend:
            label = line[0]
            value = line[1]
            color = line[2]

            percent = float(value) / float(total_val) * 100.0

            rect = self.scene.addRect(x_start + 10, y_offset, 10, 10)
            rect.setBrush(color)
            text = self.scene.addSimpleText("%s: %s (%.1f %%)" % (label, humanFilesize(value), percent,), QFont())
            text.setData(self.ROTATION, QVariant(self.NO_ROTATION))
            text.moveBy(x_start + 23, y_offset - 2)

            y_offset += 15
            cpt += 1

    def build_object_axes(self, data):

        nb_bars = len(data)
        if nb_bars < 5:
            nb_bars = 5

        width_label_offset = 10
        self.max_size_label = 0
        x_offset = -2 * self.width() / 5 + width_label_offset

        width = nb_bars * (self.width() / 15)
        xAxe = QGraphicsLineItem(x_offset, 0, width, 0)
        self.scene.addObject(xAxe)

        y_label_height = int(self.height() * 0.75)
        yAxe = QGraphicsLineItem(x_offset, 0, x_offset, -y_label_height)
        self.scene.addObject(yAxe)

        # Draw arrows for x axe
        line1 = QLineF(width, 0, width - 5, 5)
        line2 = QLineF(width, 0, width - 5, -5)

        self.scene.addLine(line1)
        self.scene.addLine(line2)

        y_max = self.current_max
        # Upper ten or so


        if self.columns[1] == 'packets' or self.columns[1] == 'nb_dropp':
            y_label = self.scene.addSimpleText("%s" % str(int(y_max)))
            y_label2 = self.scene.addSimpleText("%s" % str(int(y_max / 2)))
        else:
            y_label = self.scene.addSimpleText("%s" % humanFilesize(y_max))
            y_label2 = self.scene.addSimpleText("%s" % humanFilesize(y_max / 2))

        y_label.setData(self.ROTATION, QVariant(self.NO_ROTATION))
        y_label.moveBy(x_offset - y_label.boundingRect().width() - 5, -y_label_height)

        y_label2.setData(self.ROTATION, QVariant(self.NO_ROTATION))
        y_label2.moveBy(x_offset - y_label2.boundingRect().width() - 5, -(y_label_height) / 2)

        # Used to fit in the view
        if y_label.boundingRect().width() > self.max_size_label:
            self.max_size_label = y_label.boundingRect().width()

        # Decoration on axes
        for cpt in [0, 2]:
            rect = QGraphicsRectItem(x_offset, 0 - cpt * y_label_height / 4, width + (-x_offset), -y_label_height / 4)
            # Items in back
            rect.setZValue(-1)
            rect.setBrush(QColor(135, 135, 135, 64))
            rect.setPen(QPen(QColor(0, 0, 0, 0)))
            self.scene.addObject(rect)

    def build_objects_barchart(self, data, order):
        total_width = self.width()
        bar_width = (total_width / 15)
        total_height = self.height()
        bar_max_height = total_height * 0.75
        y_scale_factor = -1 * (bar_max_height / self.current_max)

        pen = QPen()
        width_label_offset = 10
        x_offset = -2 * total_width / 5 + width_label_offset
        cpt = 0
        sum = 0


        self.scene.addText("") # ugly hack
        for dat in data:
            (label, value) = dat

            rect = QRectF(x_offset, 0, bar_width, value * y_scale_factor)
            bar = BarItem(rect, BARCHART)

            if label in order:
                pos = order.index(label)
            else:
                pos = len(order) + 1

            bar.pos = pos
            bar.pen = pen
            bar.color = self.brushes[cpt]
            bar.value = value
            bar.label = label

            self.scene.addObject(bar)
            text = self.scene.addSimpleText("%s" % label)
            text.setData(self.ROTATION, QVariant(self.ROTATION))
            text.rotate(-45)
            text.moveBy(x_offset + (bar_width - text.boundingRect().width()) / 1.3, cos(45) * text.boundingRect().width() + 20)

            x_offset += (bar_width * 1.5)
            sum += value
            cpt += 1
            if (self.max_slices and cpt >= self.max_slices):
                break

    def build_pie_legend(self):
        ellipse_radius = self.height() * 0.6

        rect = QRectF(ellipse_radius, 0, ellipse_radius, ellipse_radius)
        rect.moveCenter(self.scene.sceneRect().center())
        rect.moveRight(1.75 * ellipse_radius)
        max_color = len(self.brushes)

        x_start = rect.left()
        y_offset = rect.y() + 10

        cpt = 0

        for item in self.scene.items():
            if isinstance(item, QGraphicsEllipseItem):
                label = item.label
                value = item.value

                rect = self.scene.addRect(x_start + 10, y_offset, 10, 10)
                rect.setBrush(item.getColor())

                unit = arg_types[self.columns[1]].unit
                legend = str(value) + " " + unit if self.columns[1] != "size" else humanFilesize(value)

                text = self.scene.addSimpleText("%s: %s (%.1f %%)" % (label, legend, (value * 100 / self.current_sum),), QFont())
                text.setData(self.ROTATION, QVariant(self.NO_ROTATION))
                text.moveBy(x_start + 23, y_offset - 2)

                y_offset += 15
                cpt += 1

    def build_objects_piechart(self, data, order):
        pen = QPen()
        total_height = self.height()
        ellipse_radius = (total_height * 0.6)
        max_color = len(self.brushes)

        cpt = 0
        sum = 0

        self.scene.addText("") # ugly hack
        for dat in data:
            (label, value) = dat
            start_angle = sum
            span = (value * 360 / self.current_sum)

            rect = QRectF(0, -ellipse_radius, ellipse_radius, ellipse_radius)
            rect.moveCenter(self.scene.sceneRect().center())
            ellipse = PieItem(rect, PIECHART)

            ellipse.setStartAngle(start_angle)
            ellipse.setSpanAngle(span)
            ellipse.setPen(pen)
            ellipse.setColor(self.brushes[cpt % max_color])
            ellipse.pos = order.index(label)
            ellipse.value = value
            ellipse.label = label
            ellipse.setZValue(0)

            sum += span
            cpt += 1
            self.scene.addObject(ellipse)

    def sizeHint(self):
        size = QGraphicsView.sizeHint(self)
        MIN_WIDTH = 100
        MIN_HEIGHT = 100
        if size.width() < MIN_WIDTH:
            size.setWidth(MIN_WIDTH)
        if size.height() < MIN_HEIGHT:
            size.setHeight(MIN_HEIGHT)
        return size

    def makeColors(self):
        colors = []
        for row in range(0, 10):
            color = QColor(102, row * 25, row * 25)
            colors.append(color)
        return colors

    def create_brushes(self, nb_el):
        self.brushes = []
        self.colors = GraphFragmentView.colours

        cpt = 0
        for color in self.colors:
            self.brushes.append(QColor(color))
            cpt += 1
            if nb_el == cpt:
                break

    def updateScene(self):
        if not self.is_closed and self.in_animation:
            self.setUpdatesEnabled(True)

            self.count += 1
            # finished ? stop animation
            if self.count >= 20:
                self.timer.stop()
                self._animation_done()
            else:
                #update all objects
                for item in self.scene.items():
                    if isinstance(item, QGraphicsRectItem) or isinstance(item, QGraphicsLineItem):
                        item.resetTransform()
                        item.scale(1, self.count * self.INITIAL_SCALE)
                    elif isinstance(item, PieItem):
                        current_step = (self.count + 1) * (360 / 20)
                        if (current_step >= item.startAngle):
                            item.setVisible(True)
                            if current_step > item.startAngle + item.spanAngle:
                                QGraphicsEllipseItem.setSpanAngle(item, item.spanAngle * 16)
                            else:
                                QGraphicsEllipseItem.setSpanAngle(item, 16 * (current_step - item.startAngle))
                        else:
                            item.setVisible(False)
                    elif isinstance(item, QGraphicsSimpleTextItem):
                        item.resetTransform()
                        if int(item.data(self.ROTATION).toString()) != self.NO_ROTATION:
                            item.rotate(-45)
        else:
            self.timer.stop()

    def _animation_done(self):
        GraphicsView.IN_ANIMATION = False
        self.in_animation = False

        self.emit(SIGNAL("showButtons"))

        if GraphicsView.UPDATE_ALL:
            self.emit(SIGNAL("animation_done(QString)"), self.uuid)

    def wheelEvent(self, event):
        pos = self.mapToScene(event.pos())
        scale = 1.2

        if event.delta() < 0:
            x = y = 1 / 1.2
        else:
            x = y = 1.2
            scale *= -1
        self.scene.setTextSize(self.scene.getTextSize() + scale)
        self.scale(x, y)
