#!/usr/bin/env python
from __future__ import with_statement
from time import time, sleep
from datetime import datetime
from math import floor
from os import getpid
import syslog
from optparse import OptionParser
from ConfigParser import ConfigParser

import psycopg2

from nucentral.backend.variables_store import VariablesStore
from nudpi.backend.database import DataBaseManager


class Compression(ConfigParser):

    RANGE_NAME = ('quarter', 'hour', 'day', 'week', 'month', 'year',)

    TIME_RANGE = {'quarter' : 60*15, # 900 seconds
                  'hour' : 60*60,
                  'day' : 60*60*24,
                  'week' : 60*60*24*7,
                  'month' : 60*60*24*31,
                  'year' : 60*60*24*365
                 }

    COMPRESSED_RANGE = {'quarter' : 60*15 / 30, # 900 seconds
                        'hour' : 60*60 / 60,
                        'day' : 60*60*24 / 60,
                        'week' : 60*60*24*7 / 60,
                        'month' : 60*60*24*31 / 30,
                        'year' : 60*60*24*365 / 30
                       }

    QUERY = """
            INSERT INTO dpi(ct_pckts, ct_bytes, ct_flows, ip_saddr_str, ip_daddr_str, tcp_dport, proto_path, t_start, t_end)
            (SELECT sum(ct_pckts) as pckts, sum(ct_bytes) as bytes, sum(ct_flows) as flow, ip_saddr_str,
                    ip_daddr_str, tcp_dport, proto_path,
                    %d, %d
             FROM dpi
             WHERE t_start >= %d AND t_start < %d AND t_end - t_start < %d
             GROUP BY ip_saddr_str, ip_daddr_str, tcp_dport, proto_path);
            DELETE FROM dpi WHERE t_start >= %d AND t_start < %d AND t_end - t_start < %d;
             """
    COMPRESSION_CONF_PATH = '/var/lib/nucentral/nudpi-compression.conf'
    DB_CONF_PATH = '/var/lib/nucentral/nudpi.xml'
    DEFAULT_SECTION = "config"


    def __init__(self):
        ConfigParser.__init__(self)
        self.dbconf = VariablesStore()
        self.database = DataBaseManager()
        self.dbconf.load(self.DB_CONF_PATH)
        self.database.load(self.dbconf)

    def launch(self, range, start_time, end_time):
        for compress_start in xrange(start_time, end_time, self.COMPRESSED_RANGE[range]):
            compress_end = compress_start + self.COMPRESSED_RANGE[range]
            # Get ids to compress
            query = self.QUERY % (compress_start, compress_end,
                                compress_start, compress_end, self.COMPRESSED_RANGE[range],
                                compress_start, compress_end, self.COMPRESSED_RANGE[range])
            self.database.execute(query, commit=True)

    def readConfiguration(self):
        loaded_conf = {}
        self.read(self.COMPRESSION_CONF_PATH)
        for section in self.sections():
            for option in self.options(section):
                loaded_conf[option] = self.get(section, option)
        return loaded_conf

    def saveConfiguration(self, current_time):
        self.read(self.COMPRESSION_CONF_PATH)
        if self.DEFAULT_SECTION not in self.sections():
            self.add_section(self.DEFAULT_SECTION)

        self.set(self.DEFAULT_SECTION, 'previous_time', current_time)

        with open(self.COMPRESSION_CONF_PATH, 'w') as fd:
            self.write(fd)

def getCompressionEndTime(time, range):
    return int(floor(time / Compression.COMPRESSED_RANGE[range]) * Compression.COMPRESSED_RANGE[range]) - Compression.TIME_RANGE[range]

if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-v", "--verbose", action="store_true", help="write compression informations to syslog")
    (options, args) = parser.parse_args()

    syslog.openlog('nudpi-compression[%i]' % getpid())
    syslog.syslog('Starting nudpi-compression')
    compression = Compression()
    loaded_conf = compression.readConfiguration()

    if 'previous_time' in loaded_conf:
        previous_time = int(loaded_conf['previous_time'])
        syslog.syslog('Last start at %s' % datetime.fromtimestamp(previous_time).isoformat())
    else:
        syslog.syslog('First start')
        previous_time = time()

    while True:
        if not compression.database.connect():
            syslog.syslog(syslog.LOG_ERR, "Can't connect to database ...")
            sleep(30)
            continue

        current_time = int(time())

        try:
            for range in Compression.RANGE_NAME:
                # Previous compression end_time for this range
                previous_endtime = getCompressionEndTime(previous_time, range)
                current_endtime = getCompressionEndTime(current_time, range)

                for start_time in xrange(previous_endtime, current_endtime, Compression.COMPRESSED_RANGE[range]):
                    if options.verbose:
                        syslog.syslog(syslog.LOG_INFO, "compressing range %s : %010i -> %010i" % (range, start_time, start_time + Compression.COMPRESSED_RANGE[range]))
                    compression.launch(range, start_time, start_time + Compression.COMPRESSED_RANGE[range])

            previous_time = current_time
            compression.saveConfiguration(previous_time)
        except (psycopg2.Error, psycopg2.Warning), e:
            syslog.syslog(syslog.LOG_ERR, "Compression counldn't be done : %s" % e)
            try:
                compression.database.disconnect()
            except (psycopg2.Error, psycopg2.Warning), e:
                pass
            sleep(30)
            continue

        try:
            compression.database.disconnect()
        except (psycopg2.Error, psycopg2.Warning), e:
            pass
        next = current_time + 30
        now = time()
        if now < next:
            sleep(next - now)

