
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

import sys
from PyQt4.QtGui import QPalette, QWidget, QApplication, QPainter, QBrush, QPen
from PyQt4.QtCore import Qt, QPoint, QPointF, QDateTime, QTime, QString, QSize
import PyQt4.Qwt5 as Qwt
from PyQt4.Qwt5.anynumpy import arange


class TextScaleDraw(Qwt.QwtScaleDraw):
    """
    Class to implement the placement of text on axis scales
    """
    def __init__(self, labelStrings, *args):
        """
        Initialize text scale draw with label strings and any other arguments
        """
        Qwt.QwtScaleDraw.__init__(self, *args)
        self.labelStrings = labelStrings

    def label(self, value):
        """
        Apply the label at location 'value' . Since this class is to be used
        for BarPlots or LinePlots, every item in 'value' should be an integer.
        """
        label = QString(self.labelStrings[int(value)])
        return Qwt.QwtText(label)

class CartesianAxis(Qwt.QwtPlotItem):
    """
    Supports a coordinate system similar to
    http://en.wikipedia.org/wiki/Image:Cartesian-coordinate-system.svg
    """

    def __init__(self, masterAxis, slaveAxis):
        """
        Valid input values for masterAxis and slaveAxis are QwtPlot.yLeft,
        QwtPlot.yRight, QwtPlot.xBottom, and QwtPlot.xTop. When masterAxis is
        an x-axis, slaveAxis must be an y-axis; and vice versa.
        """
        Qwt.QwtPlotItem.__init__(self)
        self.__axis = masterAxis
        if masterAxis in (Qwt.QwtPlot.yLeft, Qwt.QwtPlot.yRight):
            self.setAxis(slaveAxis, masterAxis)
        else:
            self.setAxis(masterAxis, slaveAxis)
        self.scaleDraw = Qwt.QwtScaleDraw()
        self.scaleDraw.setAlignment((Qwt.QwtScaleDraw.LeftScale,
                                     Qwt.QwtScaleDraw.RightScale,
                                     Qwt.QwtScaleDraw.BottomScale,
                                     Qwt.QwtScaleDraw.TopScale)[masterAxis])

    def draw(self, painter, xMap, yMap, rect):
        """
        Draw an axis on the plot canvas
        """
        if self.__axis in (Qwt.QwtPlot.yLeft, Qwt.QwtPlot.yRight):
            self.scaleDraw.move(round(xMap.xTransform(0.0)), yMap.p2())
            self.scaleDraw.setLength(yMap.p1()-yMap.p2())
        elif self.__axis in (Qwt.QwtPlot.xBottom, Qwt.QwtPlot.xTop):
            self.scaleDraw.move(xMap.p1(), round(yMap.xTransform(0.0)))
            self.scaleDraw.setLength(xMap.p2()-xMap.p1())
        self.scaleDraw.setScaleDiv(self.plot().axisScaleDiv(self.__axis))
        self.scaleDraw.draw(painter, self.plot().palette())

class CartesianPlot(Qwt.QwtPlot):
    """
    Creates a coordinate system similar system
    http://en.wikipedia.org/wiki/Image:Cartesian-coordinate-system.svg
    """

    def __init__(self, filename):
#         self.times = coords[0]
#         self.values = coords[1]

        Qwt.QwtPlot.__init__(self)
        self.setTitle(filename)
        self.setAxisTitle(Qwt.QwtPlot.yLeft, "Load");
        # create a plot with a white canvas
        self.setCanvasBackground(Qt.white)
        # set plot layout
        self.plotLayout().setMargin(0)
        self.plotLayout().setCanvasMargin(0)
        self.plotLayout().setAlignCanvasToScales(True)
        # attach a grid
        grid = Qwt.QwtPlotGrid()
        grid.attach(self)
        grid.setPen(QPen(Qt.black, 0, Qt.DotLine))
        # attach a x-axis
        xaxis = CartesianAxis(Qwt.QwtPlot.xBottom, Qwt.QwtPlot.yLeft)
        xaxis.attach(self)
        self.enableAxis(Qwt.QwtPlot.xBottom, True)
        # attach a y-axis
        yaxis = CartesianAxis(Qwt.QwtPlot.yLeft, Qwt.QwtPlot.xBottom)
        yaxis.attach(self)
        self.enableAxis(Qwt.QwtPlot.yLeft, True)
        self.setAxisLabelRotation(Qwt.QwtPlot.xBottom, -50.0);
#         self.draw("HOUR")

#         self.__initZooming()

#     def __initZooming(self):
#         """Initialize zooming
#         """
#
#         self.zoomer = Qwt.QwtPlotZoomer(Qwt.QwtPlot.xBottom,
#                                         Qwt.QwtPlot.yLeft,
#                                         Qwt.QwtPicker.DragSelection,
#                                         Qwt.QwtPicker.AlwaysOff,
#                                         self.canvas())


    def draw(self, periodString):
        """
        Draw or redraw graph depending on coordinates.
        """
        self.clear()

        start = 0
        end = len(self.times)
        xLabelHours = True

        if periodString == "HOUR":
            xLabel = "Time : 1 hour"
        elif periodString == "DAY":
            xLabel = "Time : 24 hours"
        elif periodString == "WEEK":
            xLabel = "Time : 7 days"
            xLabelHours = False
        elif periodString == "MONTH":
            xLabel = "Time : 31 days"
            xLabelHours = False

        delta = 1.0
        self.setAxisTitle(Qwt.QwtPlot.xBottom, xLabel);
        self.setAxisLabelRotation(Qwt.QwtPlot.xBottom, -50)

        x = arange(start, end, delta)
        y = self.values
        labelStrings = []
        for i in range(len(self.times)):
            if xLabelHours:
                labelStrings.append(QDateTime.fromTime_t(self.times[i]).time().toString("hh:mm"))
            else:
                labelStrings.append(QDateTime.fromTime_t(self.times[i]).toString("dd/MM"))

        self.setAxisScaleDraw(Qwt.QwtPlot.xBottom, TextScaleDraw(labelStrings))
        self.setAxisScale(Qwt.QwtPlot.xBottom, 0, len(labelStrings) )

        # attach a curve
        curve = Qwt.QwtPlotCurve('Plot')
        curve.attach(self)
        curve.setPen(QPen(Qt.green, 2))
        curve.setData(x, y)
        self.replot()

if __name__ == '__main__':
    app = QApplication(sys.argv)
    plot = CartesianPlot()
    plot.show()
    sys.exit(app.exec_())

