"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Sebastien Tricaud <stricaud AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id: nupki.py 5475 2009-02-19 15:18:43Z toady $
"""

import sqlite3
from sqlite3 import OperationalError

from nucentral.common.error import NUPKI
from nucentral.common.transport import parseDatetime
from nucentral.common.timezone import datetime_set_utc
from nucentral.backend import tr, ComponentError
from nucentral.backend.error import exceptionAsUnicode
from nucentral.backend.logger import Logger

from .error import NUPKI_DB

from nupki.common.pki import PKI, Certificate

CERT_FIELDS = 'type, id, cn, email, orgunit, location, state, country, status, new_status, organization, start_date, end_date'

class DatabaseError(ComponentError):
    def __init__(self, *args, **kw):
        ComponentError.__init__(self, NUPKI, NUPKI_DB, *args, **kw)

class Database(Logger):
    def __init__(self, logger, filename):
        Logger.__init__(self, "db", parent=logger)
        self.warning("Open database: %s" % filename)
        try:
            self.conn = sqlite3.connect(filename)
        except OperationalError, err:
            raise DatabaseError(
                tr("Unable to open the database (%s): %s"),
                filename, exceptionAsUnicode(err))
        self.cursor = self.conn.cursor()

    def commit(self):
        self.debug("Commit")
        self.conn.commit()

    def rollback(self):
        self.debug("Rollback")
        self.conn.rollback()

    def runRequest(self, request, *args):
        message = "SQL query: %s;" % request
        if args:
            message += " # %s" % (args,)
        self.debug(message)
        return self.cursor.execute(request, args)

    def close(self):
        self.cursor.close()

    def _where(self, where, arguments):
        if not where:
            return ''
        conditions = []
        for key, value in where.iteritems():
            conditions.append('%s=?' % key)
            arguments.append(value)
        return " WHERE %s" % ' AND '.join(conditions)

    def _select(self, table, fields, where=None, order_by=None):
        arguments = []
        sql = 'SELECT %s FROM %s' % (fields, table)
        sql += self._where(where, arguments)
        if order_by:
            sql += ' ORDER BY %s' % order_by
        self.runRequest(sql, *arguments)

    def _selectCert(self, pki_name, fields, where=None, order_by=None):
        where['pkiid'] = self.getPKI_ID(pki_name)
        self._select('certificate', fields, where, order_by=order_by)

    def _update(self, table, values, where, commit=True):
        if not values:
            # Nothing to do
            return
        arguments = []
        keys = []
        for key, value in values.iteritems():
            keys.append(key)
            arguments.append(value)
        sql = "UPDATE %s SET " % table
        sql += ', '.join(('%s=?' % key) for key in keys)
        sql += self._where(where, arguments)
        self.runRequest(sql, *arguments)
        rowcount = self.cursor.rowcount
        if commit:
            self.commit()
        return rowcount

    def _insert(self, table, values):
        sql = 'INSERT INTO %s' % table
        fields = []
        arguments = []
        for key, value in values.iteritems():
            fields.append(key)
            arguments.append(value)
        sql += ' (%s)' % ', '.join(fields)
        sql += ' VALUES (%s)' % ', '.join('?' for index in xrange(len(fields)))
        self.runRequest(sql, *arguments)

    def getFromRequest(self, request, *args):
        self.runRequest(request, args)
        return self.cursor.fetchone()

    def insertNewPKI(self, pkiname, orgunit, org, location, state, country):
        self._insert('pki',
            {'name': pkiname,
             'lastserial': 0,
             'orgunit': orgunit,
             'organization': org,
             'location': location,
             'state': state,
             'country': country})
        self.commit()

    def getPKI_ID(self, pkiname, error=True):
        """
        Retrieves the PKI ID
        """
        self._select('pki', 'id', {'name': pkiname})
        row = self.cursor.fetchone()
        if not row:
            if error:
                raise DatabaseError(tr("There is no PKI called %s"), repr(pkiname))
            else:
                return None
        return row[0]

    def hasPKI(self, name):
        return (self.getPKI_ID(name, error=False) is not None)

    def deletePKI(self, pkiname):
        pkiid = self.getPKI_ID(pkiname, error=False)
        if pkiid is None:
            return False
        self.runRequest("DELETE FROM certificate WHERE pkiid = ?",
                        pkiid)
        self.runRequest("DELETE FROM pki WHERE name = ?",
                        pkiname)
        self.commit()
        return True

    def listPKI(self):
        """
        List all entries in PKI
        and run the callback 'cb' for each entry
        """
        self._select('pki',
            'id, name, orgunit, organization, location, state, country, lastserial')
        return [PKI(*columns) for columns in self.cursor]

    def getPKI(self, pkiname):
        self._select('pki',
            'id, name, orgunit, organization, location, state, country, lastserial',
            {'name': pkiname})
        columns = self.cursor.fetchone()
        if columns is None:
            raise DatabaseError(tr("There is no such PKI: %s"), pkiname)
        return PKI(*columns)

    def _getCertificate(self, pki_name, columns):
        cert = Certificate(*columns)
        cert.pki_name = pki_name
        if cert.start_date is not None:
            cert.start_date = parseDatetime(cert.start_date)
            cert.start_date = datetime_set_utc(cert.start_date)
        if cert.end_date is not None:
            cert.end_date = parseDatetime(cert.end_date)
            cert.end_date = datetime_set_utc(cert.end_date)
        return cert

    def getCert(self, pkiname, cname):
        self._selectCert(pkiname, CERT_FIELDS, {'cn': cname})
        columns = self.cursor.fetchone()
        return self._getCertificate(pkiname, columns)

    def getCertList(self, pki, **where):
        where['pkiid'] = pki.id
        self._select('certificate', CERT_FIELDS, where, order_by='cn')
        return [self._getCertificate(pki.name, columns) for columns in self.cursor]

    def getCertStatus(self, pkiname, cname):
        self._selectCert(pkiname, "status", {'cn': cname})
        return self.cursor.fetchone()[0]

    def cnameExists(self, pkiname, cname):
        self._selectCert(pkiname, "cn", {'cn': cname})
        row = self.cursor.fetchone()
        return (row is not None)

    def createCertificate(self, pkiname, type, cname, serial, email, organization, orgunit, location, state, country, end_date):
        pkiid = self.getPKI_ID(pkiname)
        self._insert('certificate',
            {'pkiid': pkiid,
             'type': type,
             'cn': cname,
             'email': email,
             'organization': organization,
             'orgunit': orgunit,
             'location': location,
             'state': state,
             'country': country,
             'serial': serial,
             'end_date': end_date,
             'status': "to be created"})
        self.commit()

    def updatePKISettings(self, pkiname, orgunit, organization, location, state, country):
        self.runRequest('UPDATE pki SET orgunit = ?, organization = ?, location = ?, state = ?, country = ? '
                        'WHERE name = ?',
                        orgunit, organization, location, state, country,
                        pkiname)
        self.commit()

    def setNewStatus(self, pkiname, cname, new_status):
        if new_status not in ("to be renewed", "to be revoked"):
            return False
        pkiid = self.getPKI_ID(pkiname)
        rowcount = self._update('certificate',
            {'new_status': new_status},
            {'pkiid': pkiid, 'cn': cname})
        return (rowcount == 1)

    def setStatus(self, pkiname, cname, status, **values):
        pkiid = self.getPKI_ID(pkiname)
        values['status'] = status
        values['new_status'] = None
        self._update('certificate',
            values,
            {'pkiid': pkiid, 'cn': cname})

    def undo(self, pkiname):
        modified = False
        pkiid = self.getPKI_ID(pkiname)
        # Delete certificate creation requests
        self.runRequest(
            "DELETE FROM certificate WHERE pkiid=? AND status=?",
            pkiid, "to be created")
        modified |= (self.cursor.rowcount != 0)
        # Undo renew/revocation requests
        self.runRequest(
            "UPDATE certificate SET new_status=NULL "
            "WHERE pkiid=? AND new_status IS NOT NULL",
            pkiid)
        modified |= (self.cursor.rowcount != 0)
        self.commit()
        return modified

    #--- Deprecrated methods kept for backward compatibility ----------------
    #--- (compatibility with old clients) ------------------------------------

    # Deprecated, replaced by createCertificate()
    def appendCertificate(self, pkiname, type, cname):
        pkiid = self.getPKI_ID(pkiname)
        serial = 0 # XXX
        self._insert('certificate',
            {'pkiid': pkiid,
             'type': type,
             'cn': cname,
             'serial': serial,
             'status': "to be created"})
        # DON'T commit: it will be done in the next service call,
        # updateCertificateSettings(), which use rollback() on error to keep
        # database consistency

    # Deprecated, replaced by createCertificate()
    def updateCertificateSettings(self, pkiname, cname,
    serial, email, organization, orgunit, location, state, country, end_date):
        pkiid = self.getPKI_ID(pkiname)
        try:
            self._update('certificate',
                {'email': email,
                 'organization': organization,
                 'orgunit': orgunit,
                 'location': location,
                 'state': state,
                 'country': country,
                 'end_date': end_date},
                {'pkiid': pkiid, 'cn': cname},
                commit=False)
        except:
            self.rollback()
            raise
        self.commit()

