"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Sebastien Tricaud <stricaud AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id$
"""

from __future__ import with_statement
from os.path import join as path_join, exists
import shutil
import re

from nucentral.common import tr
from nucentral.common.transport import parseDatetime
from nucentral.common.timezone import datetime_utcnow, datetime_set_utc

from nupki.common.pki import DEFAULT_CERT_EXPIRE, MIN_DATE, MAX_DATE
from nupki.common.regex import (PKI_NAME_REGEX, CN_NAME_REGEX,
    ORGANIZATION_UNIT_REGEX, STATE_REGEX,
    EMAIL_REGEX, COUNTRY_REGEX)
from .error import NuPKIError
from .x509 import Certificate
from .database import Database
from .ssl_wrapper import SSLWrapper, escapeFilename

class PKIManager(object):
    """
    Manages the PKI itself. There are two kinds of PKI models:
    * simple: where best default choices are choosen for the user. Fonction names like fooSimpleBar(self, ...)
    * advanced: where every detail can be tuned. Fonction names like fooBar(self, ...)
    """
    def __init__(self, component, conf, pkiname="default"):
        self.component = component
        self.logger = component

        core = component.nucentral_core
        self.notify = core.notify
        if not conf:
            vardir = core.var_dir
            conf = {'nupkidir': path_join(vardir, 'nupki2'),
                     'pkidir': path_join(vardir, 'nupki2', 'pki'),
                     'templates': path_join(vardir, 'nupki2', 'templates'),
                     'dbfile': path_join(vardir, 'nupki2','certs.db'),
                     'default_orgunit': 'Research and Development',
                     'default_org': 'Edenwall',
                     'default_location': 'Paris',
                     'default_state': 'Ile-de-France',
                     'default_country': 'FR',
                    }

        self.conf = conf
        self.ssl = Certificate()
        self.workdir = conf['pkidir']
        self.expire_time = 60*60*24*365*5     # five years
        self.pkiname = pkiname

        # Update the schema if needed
        self.db = Database(self.logger, conf['dbfile'])
        self.updateDatabaseSchema()

    def checkPKIName(self, name):
        if not re.match(PKI_NAME_REGEX, name):
            raise NuPKIError(tr("Invalid PKI name: %s"), repr(name))

    def checkCertName(self, cname):
        if not re.match(CN_NAME_REGEX, cname):
            raise NuPKIError(
                tr('Invalid certificate name: %s'),
                repr(cname))
        # Checks if the PKI has no weird char that can cause
        # shell misuse
        if cname.find("..") >= 0 or cname.find("%2e%2e") >= 0:
            raise NuPKIError(
                tr('Invalid certificate name: directory escaping.'))

    def reload(self):
        self.db.close()
        self.db = Database(self.component, self.conf['dbfile'])
        self.updateDatabaseSchema()

    def pkiExists(self, name):
        return self.db.hasPKI(name)

    def checkParameter(self, name, value, regex):
        if not re.match(regex, value):
            raise NuPKIError(
                tr('Invalid %s parameter value: %s'),
                name, str(value))

    def checkDatetime(self, name, timestamp):
        if not(MIN_DATE <= timestamp <= MAX_DATE):
            raise NuPKIError(
                tr('Invalid %s parameter: %s'),
                name, str(timestamp))

    def newPKI(self, pkiname, orgunit, org, location, state, country, expire):
        self.checkPKIName(pkiname)
        self.checkParameter("organization unit", orgunit, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("organization", org, STATE_REGEX)
        self.checkParameter("location", location, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("state", state, STATE_REGEX)
        self.checkParameter("country", country, COUNTRY_REGEX)
        if self.pkiExists(pkiname):
            raise NuPKIError(
                tr('There is already a PKI called "%s".'),
                pkiname)
        sslwrap = SSLWrapper(self, pkiname)
        sslwrap.createCA(pkiname, orgunit, org, location, state, country, expire)
        self.db.insertNewPKI(pkiname, orgunit, org, location, state, country)

    def getPKI(self, pkiname):
        self.checkPKIName(pkiname)
        return self.db.getPKI(pkiname)

    def getCertificate(self, pkiname, cname):
        return self.db.getCert(pkiname, cname)

    def deletePKI(self, pkiname):
        self.checkPKIName(pkiname)
        pkidir = path_join(self.workdir, pkiname)
        try:
            # We remove the directory containing everything about this PKI
            shutil.rmtree(pkidir)
            removed = True
        except OSError:
            removed = False
        removed |= self.db.deletePKI(pkiname)
        return removed

    def _doCreateCertificate(self, sslwrap, cname, type):
        sslwrap.createCertRequest(type, cname)
        sslwrap.signCertRequest(type, cname)
        start_date = self._getX509Field(sslwrap.pkiname, cname, "startdate")
        end_date = self._getX509Field(sslwrap.pkiname, cname, "enddate")
        self.db.setStatus(sslwrap.pkiname, cname, 'valid',
            start_date=start_date,
            end_date=end_date)

    def doCreateAllCertificates(self, pki):
        """
        Forces the creation of every certificate in the pki 'pkiname'
        """
        sslwrap = SSLWrapper(self, pki.name)
        for cert in self._listAllCerts(pki, status="to be created"):
            self._doCreateCertificate(sslwrap, cert.cname, cert.type)

    def revokeCertificate(self, sslwrap, cname):
        sslwrap.revokeCert(cname)
        self.db.setStatus(sslwrap.pkiname, cname, 'revoked')
        self.notify.emit('nupki', 'updateCRL', pki_name=sslwrap.pkiname)

    def revokeAllCertificates(self, pki):
        """
        Forces the revokation of every certificate in the pki 'pkiname'
        """
        sslwrap = SSLWrapper(self, pki.name)
        for cert in self._listAllCerts(pki,  new_status="to be revoked"):
            self.revokeCertificate(sslwrap, cert.cname)

    def renewCertificate(self, sslwrap, cert):
        sslwrap.renewCert(cert)
        self.db.setStatus(sslwrap.pkiname, cert.cname, 'valid')

    def renewAllCertificates(self, pki):
        """
        Forces the renewal of every certificate in the pki 'pkiname'
        """
        sslwrap = SSLWrapper(self, pki.name)
        for cert in self._listAllCerts(pki, new_status="to be renewed"):
            self.renewCertificate(sslwrap, cert)

    def applyAllPKIActions(self, only_pki=None):
        """
        Handles the cron task and the "force pending action" button action
        """
        if only_pki:
            self.checkPKIName(only_pki)

        pkis = self.listPKI()
        for pki in pkis:
            name = pki.name
            if only_pki and (name != only_pki):
                continue
            self.doCreateAllCertificates(pki)
            self.revokeAllCertificates(pki)
            self.renewAllCertificates(pki)

    def signRequest(self, pkiname, cname, req, type):
        self.checkPKIName(pkiname)
        # TODO: retrieve the cname from the certificate and make sure it's not already been signed
        sslwrap = SSLWrapper(self, pkiname)
        sslwrap.writeSigningRequest(cname, req)
        sslwrap.signCertRequest(type, cname)

    def listCerts(self, pkiname, cert_type):
        pki = self.getPKI(pkiname)
        return self.db.getCertList(pki, type=cert_type)

    def _listAllCerts(self, pki, **where):
        return self.db.getCertList(pki, **where)

    def listPKI(self):
        return self.db.listPKI()

    def close(self):
        self.db.close()

    def getCertStatus(self, pkiname, cname):
        self.checkPKIName(pkiname)
        return self.db.getCertStatus(pkiname, cname)

    def getPKIPath(self, pkiname):
        self.checkPKIName(pkiname)
        return path_join(self.workdir, pkiname)

    def getCACertPath(self, pkiname):
        return path_join(self.getPKIPath(pkiname), "ca.crt")

    def getCrlPath(self, pkiname):
        return path_join(self.getPKIPath(pkiname), "crl.pem")

    def getCAPrivPath(self, pkiname):
        return path_join(self.getPKIPath(pkiname), "ca.key")

    def getCACert(self, pkiname):
        with open(self.getCACertPath(pkiname)) as fp:
            return fp.read()

    def getCACertX509Text(self, pkiname):
        self.checkPKIName(pkiname)
        sslwrap = SSLWrapper(self, pkiname)
        return sslwrap.getCACertText()

    def getCRLText(self, pki_name):
        sslwrap = SSLWrapper(self, pki_name)
        filename = self.getCrlPath(pki_name)
        return sslwrap.getCRLText(filename )

    def _getCertPath(self, pkiname, cname, suffix):
        self.checkCertName(cname)
        filename = escapeFilename(cname) + suffix
        return path_join(self.getPKIPath(pkiname), filename)

    def getCertRequestPath(self, pkiname, cname):
        return self._getCertPath(pkiname, cname, ".csr")

    def getCertPath(self, pkiname, cname):
        return self._getCertPath(pkiname, cname, ".cert")

    def getPrivateCertPath(self, pkiname, cname):
        return self._getCertPath(pkiname, cname, ".pkey")

    def getCertContent(self, pkiname, cname):
        with open(self.getCertPath(pkiname, cname)) as fp:
            return fp.read()

    def getCRLBuffer(self, pkiname):
        with open(path_join(self.getPKIPath(pkiname), "crl.pem")) as fp:
            return fp.read()

    def getPrivateCert(self, pkiname, cname):
        with open(self.getPrivateCertPath(pkiname, cname)) as fp:
            return fp.read()

    def checkCertFileExists(self, pkiname, cname):
        filename = self.getCertPath(pkiname, cname)
        if not exists(filename):
            raise NuPKIError(
                tr('The "%s" PKI does not contain the "%s" certificate, or this certificate has not been created.'),
                pkiname, cname)

    def setNewStatus(self, pkiname, cname, new_status):
        if new_status not in ('to be revoked', 'to be renewed'):
            return False
        self.checkPKIName(pkiname)
        self.checkCertFileExists(pkiname, cname)
        return self.db.setNewStatus(pkiname, cname, new_status)

    def doRevoke(self, pkiname, cname, serial):
        """
        from "to be revoked" to "revoked"
        """
        self.checkPKIName(pkiname)
        self.checkCertName(cname)
        sslwrap = SSLWrapper(self, pkiname)
        sslwrap.revokeCert(cname)
        self.db.setStatus(pkiname, cname, 'revoked')
        self.notify.emit('nupki', 'updateCRL', pki_name=pkiname)

    def regenCRL(self, pkiname):
        sslwrap = SSLWrapper(self, pkiname)
        result = sslwrap.regenCRL()
        self.notify.emit('nupki', 'updateCRL', pki_name=pkiname)
        return result

    def updatePKISettings(self, pkiname, orgunit, organization, location, state, country):
        self.checkPKIName(pkiname)
        self.checkParameter("organization unit", orgunit, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("organization", organization, STATE_REGEX)
        self.checkParameter("location", location, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("state", state, STATE_REGEX)
        self.checkParameter("country", country, COUNTRY_REGEX)
        self.db.updatePKISettings(pkiname, orgunit, organization, location, state, country)
        # FIXME: update organization field of all PKI certificates

    def createCertificate(self, attr):
        pkiname = attr['pki']
        self.checkPKIName(pkiname)

        cname = attr['cname']
        type = attr.get('type', 'client')
        email = attr.get('email', u'')
        orgunit = attr.get('organization_unit', u'')
        location = attr.get('location', u'')
        state = attr.get('state', u'')
        country = attr.get('country', u'')
        serial = attr.get('serial', 0)
        force = attr.get('force', False)
        try:
            organization = attr['organization']
        except KeyError:
            pki = self.db.getPKI(pkiname)
            organization = pki.organization
        try:
            end_date = parseDatetime(attr['end_date'])
            if end_date.tzinfo is None:
                end_date = datetime_set_utc(end_date)
        except KeyError:
            end_date = datetime_utcnow() + DEFAULT_CERT_EXPIRE

        self.checkCertName(cname)
        self.checkParameter("email", email, EMAIL_REGEX)
        self.checkParameter("organization", organization, STATE_REGEX)
        self.checkParameter("organization unit", orgunit, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("location", location, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("state", state, STATE_REGEX)
        self.checkParameter("country", country, COUNTRY_REGEX)
        self.checkDatetime("end date", end_date)

        if self.db.cnameExists(pkiname, cname):
            raise NuPKIError(
                tr('The "%s" PKI already contains a certificate called "%s".'),
                pkiname, cname)
        self.db.createCertificate(pkiname, type, cname, serial, email, organization, orgunit, location, state, country, end_date)
        if force:
            sslwrap = SSLWrapper(self, pkiname)
            self._doCreateCertificate(sslwrap, cname, type)
        return None

    def _getX509Field(self, pki_name, cname, field):
        sslwrap = SSLWrapper(self, pki_name)
        value = sslwrap.getx509param(cname, field)

        if field == "startdate" or field == "enddate":
            value = sslwrap.parseDatetime(value)
        return value

    def getX509Field(self, pki_name, cname, field):
        self.checkPKIName(pki_name)
        self.checkCertName(cname)
        self.checkCertFileExists(pki_name, cname)
        return self._getX509Field(pki_name, cname, field)

    def undo(self, pki_name):
        self.checkPKIName(pki_name)
        return self.db.undo(pki_name)

    def updateDatabaseSchema(self):
        """
        Responsible of any database update
        Implicit transaction between every commit
        """
        self.db.runRequest("SELECT version FROM _schema WHERE name='nupki'")
        columns = self.db.cursor.fetchone()
        dbversion = columns[0]

        if dbversion == "2.0":
            try:
                self._upgrade2_0()
            except Exception, err:
                self._handle_upgrade_error(err)
            else:
                self.db.commit()
                dbversion = "2.1"

        if dbversion == "2.1":
            try:
                self._upgrade2_1()
            except Exception, err:
                self._handle_upgrade_error(err)
            else:
                self.db.commit()
                dbversion = "2.2"

        if dbversion == "2.2":
            try:
                self._upgrade2_2()
            except Exception, err:
                self._handle_upgrade_error(err)
            else:
                self.db.commit()
                dbversion = "2.3"

    def _handle_upgrade_error(self, err):
        self.db.rollback()
        self.component.writeError(err, "Error while upgrading")

    def _upgrade2_0(self):
        self.logger.error("Upgrade database schema 2.0 to version 2.1")
        self.db.runRequest('UPDATE _schema SET version="2.1" WHERE name = "nupki"')
        self.db.runRequest('ALTER TABLE certificate ADD COLUMN expire TEXT')

    def _upgrade2_1(self):
        self.logger.error("Upgrade database schema 2.1 to version 2.2")
        self.db.runRequest('UPDATE _schema SET version="2.2" WHERE name = "nupki"')
        self.db.runRequest('ALTER TABLE certificate ADD COLUMN new_status TEXT')
        for status in ("to be renewed", "to be revoked"):
            self.db.runRequest('UPDATE certificate SET new_status=?, status="valid" WHERE status=?',
                status, status)

    def _upgrade2_2(self):
        self.logger.error("Upgrade database schema 2.2 to version 2.3")
        self.db.runRequest('UPDATE _schema SET version="2.3" WHERE name = "nupki"')
        self.db.runRequest('ALTER TABLE certificate ADD COLUMN organization TEXT')
        self.db.runRequest('ALTER TABLE certificate ADD COLUMN start_date TIMESTAMP')
        self.db.runRequest('ALTER TABLE certificate ADD COLUMN end_date TIMESTAMP')
        for pki in self.listPKI():
            self.db._update('certificate',
                {'organization': pki.organization},
                {'pkiid': pki.id},
                commit=False)
            for cert in self._listAllCerts(pki):
                if cert.status != "to be created":
                    start_date = self._getX509Field(pki.name, cert.cname, "startdate")
                    end_date = self._getX509Field(pki.name, cert.cname, "enddate")
                else:
                    start_date = None
                    end_date = datetime_utcnow() + DEFAULT_CERT_EXPIRE
                self.db._update('certificate',
                    {'start_date': start_date,
                     'end_date': end_date},
                    {'pkiid': pki.id, 'cn': cert.cname})

    def getCertText(self, pki_name, cname):
        self.checkPKIName(pki_name)
        self.checkCertName(cname)
        sslwrap = SSLWrapper(self, pki_name)
        filename = self.getCertPath(pki_name, cname)
        return sslwrap.getCertText(filename)

    #--- Deprecrated methods kept for backward compatibility ----------------
    #--- (compatibility with old clients) ------------------------------------

    # Deprecated, replaced by createCertificate()
    def newCert(self, pkiname, cname, type="client", expire=None):
        self.checkPKIName(pkiname)
        self.checkCertName(cname)
        if self.db.cnameExists(pkiname, cname):
            raise NuPKIError(
                tr('The "%s" PKI already contains a certificate called "%s".'),
                pkiname, cname)
        self.db.appendCertificate(pkiname, type, cname)

    # Deprecated, replaced by createCertificate()
    def updateCertificateSettings(self, pkiname, cname, serial, email, organization, orgunit, location, state, country):
        end_date = datetime_utcnow() + DEFAULT_CERT_EXPIRE

        self.checkPKIName(pkiname)
        self.checkCertName(cname)
        self.checkParameter("email", email, EMAIL_REGEX)
        self.checkParameter("organization", organization, STATE_REGEX)
        self.checkParameter("organization unit", orgunit, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("location", location, ORGANIZATION_UNIT_REGEX)
        self.checkParameter("state", state, STATE_REGEX)
        self.checkParameter("country", country, COUNTRY_REGEX)
        self.checkDatetime("end date", end_date)

        self.db.updateCertificateSettings(pkiname, cname,
            serial, email, organization, orgunit, location, state, country,
            end_date)

    # Deprecated: replaced by the applyAllPKIActions()
    def forceCertCreation(self, pkiname, cname, cert_type):
        self.checkPKIName(pkiname)
        self.checkCertName(cname)
        sslwrap = SSLWrapper(self, pkiname)
        self._doCreateCertificate(sslwrap, cname, cert_type)

