#!/usr/bin/python
"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Sebastien Tricaud <stricaud AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id: OpenSSLWrapper.py 14445 2009-11-17 16:23:01Z toady $
"""

from __future__ import with_statement

import subprocess
from os import makedirs, umask
import codecs
from os.path import join as path_join
from logging import getLogger
import re
from jinja import Environment
from tempfile import NamedTemporaryFile
from datetime import datetime
from locale import setlocale, LC_ALL

from nucentral.common import tr
from nucentral.common.process import createProcess
from nucentral.common.timezone import datetime_utcnow
from nucentral.backend.process import runCommand

from nupki.common.pki import DEFAULT_CERT_EXPIRE, CERT_RENEW_ENDTIME
from nupki.common.regex import PKI_NAME_REGEX
from .error import NuPKIError
from .environment import NupkiEnv

def escapeConf(value):
    if not isinstance(value, unicode):
        value = unicode(value)
    return value.replace("$", "\$").replace("#", r"\#")

def escapeSubject(value):
    return value.replace("/", r"\/")

def escapeFilename(filename):
    return filename.replace(" ", "%20").replace("/", "%2f")

def checkCountry(country):
    if len(country) < 2:
        raise NuPKIError(tr('Country code is too short: %s'), repr(country))

class SSLWrapper:
    def __init__(self, pki, pkiname=None, prefix_path=None):
        if (pkiname is not None) \
        and not re.match(PKI_NAME_REGEX, pkiname):
            raise NuPKIError(tr("Invalid PKI name: %s"), repr(pkiname))

        self.conf = pki.conf
        self.env = NupkiEnv()
        self.envconst = self.env.constant
        self.logger = getLogger()

        self.opensslCmd = {
            'genrsa': ['openssl', 'genrsa'],
            'rsa': ['openssl', 'rsa', '-pubout'],
            'x509': ['openssl', 'x509'],
            'buildCA': ['openssl', 'req', '-days', '3650', '-nodes', '-new', '-x509', '-batch'],
        }

        self.nupki = pki

        self.pkiname = pkiname

        if self.pkiname:
            self.pkiworkdir = path_join(self.conf['pkidir'], self.pkiname)
            cacert = self._getCACertPath(self.pkiname)
            cakey = self._getPrivateCACertPath(self.pkiname)
            self.opensslcfg = path_join(self.pkiworkdir, "openssl.cfg")

            self.opensslCmd['ca'] = ['openssl', 'ca', '-batch', '-notext', '-cert', '-config', self.opensslcfg, cacert, '-keyfile', cakey]
            self.opensslCmd['req'] = ['openssl', 'req', '-config', self.opensslcfg, '-batch']


    def __extractVal(self, value):
        # return "May 27 15:46:59 2014 GMT"
        return value.strip().partition("=")[2]

    def __getCertPath(self, pkiname, cname):
        return path_join(self.pkiworkdir, escapeFilename(cname) + self.envconst['cert-suffix'])

    def _getCACertPath(self, pkiname):
        return path_join(self.pkiworkdir, self.envconst['ca-cert'])

    def _getPrivateCACertPath(self, pkiname):
        return path_join(self.pkiworkdir, self.envconst['ca-privatekey'])

    def checkCommandStatus(self, status):
        if status != 0:
            raise NuPKIError(tr("Command failed (exit code %s)"), status)

    def runCommand(self, command, check_status=True):
        process, status = runCommand(self.logger, command)
        self.checkCommandStatus(status)

    def __run(self, command, parameters):
        ssl_cmd = []
        for item in self.opensslCmd[command]:
            ssl_cmd.append(item)
        for item in parameters:
            ssl_cmd.append(item)

        handler = createProcess(self.logger, ssl_cmd, stdout=subprocess.PIPE)
        output = handler.communicate()[0]
        self.checkCommandStatus(handler.returncode)
        return self.__extractVal(output)

    def __createWrite(self, file, buffer):
        with codecs.open(file, "w", encoding="utf8") as fp:
            fp.write(buffer)
            fp.flush()

    def __setDefaultVariablesPath(self, variables):
        workdir = self.pkiworkdir
        variables['dir']           = workdir
        variables['CA_dir']        = workdir
        variables['certs']         = path_join(workdir, "cert")
        variables['crl_dir']       = path_join(workdir, "crl")
        variables['database']      = path_join(workdir, "index.txt")
        variables['new_certs_dir'] = path_join(workdir, "newcerts")
        variables['serial']        = path_join(workdir, "serial")
        variables['crlnumber']     = path_join(workdir, "crlnumber")
        variables['crl']           = path_join(workdir, "crl.pem")
        # DO NOT CHANGE THIS. ca.crt is used in the ca command from openssl
        variables['certificate']   = path_join(workdir, "ca.crt")
        # DO NOT CHANGE THIS. ca.key is used in the ca command from openssl
        variables['private_key']   = path_join(workdir, "ca.key")

    def createKeys(self, privkeyname, pubkeyname=None, size=1024):
        """
        Create private key and public key of size 'size',
        It uses the RSA algorithm.
        """
        self.__run('genrsa', ['-out', privkeyname, unicode(size)])
        if pubkeyname:
            self.__run('rsa', ['-in', privkeyname, '-out', pubkeyname])

    def formatSubject(self, country, state, org, location, pkiname, orgunit):
        checkCountry(country)
        return '/C=%s/ST=%s/O=%s/localityName=%s/commonName=%s/organizationalUnitName=%s/' % (
            escapeSubject(country), escapeSubject(state), escapeSubject(org),
            escapeSubject(location), escapeSubject(pkiname),
            escapeSubject(orgunit))

    def renderTemplate(self, template, filename,  variables):
        # Escape variables
        variables = dict(
            (key, escapeConf(value))
            for key, value in variables.iteritems())

        with codecs.open(filename, "w", encoding="utf8") as fp:
            buffer = template.render(variables)
            fp.write(buffer)

    def createCA(self, pkiname, orgunit, org, location, state, country, expire):
        """
        First function called when creating a new PKI
        """
        checkCountry(country)
        if not expire:
            expire = "3650"

        variables = self.getOpenVPNConfig()
        jinjaenv = Environment()

        # Read template
        opensslcfg_tmpl = path_join(self.conf['templates'], "openssl_vpn.cfg.tmpl")
        with codecs.open(opensslcfg_tmpl, "r", encoding="utf8") as fp:
            template = fp.read()
        tmpl = jinjaenv.from_string(template)

        variables['req_default_bits'] = "1024"

        variables['req_dn_0.organizationName_default'] = org
        variables['req_dn_organizationalUnitName_default'] = orgunit
        variables['req_dn_localityName_default'] = location
        variables['req_dn_stateOrProvinceName_default'] = state
        variables['req_dn_countryName_default'] = country

        # Change variables
        self.__setDefaultVariablesPath(variables)

        makedirs(variables['dir'])
        makedirs(variables['certs'])
        makedirs(variables['crl_dir'])
        makedirs(variables['new_certs_dir'])

        # Write pki openssl config with variables
        # /var/lib/nucentral/nupki2/mypki/openssl.cfg
        openssl_config = path_join(self.pkiworkdir, "openssl.cfg")
        self.renderTemplate(tmpl, openssl_config, variables)

        self.__createWrite(variables['serial'], "01\n")
        self.__createWrite(variables['crlnumber'], "01\n")
        self.__createWrite(variables['database'], "")

        subject = self.formatSubject(country, state, org, location, pkiname, orgunit)

        runme = (
            "openssl", "req",
            "-days", str(expire),
            "-nodes", "-new", "-x509", "-batch",
            "-keyout", variables['private_key'],
            "-out", variables['certificate'],
            "-config", openssl_config,
            "-subj", subject)
        self.runCommand(runme)

        # Once the CA is created, we create a CRL
        runme = ("openssl", "ca", "-gencrl",
            "-keyfile", variables['private_key'],
            "-cert", variables['certificate'],
            "-out", variables['crl'],
            "-config", openssl_config)
        self.runCommand(runme)

    def writeOpensslConfig(self, cname):
        variables = self.getOpenVPNConfig()
        variables['CA_dir'] = self.pkiworkdir
        variables['CA_commonName_value'] = cname
        variables['req_commonName_default'] = cname
        if self.pkiname:
            cert = self.nupki.getCertificate(self.pkiname, cname)
            variables['req_dn_0_organizationName_default'] = cert.organization
            variables['req_dn_organizationalUnitName_default'] = cert.organization_unit
            variables['req_dn_localityName_default'] = cert.location
            variables['req_dn_stateOrProvinceName_default'] = cert.state
            checkCountry(cert.country)
            variables['req_dn_countryName_default'] = cert.country
            variables['req_dn_emailAddress_default'] = cert.email

        variables['req_default_bits'] = 1024

        opensslcfg_tmpl = path_join(self.conf['templates'], "openssl_vpn.cfg.tmpl")

        jinjaenv = Environment()

        # Read template
        with codecs.open(opensslcfg_tmpl, "r", encoding="utf8") as fp:
            template = fp.read()
        tmpl = jinjaenv.from_string(template)
        openssl_config = path_join(self.pkiworkdir, "openssl.cfg")
        self.renderTemplate(tmpl, openssl_config, variables)

        return openssl_config

    def writeSigningRequest(self, cname, req):
        cname_path = path_join(self.pkiworkdir, escapeFilename(cname))
        cname_csr = cname_path +  ".csr"
        with open(cname_csr, "wb") as fp:
            fp.write(req)

    def checkCertType(self, cert_type):
        if cert_type in ('server', 'client'):
            return
        raise NuPKIError(tr('Unknown certificate type: %s'), cert_type)

    def createCertRequest(self, certtype, cname):
        """
        create a new Certificate request with the CA authoring in the PKI context
        certype: user or server
        cname: common name (username or dns record) to apply for the PKI. Must be unique.
        """
        # FIXME: use a valid expire value
        expire = "3650"

        self.checkCertType(certtype)

        openssl_config = self.writeOpensslConfig(cname)

        cname_path = path_join(self.pkiworkdir, escapeFilename(cname))
        cname_csr = cname_path +  ".csr"
        cname_key = cname_path + self.envconst['privatekey-suffix']

        self.__run('genrsa', ['-out', cname_key])

        req_cmd = ("openssl", "req",
            "-days", expire,
            "-nodes", "-new", "-batch",
            "-keyout", cname_key,
            "-out", cname_csr,
            "-config", openssl_config)
        if certtype == "server":
            req_cmd += ("-extensions", "server")

        self.runCommand(req_cmd)

    def signCertRequest(self, certtype, cname, end_date=None):
        """
        create a new Certificate request with the CA authoring in the PKI context
        certype: user or server
        cname: common name (username or dns record) to apply for the PKI. Must be unique.
        """
        self.checkCertType(certtype)

        openssl_config = self.writeOpensslConfig(cname)

        cname_path = path_join(self.pkiworkdir, escapeFilename(cname))
        cname_csr = cname_path + ".csr"
        cname_cert = cname_path + self.envconst['cert-suffix']

        if not end_date:
            if self.pkiname:
                cert = self.nupki.getCertificate(self.pkiname, cname)
                end_date = cert.end_date
            else:
                end_date = datetime_utcnow() + DEFAULT_CERT_EXPIRE

            # UTC timestamp, year is written in two digits so 1950<=year<=2049
            # 'Z' suffix is the time zone (UTC)
        end_date = end_date.strftime("%y%m%d%H%M%SZ")

        sign_cmd = ("openssl", "ca",
            "-batch",
            "-enddate", end_date,
            "-out", cname_cert,
            "-in", cname_csr,
            "-config", openssl_config)
        if certtype == "server":
            sign_cmd += ("-extensions", "server")
        self.runCommand(sign_cmd)


    def getCertBufferFromReqBuffer(self, certtype, cname, reqbuf):
        reqoutfile = path_join(self.pkiworkdir, "req.tmp")
        with open(reqoutfile, "w") as reqoutfp:
            reqoutfp.write(reqbuf)

        if certtype == "client":
            cname_path = path_join(self.pkiworkdir, escapeFilename(cname))
            cname_req = reqoutfile
            cname_cert = cname_path +  "cert.tmp"
            ca_key = path_join(self.pkiworkdir, self.envconst['ca-privatekey'])
            ca_cert = path_join(self.pkiworkdir, self.envconst['ca-cert'])
            ca_serial = path_join(self.pkiworkdir, "serial")
            self.__run('x509', ['-extensions', 'usr_cert', '-CA', ca_cert, '-CAkey', ca_key, '-req', '-in', cname_req, '-out', cname_cert, '-CAcreateserial', '-CAserial', ca_serial])

            with open(cname_cert, "r") as reqoutfp:
                certbuf = reqoutfp.read()

            return certbuf

    def getx509param(self, certfile, param):
        certificate = self.__getCertPath(self.pkiname, certfile)
        parameters = ['-in', certificate, '-noout', "-%s" % (param)]
        return self.__run('x509', parameters)

    def revokeCert(self, cname):
        """
        Revoke the certificate provided as parameter
        """
        revoke_cert = path_join(self.pkiworkdir, escapeFilename(cname) + ".cert")
        ca_key = path_join(self.pkiworkdir, "ca.key")
        ca_cert = path_join(self.pkiworkdir, "ca.crt")
        crl_pem = path_join(self.pkiworkdir, "crl.pem")
        openssl_config = path_join(self.pkiworkdir, "openssl.cfg")
        runme = ("openssl", "ca",
            "-revoke", revoke_cert,
            "-keyfile", ca_key,
            "-cert", ca_cert,
            "-config", openssl_config)
        self.runCommand(runme)
        self.regenCRL()

    def _getCRLField(self, filename, argument):
        command = ("openssl", "crl", "-in", filename, argument, "-noout")
        handler = createProcess(self.logger, command, stdout=subprocess.PIPE)
        output = handler.communicate()[0]
        self.checkCommandStatus(handler.returncode)
        return self.__extractVal(output)

    def regenCRL(self):
        ca_key = path_join(self.pkiworkdir, "ca.key")
        ca_cert = path_join(self.pkiworkdir, "ca.crt")
        crl_pem = path_join(self.pkiworkdir, "crl.pem")
        openssl_config = path_join(self.pkiworkdir, "openssl.cfg")
        runme = ("openssl", "ca", "-gencrl",
            "-keyfile", ca_key,
            "-cert", ca_cert,
            "-out", crl_pem,
            "-config", openssl_config)
        self.runCommand(runme)

        last_update = self._getCRLField(crl_pem, "-lastupdate")
        last_update = self.parseDatetime(last_update)

        next_update = self._getCRLField(crl_pem, "-nextupdate")
        next_update = self.parseDatetime(next_update)

        return (last_update, next_update)

    def renewCert(self, cert):
        self.revokeCert(cert.cname)
        current_cert = path_join(self.pkiworkdir, escapeFilename(cert.cname) + ".cert")
        newreq = path_join(self.pkiworkdir, escapeFilename(cert.cname) + ".csr")
        private_key = path_join(self.pkiworkdir, escapeFilename(cert.cname) + ".pkey")

        runme = ("openssl", "x509", "-x509toreq",
            "-in", current_cert,
            "-out", newreq,
            "-signkey", private_key)
        self.runCommand(runme)

        now = datetime_utcnow()
        if cert.end_date > now:
            end_date = datetime.now() + CERT_RENEW_ENDTIME
        else:
            end_date = cert.end_date + CERT_RENEW_ENDTIME

        self.signCertRequest(cert.type, cert.cname, end_date)

    def _getProcessOutput(self, *command):
        process = createProcess(self.logger, command, stdout=subprocess.PIPE)
        stdout = process.communicate()[0]
        self.checkCommandStatus(process.returncode)
        return stdout

    def getCACertText(self):
        return self._getProcessOutput("openssl", "x509",
            "-in", path_join(self.pkiworkdir, "ca.crt"),
            "-text", "-noout")

    def getCertText(self, filename):
        return self._getProcessOutput("openssl", "x509",
            "-in", filename,
            "-text", "-noout")

    def getCRLText(self, filename):
        return self._getProcessOutput("openssl", "crl",
            "-in", filename,
            "-text", "-noout")

    def getOpenVPNConfig(self):
        return {
            "HOME" : ".",
            "RANDFILE" : "$ENV::HOME/.rnd",
            # Please define your PKI dir. pki/VPN is arbitrary
            "CA_dir" : "/var/lib/nucentral/nupki2/pki/VPN",
            "CA_x509_extensions" : "usr_cert",
            "CA_default_days" : "3650",
            "CA_default_crl_days" : "30",
            "CA_default_md" : "md5",
            "CA_preserve" : "no",
            "req_default_bits" : "$ENV::KEY_SIZE",
            "req_default_keyfile" : "privkey.pem",
            "req_dn_countryName_default" : "FR",
            "req_dn_stateOrProvinceName_default" : "Ile-de-France",
            "req_dn_organizationalUnitName_default" : "RetD",
            "req_dn_localityName_default" : "Paris",
            "req_dn_0_organizationName_default" : "Edenwall",
            "req_dn_emailAddress_default" : "root@edenwall.com"}

    def parseDatetime(self, value):
        old_locale = setlocale(LC_ALL, "C")
        try:
            try:
                # 'Mar 23 19:10:12 2010 GMT'
                return datetime.strptime(value, "%b %d %H:%M:%S %Y %Z")
            except ValueError:
                # 'Mar 23 19:10:12 2010'
                return datetime.strptime(value, "%b %d %H:%M:%S %Y")
        finally:
            setlocale(LC_ALL, old_locale)

