"""
Copyright (C) 2008-2011 EdenWall Technologies
Written by Sebastien Tricaud <stricaud AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""
from __future__ import with_statement

from base64 import b64decode, b64encode
from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED
from os.path import join as path_join
from shutil import copyfile
from time import mktime
from cStringIO import StringIO
from os import umask

from twisted.internet.defer import inlineCallbacks, returnValue

#from nucentral.common import tr
from nucentral.backend import Component
from nucentral.backend.cron import scheduleDaily, scheduleRepeat
from nucentral.backend.error import CoreError
from nucentral.core import events
from nucentral.core.context import Context
from nucentral.backend.exceptions import ConfigError

from nupki.version import VERSION
#from nupki.core.error import NuPKIError
from nupki.core.pki import PKIManager
from nupki.core.ssl_wrapper import SSLWrapper
from .client import Client

# Trigger CRL regeneration every week:
CRL_REGENERATE_PERIOD = 60 * 60 * 24 * 7

class NupkiModule(Component):
    NAME = "nupki"
    VERSION = VERSION
    API_VERSION = 1
    REQUIRES = ('config',)
    ACLS = {'config': ('getValue',),
            'ha' : ('getHAActive',),
            'nupki': ('getCertPath', 'getPrivateCertPath', 'getCACertPath')
    }
    ROLES = {
        'nupki_read': set((
            "listPKI", "listClientCerts", "listServerCerts", "getCertStatus",
            "getX509",
            "getCertPath", "getCert", "getCACertPath", "getCACert",
            "getPrivateCertPath", "getPrivateCert",
            "getZIPCert", "getCACertX509Text",
            "getCRLBuffer", "setupClient",
            "getPKIsettings", "getCertText", "getCRLText",
        )),
        'nupki_write': set(("@nupki_read",
            "copyPKI", "createPKI", "deletePKI",
            "createCert",
            "revokeCert", "revoke",
            "tagCertificates",
            "updateCertSettings", "updatePKISettings", "newCert",
            "forceCertCreation", "forceEachCertAction", "getCertRequest",
            "signRequest", "createCertificate", "undo", "regenCRL",
        )),
    }
    DEFAULT_RESULT = 'ok'

    def init(self, nucentral_core):
        self.nucentral_core = nucentral_core
        vardir = nucentral_core.var_dir
        self.conf = {'nupkidir': path_join(vardir, 'nupki2'),
                     'pkidir': path_join(vardir, 'nupki2', 'pki'),
                     'templates': path_join(vardir, 'nupki2', 'templates'),
                     'dbfile': path_join(vardir, 'nupki2','certs.db'),
                     'default_orgunit': 'Research and Development',
                     'default_org': 'Edenwall',
                     'default_location': 'Paris',
                     'default_state': 'Ile-de-France',
                     'default_country': 'FR',
                     'crontask_hour': '6',
                     'crontask_minute': '20',
                     }

        for key in self.conf.keys():
            try:
                self.conf[key] = self.nucentral_core.config_manager.getValue('nupki', 'config', key)
            except ConfigError:
                pass

        self.nupki = PKIManager(self, self.conf)

        events.connect('nucentralServerStarted', self._nucentralStarted)

    def _nucentralStarted(self):
        scheduleDaily(int(self.conf['crontask_hour']),
                      int(self.conf['crontask_minute']),
                      self.dailyCronTask)
        scheduleRepeat(CRL_REGENERATE_PERIOD, self.periodicCRLRegeneration)

    def __nonzero__(self):
        # "if attr: ..." is always True
        return True

    @inlineCallbacks
    def dailyCronTask(self):
        """
        cron task that must be run daily
        """
        core = self.nucentral_core
        context = Context.fromComponent(self)
        try:
            is_active = yield core.callService(context, 'ha', 'getHAActive')

            if is_active:
                self.nupki.applyAllPKIActions()
            else:
                self.debug('Ignoring certificate creation as this is a passive host')
        except CoreError, err:
            # Force the creation when the 'ha' component is not loaded
            self.nupki.applyAllPKIActions()

    def periodicCRLRegeneration(self):
        # Refresh CRLs
        pkis = self.nupki.listPKI()
        self.warning('Regenerating CRLs')
        for pki in pkis:
            self.nupki.regenCRL(pki.name)

    def getClient(self, context):
        session = context.getSession()
        if 'nupki_client' not in session:
            client = Client({})
            session['nupki_client'] = client
        return session['nupki_client']

    def service_setupClient(self, context, attr):
        """
        Setup the client. Attributes:

         * version (unicode): Client version

        Return a dictionary with keys:

         * version (unicode): Server version
        """
        client = Client(attr)
        session = context.getSession()
        session['nupki_client'] = client
        return client.exportXMLRPC()

    def setModified(self):
        """
        return a deferred
        """
        return self.nucentral_core.notify.emit(self.NAME, 'configModified')

    #
    # PKI stuff
    #

    def service_createPKI(self, context, pkiname, orgunit, org, location, state, country, expiretime=None):
        self.nupki.newPKI(pkiname, orgunit, org, location, state, country, expiretime)
        return self.setModified()

    @inlineCallbacks
    def service_deletePKI(self, context, pkiname):
        removed = self.nupki.deletePKI(pkiname)
        yield self.setModified()
        returnValue(removed)

    def service_listPKI(self, context):
        client = self.getClient(context)
        pkis = self.nupki.listPKI()
        if client.use_named_list:
            # client using the new namedlist API
            return [pki.exportXMLRPC() for pki in pkis]
        else:
            # old client using tuple API
            return [tuple(pki) for pki in pkis]

    def service_getPKIsettings(self, context, pkiname):
        client = self.getClient(context)
        pki = self.nupki.getPKI(pkiname)
        if client.use_named_list:
            # client using the new namedlist API
            return pki.exportXMLRPC()
        else:
            # old client using tuple API
            return (pki.organization_unit, pki.organization, pki.location, pki.state, pki.country)
    #
    # Certificates stuff

    @inlineCallbacks
    def service_createCertificate(self, context, attr):
        """
        Create a new certificate.

        Mandatory attributes:
         * pki (unicode): PKI name
         * cname (unicode): Certificate name

        Optional attributes:
         * type (unicode): "client" or "server" (default: client)
         * email (unicode)
         * organization_unit (unicode)
         * location (unicode)
         * state (unicode)
         * country (unicode)
         * serial (int): serial number (default: 0)
         * force (bool): create directly the certificate if force=True,
           otherwise only create the certification request (default: False).
        """
        result = self.nupki.createCertificate(attr)
        yield self.setModified()
        returnValue(result)

    def service_forceEachCertAction(self, context, pkiname):
        self.nupki.applyAllPKIActions(pkiname)
        return self.setModified()

    # FIXME: broken (unused) service (see ticket #1187)
    def service_updatePKISettings(self, context, pkiname, orgunit, organization, location, state, country):
        self.nupki.updatePKISettings(pkiname, orgunit, organization, location, state, country)
        return self.setModified()

    def service_revokeCert(self, context, pkiname, cname):
        sslwrap = SSLWrapper(self.nupki, pkiname)
        sslwrap.revokeCert("%s" % cname)
        return self.setModified()

    def listCerts(self, context, pki_name, cert_type):
        client = self.getClient(context)
        certs = self.nupki.listCerts(pki_name, cert_type)
        return [client.exportCertificate(cert) for cert in certs]

    def service_listClientCerts(self, context, pki_name):
        return self.listCerts(context, pki_name, "client")

    def service_listServerCerts(self, context, pki_name):
        return self.listCerts(context, pki_name, "server")

    def service_getCertRequest(self, context, pkiname, cname, type):
        sslwrap = SSLWrapper(self.nupki, pkiname)
        sslwrap.createCertRequest(type, cname)

        keyfile = self.nupki.getPrivateCertPath(pkiname, cname)
        with open(keyfile) as f:
            key = f.read()

        reqfile = self.nupki.getCertRequestPath(pkiname, cname)
        with open(reqfile) as f:
            req = f.read()

        return (req, key)

    def service_signRequest(self, context, pkiname, cname, req, type):
        self.nupki.signRequest(pkiname, cname, req, type)
        return self.setModified()

    def service_getCertPath(self, context, pkiname, cname):
        return self.nupki.getCertPath(pkiname, cname)

    def service_getCert(self, context, pkiname, cname):
        return self.nupki.getCertContent(pkiname, cname)

    def service_getCRLBuffer(self, context, pkiname):
        return self.nupki.getCRLBuffer(pkiname)

    def service_getPrivateCertPath(self, context, pkiname, cname):
        return self.nupki.getPrivateCertPath(pkiname, cname)

    def service_getPrivateCert(self, context, pkiname, cname):
        return self.nupki.getPrivateCert(pkiname, cname)

    def service_getCACertPath(self, context, pkiname):
        return self.nupki.getCACertPath(pkiname)

    def service_getCAPrivPath(self, context, pkiname):
        return self.nupki.getCAPrivPath(pkiname)

    def service_getCACert(self, context, pkiname):
        return self.nupki.getCACert(pkiname)

    def service_getCACertX509Text(self, context, pkiname):
        return self.nupki.getCACertX509Text(pkiname)

    def service_getCRLText(self, context, pkiname):
        return self.nupki.getCRLText(pkiname)

    def service_getZIPCert(self, context, pki_name, cname, type="client", vpnconf=None):
        cafile = self.nupki.getCACertPath(pki_name)
        crlfile = self.nupki.getCrlPath(pki_name)
        public = self.nupki.getCertPath(pki_name, cname)
        priv = self.nupki.getPrivateCertPath(pki_name, cname)
        tmpfile = StringIO()

        zf = ZipFile(tmpfile, 'w', compression=ZIP_DEFLATED)
        zf.write(cafile, "ca.crt")
        zf.write(crlfile, "crl.crt")
        if type == "client":
            zf.write(priv, "client.key")
            zf.write(public, "client.crt")
            if vpnconf:
                vpnconf = b64decode(vpnconf)
                info = ZipInfo(filename="client.ovpn")
                info.compress_type = zf.compression
                # Unix attributes
                info.external_attr = 0600 << 16
                zf.writestr(info, vpnconf)
        else:
            zf.write(priv, "server.key")
            zf.write(public, "server.crt")
        zf.close()

        destbuf = tmpfile.getvalue()
        return b64encode(destbuf)

    @inlineCallbacks
    def service_tagCertificates(self, context, cert_list, new_status):
        """
        Set the new status of a certificate list:
         - cert_list is a list of dictionaries with keys 'pkiname' and 'cname'
         - new_status is 'to be revoked' or 'to be renewed'.

        Return the number of updated certificates.
        """
        updated = 0
        for cert in cert_list:
            ok = self.nupki.setNewStatus(cert['pkiname'], cert['cname'], new_status)
            if ok:
                updated += 1
        if updated:
            yield self.setModified()
        returnValue(updated)

    @inlineCallbacks
    def service_revoke(self, context, pkiname, cname, serial):
        self.nupki.doRevoke(pkiname, cname, serial)
        yield self.setModified()
        returnValue('ok')

    def service_runtimeFiles(self, context):
        return {
                'deleted': (self.conf['pkidir'],self.conf['dbfile'],),
                'added': ((self.conf['pkidir'], 'dir'),(self.conf['dbfile'], 'sqlite'),),
            }

    def service_runtimeFilesModified(self, context):
        """
        What we do when something is modified by a third party? -> nothing!
        """
        self.nupki.reload()

    def service_copyPKI(self, context, pkiname, cname, types):
        """
        Copy the files of the certificate "cname" of the PKI "pkiname" to the
        specified paths. types is a dictionary: file type => full path name.
        File types: "certificate", "key", "ca".

        Eg. pkiname="nucentral", cname="server",
            types={'ca': '/etc/nucentral/ca.pem',
                   'certificate': '/etc/nucentral/cert.pem',
                   'key': '/etc/nucentral/key.pem'}
        """
        # FIXME: block user context, but it's used by roadwarrior,
        # authentication and setup_ew4.py
#        if context.isUserContext():
#            # This service allows to copy a PKI files to an arbitrary
#            # directory. Don't trust users.
#            raise NuPKIError(
#                tr("The service copyPKI() can only used by a component."))
        paths = {
            'ca': self.nupki.getCACertPath(pkiname),
            'crl': self.nupki.getCrlPath(pkiname),
            'certificate': self.nupki.getCertPath(pkiname, cname),
            'key': self.nupki.getPrivateCertPath(pkiname, cname),
        }

        umask(0177)
        for key, source in paths.iteritems():
            if key not in types:
                continue
            dest = types[key]
            copyfile(source, dest)

    def service_undo(self, context, pki_name):
        """
        Undo last actions of the specified PKI: remove certificate creation
        requests.

        Return the number of deleted certificates.
        """
        return self.nupki.undo(pki_name)

    def service_getCertText(self, context, pki_name, cname):
        """
        Get a certificate description as text.
        """
        return self.nupki.getCertText(pki_name, cname)

    # copyCRL should only used by other components, and so it's not declared
    # in the ROLES list
    def service_copyCRL(self, context, pki_name, dest_filename):
        """
        Copy the CRL file to the destination.
        """
        src_filename = self.nupki.getCrlPath(pki_name)
        umask(0177)
        copyfile(src_filename, dest_filename)

    #--- Deprecrated services kept for backward compatibility ----------------
    #--- (compatibility with old clients) ------------------------------------

    # Deprecated service: replaced by newCertificate()
    # Used by multisite master <= 4.0.12
    @inlineCallbacks
    def service_createCert(self, context, pkiname, cname, type, serial, email,
    orgunit, organization, location, state, country):
        """
        Create a new certificate.

        Note: createCertificate() and newCert() create a certificate creation
              request, but createCreate() creates directly the certificate.
        """
        attr = {
            'pki': pkiname,
            'cname': cname,
            'type': type,
            'serial': serial,
            'email': email,
            'organization_unit': orgunit,
            'location': location,
            'state': state,
            'country': country,
            'force': True,
            # organization value was not set by old nupki clients
        }
        result = self.nupki.createCertificate(attr)
        yield self.setModified()
        returnValue(result)

    # Deprecated service: replaced by newCertificate()
    # Used by nupki client < 4.0.1 and multisite slave <= 4.0.12
    def service_newCert(self, context, pkiname, cname, type, expiretime=None):
        """
        Create a certificate creation request. You have to call
        updateCertSettings() to finish the creation.
        """
        self.nupki.newCert(pkiname, cname, type, expiretime)
        return self.setModified()

    # Deprecated: replaced by the createCertificate() service
    # Used by nupki client < 4.0.1 and multisite slave <= 4.0.12
    @inlineCallbacks
    def service_updateCertSettings(self, context, pkiname, cname, serial, email, orgunit, organization, location, state, country):
        self.nupki.updateCertificateSettings(pkiname, cname, serial, email, organization, orgunit, location, state, country)
        yield self.setModified()
        returnValue("ok")

    # Deprecated: replaced by the forceEachCertAction() service
    # Used by nupki client < 4.0.1
    def service_forceCertCreation(self, context, pkiname, cname, serial, cert_type):
        # FIXME: serial is no more used
        self.nupki.forceCertCreation(pkiname, cname, cert_type)
        return self.setModified()

    # Deprecated: listClientCerts() and listServerCerts() already include the status
    # Used by nupki client < 4.0.1
    def service_getCertStatus(self, context, pkiname, cname):
        return self.nupki.getCertStatus(pkiname, cname)

    # Deprecated: listClientCerts() and listServerCerts() already include the
    # start_date and end_date. Used by nupki client <= 4.0.1
    def service_getX509(self, context, value, pkiname, cname):
        """
        Return the value from a X509 certificate. Runs OpenSSL, extract the
        relevant field.
        """
        timestamp = self.nupki.getX509Field(pkiname, cname, value)
        timestamp = mktime(timestamp.timetuple())
        return str(int(timestamp))

    # Deprecated: was never really used
    # Used by nupki client < 4.0.1
    def service_getSerial(self, context, pkiname, type, cname):
        sslwrap = SSLWrapper(self.nupki, pkiname)
        return sslwrap.get_serial(cname)

    def service_regenCRL(self, ctx, pkiname):
        last_update, next_update = self.nupki.regenCRL(pkiname)
        return (unicode(last_update), unicode(next_update))

