
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from datetime import datetime
import locale

from nucentral.common.transport import parseDatetime, TransportError
from nucentral.common.timezone import datetime_set_utc
from nupki.common.pki import PKI, Certificate, DEFAULT_CERT_EXPIRE

class NuPKIClient:
    """
    Wrapper around service calls supporting old server API (backward
    compatibility).
    """
    def __init__(self, client, compatibility):
        self.client = client
        self.compatibility = compatibility

    def createCertificate(self, attr):
        if self.compatibility.has_create_certificate:
            if 'expire' in attr:
                del attr['expire']
            self.client.call("nupki", "createCertificate", attr)
        else:
            serial = "1"
            args = [attr['pki'], attr['cname'], attr['type']]
            if ('expire' in attr) \
            and self.compatibility.newcert_has_expire:
                args.append(unicode(attr['expire']))
            self.client.call("nupki", "newCert", *args)
            self.client.call("nupki", "updateCertSettings",
                attr['pki'], attr['cname'], serial, attr['email'],
                attr['organization_unit'], attr['organization'],
                attr['location'], attr['state'], attr['country'])

    def parseDatetime(self, result):
        dt = self._parseDatetime(result)
        if dt.tzinfo is None:
            dt = datetime_set_utc(dt)
        return dt

    def _parseDatetime(self, result):
        # nupki 4.0.3: '2010-03-22 14:49:27', datetime format
        try:
            return parseDatetime(result)
        except TransportError:
            pass

        # nupki 4.0.1/4.0.2: '1580389797', UNIX epoch
        try:
            epoch = int(result)
            return datetime.fromtimestamp(epoch)
        except ValueError:
            pass

        # nupki milestone 4: 'Jan 31 11:55:19 2020 GMT' (OpenSSL output)
        old_locale = locale.setlocale(locale.LC_ALL, "C")
        try:
            return datetime.strptime(result, "%b %d %H:%M:%S %Y %Z")
        finally:
            locale.setlocale(locale.LC_ALL, old_locale)

    def listPKI(self):
        pki_data = self.client.call("nupki", "listPKI")
        if self.compatibility.use_named_list:
            pkis = [PKI(**columns) for columns in pki_data]
        else:
            pkis = [PKI(*columns) for columns in pki_data]
        # filter hidden PKIs
        pkis = [pki for pki in pkis if not pki.name.startswith("_")]
        return pkis

    def getPKI(self, name):
        columns = self.client.call("nupki", "getPKIsettings", name)
        if self.compatibility.use_named_list:
            return PKI(**columns)
        else:
            return PKI(
                organization_unit=columns[0],
                organization=columns[1],
                location=columns[2],
                state=columns[3],
                country=columns[4],
                name=name,
            )

    def getCertList(self, pki_name, cert_type):
        pki = self.getPKI(pki_name)
        return self.getPkiCerts(pki, cert_type)

    def getPkiCerts(self, pki, cert_type):
        now = datetime.utcnow()
        if cert_type == "client":
            service = "listClientCerts"
        else:
            service = "listServerCerts"
        certs = self.client.call("nupki", service, pki.name)

        if self.compatibility.use_named_list:
            cert_list = [Certificate(**columns) for columns in certs]
        else:
            cert_list = [Certificate(*columns) for columns in certs]

        for cert in cert_list:
            if cert.organization is None:
                # compatibility with nupki < 4.0.3
                cert.organization = pki.organization
            if cert.pki_name is None:
                # compatibility with nupki < 4.0.3
                cert.pki_name = pki.name

            if cert.status == "to be created":
                if cert.end_date is not None:
                    cert.end_date = self.parseDatetime(cert.end_date)
                else:
                    cert.end_date = now + DEFAULT_CERT_EXPIRE
            else:
                if self.compatibility.use_named_list:
                    cert.start_date = self.parseDatetime(cert.start_date)
                    cert.end_date = self.parseDatetime(cert.end_date)
                else:
                    value = self.client.call("nupki", "getX509",
                        "startdate", pki.name, cert.cname)
                    cert.start_date = self.parseDatetime(value)

                    value = self.client.call("nupki", "getX509",
                        "enddate", pki.name, cert.cname)
                    cert.end_date = self.parseDatetime(value)

        return cert_list

    def getCRLText(self, pki_name):
        if self.compatibility.has_getcatext:
            return self.client.call("nupki", "getCRLText", pki_name)
        else:
            return self.client.call("nupki", "getCRLBuffer", pki_name)

