"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from __future__ import with_statement

import base64
from os.path import dirname, join as path_join

from PyQt4.QtCore import SIGNAL
from PyQt4.QtGui import QAbstractItemView, QHeaderView, QMenu, QMessageBox, QIcon, QFileDialog

from nucentral.common import tr
from nucentral.qt.tools import getSelectionRows

from .model import TableModel
from .buttons import ButtonsMask

class CertificateMenu(QMenu):
    def __init__(self, window, controler):
        QMenu.__init__(self, window)
        self.window = window
        self.controler = controler
        self.client = window.client
        self.create_before = self.add(None, tr("View certificate"), self.dump)
        self.certificate = None

    def add(self, icon, text, callback):
        if icon:
            action = self.addAction(QIcon(icon), text)
        else:
            action = self.addAction(text)
        self.connect(action, SIGNAL("triggered()"), callback)
        return action

    def dump(self):
        self.controler.displayCert(self.certificate)

    def display(self, event, certificate):
        self.certificate = certificate
        self.create_before.setEnabled(self.certificate.status != "to be created")
        if event:
            self.exec_(event.globalPos())

class TableControler(object):
    def __init__(self, window, tables, cert_type):
        self.ui = window.ui
        self.tables = tables
        self.window = window
        self.client = window.client
        self.cert_type = cert_type
        self.manager = window.manager

        self.model = TableModel(self.ui.tableClients, self.cert_type, self.window)
        if cert_type == "client":
            self.view = self.ui.tableClients
        else:
            self.view = self.ui.tableServers

        self.view.setModel(self.model)
        self.view.verticalHeader().hide()
        self.view.setShowGrid(True)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.horizontalHeader().setResizeMode(QHeaderView.Stretch)
        self.view.hideColumn(0)
        if window.compatibility.has_getcerttext:
            self.view.contextMenuEvent = self.contextMenuEvent
            self.view.connect(self.view, SIGNAL("doubleClicked(const QModelIndex &)"), self.showCertificate)

        self.menu = CertificateMenu(window, self)

        if cert_type == "server":
            filter_widget = self.ui.server_filter
        else:
            filter_widget = self.ui.client_filter
        self.view.connect(filter_widget, SIGNAL("textChanged(const QString &)" ), self.model.textFilter)

        self.view.connect(self.view.selectionModel(),
            SIGNAL("selectionChanged(QItemSelection, QItemSelection)"),
            self.selectionChanged)

        ui = self.ui
        self.buttons = {}
        for name, mask in (
        ("ToBeCreated", ButtonsMask.TOBECREATED),
        ("Valid",  ButtonsMask.VALID),
        ("Expired", ButtonsMask.EXPIRED),
        ("ToBeRevoked", ButtonsMask.TOBEREVOKED),
        ("ToBeRenewed",  ButtonsMask.TOBERENEWED),
        ("Revoked", ButtonsMask.REVOKED),
        ):
            widget = getattr(ui, self.cert_type + "_" + name)
            def createSetMask(mask):
                def setMask(checked):
                    self.setMask(checked, mask)
                return setMask
            window.connect(widget, SIGNAL("toggled(bool)"), createSetMask(mask))
            self.buttons[name] = widget

    def selectionChanged(self, selected, deselected):
        self.tables.refreshActions()

    def populate(self, pki):
        self.model.populate(pki, self.cert_type)

    def showCertificate(self, index):
        certs = self.getSelectedCerts()
        if not certs:
            return
        self.displayCert(certs[0])

    def setMask(self, enabled, button_mask):
        mask = self.model.mask
        if enabled:
            mask.add(button_mask)
        else:
            mask.remove(button_mask)
        self.model.applyMask(mask)

    def buttons_flush(self):
        self.model.buttons_flush()

    def hasSelection(self):
        selectionlist = self.view.selectionModel().selection()
        rows = getSelectionRows(selectionlist)
        return bool(rows)

    def getSelectedCerts(self):
        selectionlist = self.view.selectionModel().selection()
        return [
            self.model.getCertificate(row)
            for row in getSelectionRows(selectionlist)]

    def _button_labelize(self, button, format, count):
        widget = self.buttons[button]
        text = format % count
        widget.setText(text)
        widget.setEnabled(count != 0)

    def buttons_labelize(self):
        model = self.model
        self._button_labelize('ToBeCreated', tr("%s to be created"), model.visible_tobecreated)
        self._button_labelize('Valid', tr("%s valid", "", model.visible_valid), model.visible_valid)
        self._button_labelize('Expired', tr("%s expired", "", model.visible_expired), model.visible_expired)
        self._button_labelize('ToBeRevoked', tr("%s to be revoked"), model.visible_toberevoked)
        self._button_labelize('ToBeRenewed', tr("%s to be renewed"), model.visible_toberenewed)
        self._button_labelize('Revoked', tr("%s revoked", "", model.visible_revoked), model.visible_revoked)

    def setCertNewStatus(self, new_status):
        certs = self.getSelectedCerts()
        certs = [cert
            for cert in certs
            if cert.status not in (new_status, "to be created", "revoked")]

        if not certs:
            return

        pki = self.manager.getCurrentName()
        certIDlist = [
            {'pkiname': pki, 'cname': cert.cname, 'serial': "0"}
            for cert in certs]
        try:
            self.client.call("nupki", "tagCertificates", certIDlist, new_status)
        except Exception, err:
            self.window.exception(err)
            self.tables.refresh()
            return

        for cert in certs:
            self.model.updateCounters(new_status, cert.status)
            cert.new_status = new_status

        self.tables.refresh()

    def contextMenuEvent(self, event):
        certs = self.getSelectedCerts()
        if not certs:
            return
        self.menu.display(event, certs[0])

    def _writeCert(self, filename, pki, cname, certtype, vpnconf):
        zipbuffer = self.client.call("nupki", "getZIPCert", pki, cname, certtype, vpnconf)
        with open(filename, 'wb') as fp:
            fp.write(base64.b64decode(zipbuffer))

    def exportCertificates(self):
        certs = self.getSelectedCerts()
        if not certs:
            return

        if self.cert_type == "client":
            try:
                vpnconf = self.client.call("openvpn", "getClientConfig")
            except Exception:
                vpnconf = ""
        else:
            vpnconf = ""
        pki = self.manager.getCurrentName()

        known_storage = None
        already_asked = False
        automate = False
        for cert in certs:
            cname = cert.cname
            if cert.status != "valid":
                reply = QMessageBox.information(self.window,
                    tr("Cannot export the certificate"),
                    tr("The %s certificate is not valid and can not be exported") % cname)
                continue

            if known_storage and automate:
                filename = path_join(known_storage, cname + ".zip")
                self._writeCert(filename, pki, cname, self.cert_type, vpnconf)
            else:
                filename = "%s.zip" % cname.replace("/", "_")
                filename = QFileDialog.getSaveFileName(
                    None,
                    tr("Export a certificate"),
                    filename,
                    tr("Zipped certificate (*.zip)"))
                filename = unicode(filename)
                if filename:
                    self._writeCert(filename, pki, cname, self.cert_type, vpnconf)

                    if not already_asked:
                        known_storage = dirname(filename)

                if len(certs) > 1 and not already_asked:
                    reply = QMessageBox.question(self.window,
                        tr("Certificates storage automation"),
                        tr("Do you want to store all selected certificates in the same directory?"),
                        QMessageBox.No, QMessageBox.Yes)
                    if reply == QMessageBox.No:
                        known_storage = False
                        already_asked = True
                    else:
                        automate = True

    def displayCert(self, certificate):
        if certificate.status == "to be created":
            return
        try:
            text = self.client.call("nupki", "getCertText",
                certificate.pki_name, certificate.cname)
        except Exception, err:
            self.window.exception(err)
            return
        QMessageBox.information(self.window,
            tr("Certificate %s") % certificate.cname,
            u"<pre>%s</pre>" % text)

