# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2011 EdenWall Technologies
Written by Sebastien Tricaud <s.tricaud AT inl.fr>

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>

$Id: nupki_window.py 14505 2009-11-18 13:43:02Z toady $
"""

from __future__ import with_statement
from logging import INFO
import csv

from PyQt4.QtGui import QMessageBox, QDialog, QFileDialog
from PyQt4.QtCore import SIGNAL

from nucentral.client import NuCentralError
from nucentral.common import tr
from nucentral.common.transport import parseDatetime
from nucentral.common.error import exceptionAsUnicode
from nucentral.common.human import fuzzyDatetime
from nucentral.qt.central_window import CentralWindow, STANDALONE

from nucentral.common.logger import Logger
from nucentral.common import EDENWALL

from .ui import nupki_ui, deletepki_ui, settings_ui
from .tables import Tables
from .wizards import NewPkiWizard, NewCertificateWizard
from .compatibility import Compatibility
from .client import NuPKIClient
from .manager import PKIManager
from .user_settings import UserSettings

class MainWindow(CentralWindow):
    ROLES = set(('nupki_read', 'nupki_write'))
    TAB_CLIENT = 0
    TAB_SERVER = 1
    TAB_CA = 2
    TAB_CRL = 3
    ICON = ':/icons/certificat.png'

    # user_settings is not set if .checkLicense() calls sys.exit():
    # set it to None to avoid an error in .quit()
    user_settings = None

    def __init__(self, application, client, standalone=STANDALONE, parent=None, eas_window=None):
        CentralWindow.__init__(self, client, parent, eas_window)
        self.firstOpen = True
        self.logger = Logger("nupki")
        self.client = client
        self.setupCentralWindow(application, standalone)
        if EDENWALL:
            self.checkLicense()
        self.setupUI()
        self.user_settings = UserSettings(self)

    def checkLicense(self):
        noLicense = self.client.call("acl", "getMinimalMode")
        if noLicense:
            print "Invalid activation key."
            exit(1)

    def setupUI(self):
        self.ui = nupki_ui.Ui_MainWindow()
        self.ui.setupUi(self)

        self.deletepkiDialog = QDialog()
        self.deletePKIui = deletepki_ui.Ui_DeleteaPKI()
        self.deletePKIui.setupUi(self.deletepkiDialog)
        self.connect(self.deletePKIui.buttonBox, SIGNAL("accepted()"), self.__acceptDeletePKI)
        self.connect(self.deletePKIui.buttonBox, SIGNAL("rejected()"), self.__rejectDeletePKI)

        self.connect(self.deletePKIui.NocheckBox, SIGNAL("stateChanged(int)"), self.__Nocheckbox)

        self.settingsuiDialog = QDialog()
        self.settingsui = settings_ui.Ui_SettingsDialog()
        self.settingsui.setupUi(self.settingsuiDialog)

        self.connect(self.settingsui.buttonBox, SIGNAL("accepted()"), self.settings_accept)

        # Connect Toolbar actions
        self.connect(self.ui.actionNew_PKI, SIGNAL("triggered()"), self.__newPKIWizard)
        self.connect(self.ui.actionClose_PKI, SIGNAL("triggered()"), self.closePKI)
        self.connect(self.ui.actionDelete_PKI, SIGNAL("triggered()"), self.__deletePKI)
        self.connect(self.ui.actionNew_Certificate, SIGNAL("triggered()"), self.__newCertificateWizard)

        # FIXME: Reenable PKI settings (ticket #1187)
        #self.action_pkisettings = self.ui.menuPKI_Manager.addAction(tr("PKI parameters"), self.runSettings)
        self.action_csvimport = self.ui.menuPKI_Manager.addAction(tr("Create certificates from CSV"), self.__importCSV)
        self.ui.menuPKI_Manager.addSeparator()

        self.connect(self.ui.actionExport_CRL, SIGNAL("triggered()" ), self.downloadCRL)
        self.connect(self.ui.regenerate_crl_button, SIGNAL("clicked()" ), self.regenerateCRL)
        self.connect(self.ui.actionRenewCRL, SIGNAL("triggered()" ), self.regenerateCRL)

        self.setCentralWindowTitle(tr("PKI"), self.client)


    if EDENWALL:
        def isReadOnly(self):
            return not 'nupki_write' in self.getRoles()
    else:
        def isReadOnly(self):
            return False

    def getCurrentTab(self):
        return self.ui.tabWidget.currentIndex()

    def setCurrentTab(self, tab):
        return self.ui.tabWidget.setCurrentIndex(tab)

    def loadPKI(self, pki):
        # pki can be None
        self.manager.closePKI()
        self.tables.buttons_flush()

        if pki:
            name = pki.name
            self.setCentralWindowTitle(tr("PKI"), self.client, name)
            self.manager.setCurrentPKI(name)
            self.tables.loadPKI(pki)

            text = self.client.call("nupki", "getCACertX509Text", name)
            self.ui.CA_Edit.setText(text)

            self.refreshCRL(pki.name)
        else:
            self.setCentralWindowTitle(tr("PKI"), self.client)
            self.manager.setCurrentPKI(None)
            self.tables.loadPKI(None)

        self._refreshActions(bool(pki))


    def quit(self):
        # Save user settings
        if self.user_settings:
            try:
                self.user_settings.save()
            except Exception, err:
                self.writeError(err,
                    "Warning: Unable to save nupki configuration",
                    log_level=INFO)
            except KeyboardInterrupt:
                pass
        CentralWindow.quit(self)

    def load(self):
        try:
            self.compatibility = Compatibility(self.logger, self.client)
        except NuCentralError, err:
            standalone = self._standalone == STANDALONE
            self.error(tr("PKI backend not found! Exiting."), dialog=standalone)
            if standalone:
                exit(1)
            else:
                self.eas_window.setEnabledApplication(self, False)
                return

        self.nupki = NuPKIClient(self.client, self.compatibility)

        try:
            self.manager = PKIManager(self)
        except Exception, err:
            self.exception(err)
            return

        self.tables = Tables(self)

        current = self.manager.getCurrent()
        if current:
            # Load the first PKI
            self.loadPKI(current)
        else:
            # Disable actions
            self._refreshActions(False)

    def openThisPKI(self):
        action = self.sender()
        if action:
            name = unicode(action.text())
            pki = self.manager.pkis[name]
            self.loadPKI(pki)

    def __changedPKI(self, text):
        self.loadPKI(text)

    def information(self, message, title=None):
        if not title:
            title = tr("PKI")
        QMessageBox.information(self, title, message)

    def __acceptDeletePKI(self):
        if self.deletePKIui.NocheckBox.isChecked():
            self.__rejectDeletePKI()
            return

        if not self.deletePKIui.YescheckBox.isChecked():
            self.information(tr("Please check Yes or No"))
            return
        self.__rejectDeletePKI()

        pki_name = self.manager.getCurrentName()
        reply = QMessageBox.critical(self,
            tr("Delete selected PKI"),
            tr("Are you sure you want to delete the '%s' PKI and all associated certificates? "
               "Some certificates of this PKI can be used for other Edenwall services "
               "(VPN, Firewall, Authentication, ..). Remember to disable the "
               "associated services and apply the new change.\n"
               "This can trigger problems in your configuration.")
            % pki_name,
            QMessageBox.Abort, QMessageBox.Ok)
        if reply == QMessageBox.Abort:
            return
        self.deletepkiDialog.hide()

        try:
            self.client.call("nupki", "deletePKI", pki_name)
        except Exception, err:
            self.exception(err)
            return

        self.manager.deletePKI(pki_name)
        self.loadPKI(None)
        self.tables.refresh()

    def __rejectDeletePKI(self):
        self.deletepkiDialog.hide()

    def __Nocheckbox(self, state):
        if not state:
            self.deletePKIui.YescheckBox.setEnabled(True)
        else:
            self.deletePKIui.YescheckBox.setEnabled(False)

    def closePKI(self):
        self.ui.tabWidget.setCurrentIndex(0)
        self.loadPKI(None)

    def __deletePKI(self):
        self.deletePKIui.YescheckBox.setEnabled(False)
        self.deletePKIui.YescheckBox.setChecked(False)
        self.deletePKIui.NocheckBox.setEnabled(True)
        self.deletePKIui.NocheckBox.setChecked(True)

        self.selected_PKI = self.manager.getCurrentName()
        self.deletePKIui.PKI2delete.setText("<b>" + self.selected_PKI + "</b>")
        self.deletepkiDialog.show()

    def runSettings(self):
        pki = self.manager.getCurrent()

        self.settingsui.OrgUnitEdit.setText(pki.organization_unit)
        self.settingsui.OrgEdit.setText(pki.organization)
        self.settingsui.LocationEdit.setText(pki.location)
        self.settingsui.StateEdit.setText(pki.state)
        self.settingsui.CountryEdit.setText(pki.country)

        self.settingsuiDialog.show()

    def _refreshActions(self, has_pki):
        can_create = not self.isReadOnly()

        # Actions
        self.ui.actionNew_Certificate.setEnabled(has_pki and can_create)
        self.tables.refreshActions()
        self.ui.actionForce_Pending_Actions.setEnabled(has_pki and can_create)
        self.ui.actionRefresh.setEnabled(has_pki)
        self.ui.actionExport_CRL.setEnabled(has_pki)
        self.ui.actionRenewCRL.setEnabled(has_pki)

        # Menu
        self.ui.actionNew_PKI.setEnabled(can_create)
        # FIXME: Reenable PKI settings (ticket #1187)
        #self.action_pkisettings.setEnabled(has_pki)
        self.ui.actionClose_PKI.setEnabled(has_pki)
        self.ui.actionDelete_PKI.setEnabled(has_pki and can_create)
        self.action_csvimport.setEnabled(has_pki and can_create)
        self.ui.tabWidget.setEnabled(has_pki)
        self.ui.actionUndo.setEnabled(can_create and has_pki)

    def __newPKIWizard(self):
        """
        Run the PKI Wizard
        """
        self.pkiwiz = NewPkiWizard(self)
        self.pkiwiz.registerValidateCallback(lambda unused: self._refreshActions(True))

    def __newCertificateWizard(self):
        """
        Run the Certificate creator Wizard
        """
        self.certwiz = NewCertificateWizard(self)
        self.certwiz.registerValidateCallback(self.tables.newCertCreated)

    def settings_accept(self):
        orgunit = unicode(self.settingsui.OrgUnitEdit.text())
        org = unicode(self.settingsui.OrgEdit.text())
        location = unicode(self.settingsui.LocationEdit.text())
        state = unicode(self.settingsui.StateEdit.text())
        country = unicode(self.settingsui.CountryEdit.text())

        pki_name = self.manager.getCurrentName()
        self.client.call("nupki", "updatePKISettings", pki_name, orgunit, org, location, state, country)

    def downloadCRL(self):
        pki_name = self.manager.getCurrentName()
        filename = QFileDialog.getSaveFileName(None,
            tr("Export the CRL"),
            "crl.pem",
            tr("CRL file (crl.pem)"))
        if not filename:
            return

        crl_text = self.client.call("nupki", "getCRLBuffer", pki_name)
        with open(filename, 'w') as fp:
            fp.write(crl_text)

    def regenerateCRL(self):
        self.setCurrentTab(self.TAB_CRL)
        pki_name = self.manager.getCurrentName()
        result = self.client.call("nupki", "regenCRL", pki_name)
        if isinstance(result, (list, tuple)):
            last_update, next_update = result
            next_update = parseDatetime(next_update)
            message = tr("CRL regenerated: valid until %s.") % fuzzyDatetime(next_update)
        else:
            message = tr("CRL regenerated.")
        self.refreshCRL(pki_name)
        self.information(message)

    def __importCSV(self):
        # CSV columns:
        # my cname, 'client' or 'server', my@cname.org, myorgunit, myorg, mylocation, mystate, mycountry

        filename = QFileDialog.getOpenFileName(None,
            tr("Select the CSV file to open"),
            ".",
            tr("CSV file (*.csv)"))
        filename = unicode(filename)
        if not filename:
            return

        pki = self.manager.getCurrentName()
        with open(filename) as fp:
            csvfile = csv.reader(fp, delimiter=',')
            for row in csvfile:
                if not row:
                    # ignore empty line
                    continue
                cname, cert_type, email, orgunit, org, location, state, country = row[:8]
                try:
                    self.nupki.createCertificate({
                        'pki': pki,
                        'cname': cname,
                        'type': cert_type,
                        'email': email,
                        'organization': org,
                        'organization_unit': orgunit,
                        'location': location,
                        'state': state,
                        'country': country,
                    })
                except Exception, err:
                    self.logger.error("Error on certification creation: %s" % exceptionAsUnicode(err))
        self.tables.refresh()

    def refreshCRL(self, pki_name):
        text = self.nupki.getCRLText(pki_name)
        self.ui.CRL_Edit.setText(text)

