
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from nucentral.common.timezone import datetime_utcnow

from PyQt4.QtCore import Qt, QVariant, QAbstractTableModel, QModelIndex
from PyQt4.QtGui import QColor

from nucentral.common import tr

from .buttons import ButtonsMask

STATUS_LABEL = {
    "valid": tr("Valid"),
    "to be created": tr("To be created"),
    "to be revoked": tr("To be revoked"),
    "revoked": tr("Revoked"),
    "to be renewed": tr("To be renewed"),
}

class TableModel(QAbstractTableModel):
    """
    Handles the table model for Nupki. In short, it will get information from the database
    to populate what is required for a PKI: CA, client and server certs.

    Functions related to the table: __init__(), rowCount(), columnCount(), data(), headerData()
    Functions related to the database: populate()
    """
    COLUMNS = (
        ('cname', tr("Common Name")),
        ('email', tr("Email address")),
        ('organization', tr("Organization")),
        ('organization_unit', tr("Organizational Unit")),
        ('location', tr("Location")),
        ('state', tr("State")),
        ('country', tr("Country")),
        ('status', tr("Status")),
    )

    def __init__(self, parent, tabletype, window):
        QAbstractTableModel.__init__(self, parent)
        self.window = window
        self.tabletype = tabletype
        # Contains everything
        self.certs = []
        # Contains only what we display
        self.displayedCerts = []
        self.mask = ButtonsMask.ALL

        # Counters for buttons
        self.visible_tobecreated = 0
        self.visible_valid       = 0
        self.visible_expired     = 0
        self.visible_toberevoked = 0
        self.visible_toberenewed = 0
        self.visible_revoked     = 0

    def getCertStatus(self, cert):
        if cert.new_status:
            return cert.new_status
        else:
            return cert.status

    def applyMask(self, mask):
        self.mask = mask
        self.displayedCerts = [cert for cert in self.certs if self.getCertStatus(cert) in self.mask]
        self.reset()

    def rowCount(self, index=QModelIndex()):
        return len(self.displayedCerts)

    def columnCount(self, index=QModelIndex()):
        return len(self.COLUMNS) + 1

    def flush(self):
        self.certs = []
        self.displayedCerts = []
        self.reset()

    def isExpired(self, cert):
        if self.getCertStatus(cert) == "to be created":
            return False
        now = datetime_utcnow()
        return (now < cert.start_date) or (cert.end_date < now)

    def __colorizeData(self, cert):
        if self.isExpired(cert):
            return QVariant(QColor(149, 149, 149))

        status = self.getCertStatus(cert)
        if status == "valid":
            return QVariant(QColor(122, 180, 29))
        if status == "to be created":
            return QVariant(QColor(202, 225, 165))
        if status == "to be revoked":
            return QVariant(QColor(234, 117, 44))
        if status == "revoked":
            return QVariant(QColor(193, 41, 46))
        if status == "to be renewed":
            return QVariant(QColor(248, 233, 48))

        return QVariant()

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < len(self.displayedCerts)):
            return QVariant()

        cert = self.displayedCerts[index.row()]
        column = index.column()

        if role == Qt.BackgroundRole:
            return self.__colorizeData(cert)

        if role != Qt.DisplayRole:
            return QVariant()

        attr = self.COLUMNS[column-1][0]
        if attr == 'status':
            value = self.getCertStatus(cert)
            value = STATUS_LABEL[value]
            if self.isExpired(cert):
                value = u'%s (%s)' % (tr("Expired"), value)
        else:
            value = getattr(cert, attr)
        return QVariant(value)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self.COLUMNS[section-1][1])

        return QVariant()

    def updateCounters(self, toincrease, todecrease = None):
        """
        On top of the window you can see counters with certificate status.
        Checking/Unchecking the certificate.
        toincrease and todeacrese must be one of the following:
        * None
        * "to be created"
        * "valid"
        * "expired"
        * "to be revoked"
        * "to be renewed"
        * "revoked"
        """
        if todecrease:
            if todecrease == "to be created":
                if self.visible_tobecreated <= 0:
                    return
                self.visible_tobecreated -= 1
            if todecrease == "valid":
                if self.visible_valid <= 0:
                    return
                self.visible_valid -= 1
            if todecrease == "expired":
                if self.visible_expired <= 0:
                    return
                self.visible_expired -= 1
            if todecrease == "to be revoked":
                if self.visible_toberevoked <= 0:
                    return
                self.visible_toberevoked -= 1
            if todecrease == "to be renewed":
                if self.visible_toberenewed <= 0:
                    return
                self.visible_toberenewed -= 1
            if todecrease == "revoked":
                if self.visible_revoked <= 0:
                    return
                self.visible_revoked -= 1

        if toincrease:
            if toincrease == "to be created":
                self.visible_tobecreated += 1
            if toincrease == "valid":
                self.visible_valid += 1
            if toincrease == "expired":
                self.visible_expired += 1
            if toincrease == "to be revoked":
                self.visible_toberevoked += 1
            if toincrease == "to be renewed":
                self.visible_toberenewed += 1
            if toincrease == "revoked":
                self.visible_revoked += 1

    #
    # Methods from here concern the database
    #
    def populate(self, pki, cert_type):
        """
        Populate the table from data in the database
        """
        #client = self.window.client
        nupki = self.window.nupki

        self.flush()
        if not pki:
            return

        certificates = nupki.getPkiCerts(pki, cert_type)
        for cert in certificates:
            # Tell buttons the text to apply
            self.updateCounters(self.getCertStatus(cert))

            self.displayedCerts.append(cert)
            self.certs.append(cert)
            self.insertRows(1)

    def insertRows(self, position, rows=1, index=QModelIndex()):
        self.beginInsertRows(QModelIndex(), position, position + rows - 1)
        self.endInsertRows()

    def append(self, cn, email, orgunit, organization, location, state, country, status):
        certfields = [2, '%s' % (cn), email, orgunit, organization, location, state, country, status]
        self.displayedCerts.append(certfields)
        self.insertRows(1)

    def textFilter(self, pattern):
        pattern = unicode(pattern)
        self.displayedCerts = []
        for cert in self.certs:
            if pattern in cert.cname:
                self.displayedCerts.append(cert)
        self.reset()

    def buttons_flush(self):
        self.visible_tobecreated = 0
        self.visible_valid = 0
        self.visible_expired = 0
        self.visible_toberevoked = 0
        self.visible_toberenewed = 0
        self.visible_revoked = 0

    def getCertificate(self, row):
        return self.displayedCerts[row]

