
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from copy import deepcopy

from M2Crypto.m2 import X509_PURPOSE_SSL_SERVER

from PyQt4.QtCore import SIGNAL, QString
from PyQt4.QtGui import QWidget, QFileDialog, QMessageBox

from nucentral.common.logger import Logger
from nucentral.common.ssl.config import SSLConfig
from nucentral.common.download import getEncodedFileContent
from nucentral.common import tr

from .compatibility import Compatibility
from .client import NuPKIClient
from .ui.nupki_embed_ui import Ui_PkiEmbedWidget

class PkiEmbedWidget(QWidget):
    (CRL_OPTIONAL, CERT_SHOW, KEY_SHOW, CA_SHOW, CRL_SHOW) = ( 1 << x for x in range(5))
    SHOW_ALL = CERT_SHOW | KEY_SHOW | CA_SHOW | CRL_SHOW

    DEFAULT_CONF = {
        'use_nupki' :   True,
        'nupki_pki' :   u'',
        'nupki_cert' :  u'',
        'cert' :        u'',
        'key' :         u'',
        'ca' :          u'',
        'crl' :         u'',
        'disable_crl' : False,
    }

    ATTR_TO_LABEL = {
        'cert' :    tr('Certificate'),
        'key' :     tr('Private key'),
        'ca' :      tr('Certificate authority'),
        'crl' :     tr('Certificate revocation list'),
    }

    def __init__(self, client=None, parent=None,  component=None, flags = SHOW_ALL, modify_cb = None, cert_type = X509_PURPOSE_SSL_SERVER):
        QWidget.__init__(self, parent)
        self.ui = Ui_PkiEmbedWidget()
        self.ui.setupUi(self)
        self.modify_cb = modify_cb
        self.cert_type = cert_type
        self.component = component
        self.client = client
        self.component_has_cert_info = 'getCertificatesInfo' in self.client.call("CORE", "getServiceList", component)

        self.logger = Logger("nupki_embedded")
        self.compatibility = Compatibility(self.logger, self.client)
        self.nupki = NuPKIClient(self.client, self.compatibility)

        self.attr_to_widget = {
            'cert'  : {
                'lineedit' :    self.ui.cert_lineedit,
                'info' :        self.ui.cert_info,
                'view' :        self.ui.cert_view,
                'browse' :      self.ui.cert_browse,
                'label' :       self.ui.cert_label,
            },
            'key'  : {
                'lineedit' :    self.ui.key_lineedit,
                'info' :        self.ui.key_info,
                'view' :        self.ui.key_view,
                'browse' :      self.ui.key_browse,
                'label' :       self.ui.key_label,
            },
            'ca'  : {
                'lineedit' :    self.ui.ca_lineedit,
                'info' :        self.ui.ca_info,
                'view' :        self.ui.ca_view,
                'browse' :      self.ui.ca_browse,
                'label' :       self.ui.ca_label,
            },
            'crl'  : {
                'lineedit' :    self.ui.crl_lineedit,
                'info' :        self.ui.crl_info,
                'view' :        self.ui.crl_view,
                'browse' :      self.ui.crl_browse,
                'label' :       self.ui.crl_label,
            },
        }

        self.hideUnused(flags)

        # signals
        self.connect(self.ui.nupki_radio, SIGNAL("toggled(bool)"), self.showUsePKI)
        self.connect(self.ui.user_radio, SIGNAL("toggled(bool)"), self.showUploadCerts)
        self.connect(self.ui.nupki_pki, SIGNAL("currentIndexChanged(int)"), lambda x: self.loadCertsFromPKI())
        self.connect(self.ui.nupki_cert, SIGNAL("currentIndexChanged(int)"), lambda x: self.modify_cb())
        self.connect(self.ui.nupki_refresh, SIGNAL("clicked()"), self.feed)

        self.connect(self.ui.crl_checkbox, SIGNAL("clicked()"), self.enableCrl)
        self.connect(self.ui.crl_checkbox, SIGNAL("clicked()"), self.modify_cb)

        ## Path line edit signals
        for widget in self.attr_to_widget.itervalues():
            self.connect(widget['lineedit'], SIGNAL('textChanged(QString)'), lambda x:self.modify_cb())

        def make_lambda(func, attr):
            return lambda: func(attr)

        ## View informations signals
        for attr, widget in self.attr_to_widget.iteritems():
            self.connect(widget['view'], SIGNAL('clicked()'), make_lambda(self.viewInfo, attr))

        ## Browse signals
        for attr, widget in self.attr_to_widget.iteritems():
            self.connect(widget['browse'], SIGNAL('clicked()'), make_lambda(self.browseFile, attr))

        self.feed()

    def hideUnused(self, flags):
        flag_to_attr = {
            self.CERT_SHOW : 'cert',
            self.KEY_SHOW :  'key',
            self.CA_SHOW :   'ca',
            self.CRL_SHOW :  'crl',
        }

        for flag in [self.CERT_SHOW, self.KEY_SHOW, self.CA_SHOW, self.CRL_SHOW]:
            if not (flags & flag):
                # Flag for this attribue is not enabled,
                # hide all related widgets:
                attr = flag_to_attr[flag]
                for widget in self.attr_to_widget[attr]:
                    widget.hide()

        if (flags & self.CRL_SHOW) and not (flags & self.CRL_OPTIONAL):
                self.ui.crl_checkbox.setChecked(True)
                self.ui.crl_checkbox.hide()

        if not self.component_has_cert_info:
            for widget in self.attr_to_widget.itervalues():
                widget['info'].setText(tr('Not available'))
                widget['info'].setEnabled(False)

            for widget in self.attr_to_widget.itervalues():
                widget['view'].setEnabled(False)

    def showUsePKI(self, checked):
        self.ui.stackedWidget.setCurrentIndex(1)
        self.modify_cb()

    def showUploadCerts(self, checked):
        self.ui.stackedWidget.setCurrentIndex(0)
        self.modify_cb()

    def feed(self):
        self.enableCrl()

        pkis = self.nupki.listPKI()

        self.ui.nupki_pki.clear()
        self.ui.nupki_pki.addItems([pki.name for pki in pkis])

        if pkis:
            self.loadCertsFromPKI()

        if self.component_has_cert_info:
            self.cert_info = self.client.call(self.component, 'getCertificatesInfo')
            self.displayCertInfos()

    def loadCertsFromPKI(self):
        self.modify_cb()
        current_pki = self.ui.nupki_pki.currentText()

        if current_pki == "":
            return

        certsRows = self.nupki.getCertList(unicode(current_pki), "server")
        names = [cert.cname for cert in certsRows if cert.status == "valid"]
        self.ui.nupki_cert.clear()
        self.ui.nupki_cert.addItems(names)

    def enableCrl(self):
        enable = self.ui.crl_checkbox.isChecked()
        for widget in self.attr_to_widget['crl'].itervalues():
            widget.setEnabled(enable)
        if not self.component_has_cert_info:
            self.attr_to_widget['crl']['info'].setEnabled(False)
            self.attr_to_widget['crl']['view'].setEnabled(False)

    def browseFile(self, attr):
        selectedfile = QFileDialog.getOpenFileName(None, self.ATTR_TO_LABEL[attr], "", "PEM encoded file (*.pem);;Any (*.*)")
        if selectedfile:
            self.attr_to_widget[attr]['lineedit'].setText(selectedfile)
        self.modify_cb()

    def setConfig(self, _config):
        config = deepcopy(self.DEFAULT_CONF)
        config.update(_config)

        use_nupki = config['use_nupki']
        self.ui.nupki_radio.setChecked(use_nupki)
        self.ui.user_radio.setChecked(not use_nupki)

        pki_index = self.ui.nupki_pki.findText(QString(config['nupki_pki']))
        if pki_index != -1:
            self.ui.nupki_pki.setCurrentIndex(pki_index)

        cert_index = self.ui.nupki_cert.findText(QString(config['nupki_cert']))
        if cert_index != -1:
            self.ui.nupki_cert.setCurrentIndex(cert_index)

        if not use_nupki:
            self.ui.crl_checkbox.setChecked(not config['disable_crl'])
            self.enableCrl()

    def getConfig(self, get_cert_content = True):
        config = deepcopy(self.DEFAULT_CONF)

        config['use_nupki'] = self.ui.nupki_radio.isChecked()
        if config['use_nupki']:
            config['nupki_pki'] = unicode(self.ui.nupki_pki.currentText())
            config['nupki_cert'] = unicode(self.ui.nupki_cert.currentText())
        else:
            for attr, widget in self.attr_to_widget.iteritems():
                if widget['lineedit'].text():
                    if get_cert_content:
                        config[attr] = getEncodedFileContent(unicode(widget['lineedit'].text()))
                    else:
                        config[attr] = unicode(widget['lineedit'].text())
            config['disable_crl'] = not self.ui.crl_checkbox.isChecked()
        return config

    def validate(self):
        conf = self.getConfig(get_cert_content = False)
        if conf['use_nupki']:
            # Assume nupki's generate certificate are OK
            return None

        ssl_conf = SSLConfig()
        ssl_conf.EXPECTED_X509_PURPOSE = self.cert_type

        ssl_conf.setConfig(conf['ca'], conf['cert'], conf['key'], conf['crl'])

        # Enable checks
        ssl_conf.check = bool(conf['ca']) or bool(conf['crl'])
        ssl_conf.send_cert = bool(conf['cert']) or bool(conf['key'])

        return ssl_conf.validate()

    def displayCertInfos(self):
        if not self.component_has_cert_info:
            return

        for attr, widget in self.attr_to_widget.iteritems():
            widget['info'].setText(self.cert_info[attr][0])

    def viewInfo(self, attr):
        if attr != 'crl':
            text = tr('%s currently set:\n%s') % (self.ATTR_TO_LABEL[attr], self.cert_info[attr][0])
        else:
            text = self.cert_info[attr][0]

        msgBox = QMessageBox()
        msgBox.setWindowTitle( self.ATTR_TO_LABEL[attr] )
        msgBox.setText( tr('%s currently set:\n%s') % (self.ATTR_TO_LABEL[attr], self.cert_info[attr][0]))
        msgBox.setDetailedText(self.cert_info[attr][1])
        msgBox.setStandardButtons(QMessageBox.Ok)
        msgBox.exec_()

