#!/usr/bin/python
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from code import InteractiveConsole
from optparse import OptionParser
from sys import exit, exc_info
from threading import Thread, Lock, Event
import sys
from os.path import expanduser
import rlcompleter
import readline
import atexit
from pprint import pprint
from time import time
from logging import basicConfig, DEBUG, ERROR, INFO

from nucentral.common.error import writeError as _writeError
from nucentral.common.tools import readDocumentation
from nucentral.common import tr
from nucentral.client import NuCentralClientBase, NuCentralError, KEEP_ALIVE_SECONDS

# Readline doesn't work correctly with UnicodeStdout
from nucentral.common.unicode_stdout import uninstallUnicodeStdout, getTerminalCharset
uninstallUnicodeStdout()

def writeError(err, title, traceback=None):
    if isinstance(err, NuCentralError):
        message = unicode(err)
    else:
        message = None
    _writeError(err, title, message=message, traceback=traceback)

# unauthenticated user: recheck each 30 seconds if the user is authenticated
UNAUTHENTICATED_DELAY = 30.0

class KeepAlive(Thread):
    def __init__(self, client):
        Thread.__init__(self)
        self.client = client
        self._running = Lock()
        self._stop = Event()
        if self.client.login:
            self.next_call = time() + KEEP_ALIVE_SECONDS
        else:
            self.next_call = time() + UNAUTHENTICATED_DELAY

    def run(self):
        self._running.acquire()
        try:
            while True:
                self._stop.wait(0.250)
                if self._stop.isSet():
                    break
                if time() < self.next_call:
                    continue
                if self.client.login:
                    self.next_call = time() + KEEP_ALIVE_SECONDS
                    try:
                        self.client.call('session', 'keepAlive')
                    except NuCentralError, err:
                        writeError(err, "Keep alive error")
                else:
                    # unauthenticated user: recheck each 30 seconds
                    # if the user is authenticated
                    self.next_call = time() + UNAUTHENTICATED_DELAY
        finally:
            self._running.release()

    def stop(self):
        self._stop.set()
        self._running.acquire()
        self._running.release()

class ClientError(NuCentralError):
    def __init__(self, message):
        NuCentralError.__init__(self, "ClientError", message)

def oneLinerHelp(prefix, doc):
    if not doc:
        return prefix
    if 1 < len(doc):
        return prefix + ': "%s (...)"' % doc[0]
    else:
        return prefix +': "%s"' % doc[0]

class ServiceProxy:
    def __init__(self, component, service):
        self._component = component
        self._service = service
        client = component._client
        try:
            self.__doc__ = client('CORE', 'help', self._component._component, self._service)
        except NuCentralError:
            self.__doc__ = None

    def _help(self):
        return self.__doc__

    def __call__(self, *arguments):
        return self._component(self._service, *arguments)

    def __repr__(self):
        return "<ServiceProxy %r>" % str(self)

    def __str__(self):
        return "%s.%s" % (self._component._component, self._service)

class ComponentProxy(object):
    def __init__(self, client, component):
        self._client = client
        self._component = component
        try:
            self.__doc__ = self._client('CORE', 'help', self._component)
        except NuCentralError:
            self.__doc__ = None

    def _services(self):
        # Only works with Python 2.6+
        return self._client('CORE', 'getServiceList', self._component)

    def __dir__(self):
        # Only works with Python 2.6+
        return self._services()

    def __getattr__(self, service):
        if service.startswith("_"):
            object.__getattr__(self, service)
        if service not in self._services():
            raise ClientError(tr("The component %r has not service %r!") % (self._component, service))
        return ServiceProxy(self, service)

    def _help(self):
        doc = self.__doc__
        if doc:
            doc = list(doc)
        else:
            doc = []
        doc.extend(('', 'Services:'))
        for service in self._services():
            service_doc = self._client('CORE', 'help', self._component, service)
            if service_doc:
                service = service_doc[0]
                if service.endswith(":"):
                    service = service[:-1]
                    del service_doc[0:2]
                else:
                    del service_doc[0]
            text = oneLinerHelp(" - %s" % service, service_doc)
            doc.append(text)
        return doc

    def __call__(self, service, *arguments):
        return self._client(self._component, service, *arguments)

    def __repr__(self):
        return "<ComponentProxy %r>" % self._component

    def __str__(self):
        return self._component

class Client:
    def __init__(self):
        self.options = self.parseOptions()
        self.setupLog()

        # Connect to server
        if self.options.verbose:
            print "Connect to server"

        options = {'client_name': 'nucentral_client'}
        if self.options.http:
            options['protocol'] = 'http'
        if self.options.host:
            options['host'] = self.options.host
        if self.options.port:
            options['port'] = self.options.port
        try:
            self.server = NuCentralClientBase(**options)
        except NuCentralError, err:
            print "Connection error: %s" % err
            exit(1)
        if self.options.verbose:
            print "Connected to: %s" % self.server.url

        if self.options.username and (self.options.password is not None):
            charset = 'utf8'
            username = unicode(self.options.username, charset)
            password = unicode(self.options.password, charset)
            if not self.authenticate(username, password):
                exit(1)

    def setupLog(self):
        if self.options.debug:
            level = DEBUG
        elif self.options.verbose:
            level = INFO
        else:
            level = ERROR
        basicConfig(level=level, format="%(message)s")

    def parseOptions(self):
        parser = OptionParser(usage="%prog [options]")
        parser.add_option("--host", help='Host name or IP',
            type="str", default=None)
        parser.add_option("--http", help="Use HTTP protocol instead of HTTPS",
            action="store_true", default=False)
        parser.add_option("--port", help='Server TCP port',
            type="int", default=None)
        parser.add_option("-u", "--username", help="Username used with password for authentication",
            type="str", default=None)
        parser.add_option("-p", "--password", help="Password used with username for authentication",
            type="str", default=None)
        parser.add_option("-v", "--verbose", help="Verbose mode",
            action="store_true", default=False)
        parser.add_option("--debug", help="Debug mode (don't call any service at startup)",
            action="store_true", default=False)
        options, arguments = parser.parse_args()
        if options.debug:
            options.verbose = True
        if arguments:
            parser.print_help()
            exit(1)
        return options

    def __call__(self, *arguments):
        return self.server.call(*arguments)

    def call(self, *arguments):
        """call('component', 'service', ...): call a NuCentral service"""
        return self(*arguments)

    def authenticate(self, login, password):
        """authenticate('login', 'password'): authentication, update user session (login and groups)"""
        try:
            self.server.authenticate(login, password)
            return True
        except NuCentralError, err:
            writeError(err, "Authentication error")
            return False

    def exit(self):
        """exit(): quit"""
        exit(0)

    def quit(self):
        print("Exit client.")
        self.server.logout()
        print

    def components(self):
        """components(): get the component list"""
        return self('CORE', 'getComponentList')

    def services(self, component):
        """services('component'): get the list of services of the component"""
        return self('CORE', 'getServiceList', component)

    def proxy(self, component):
        """proxy('component'): create a component proxy which can be used like component.service()"""
        if component not in self.components():
            raise ClientError(tr("There is not component %r!") % component)
        return ComponentProxy(self, component)

    def usage(self):
        banner = [
            "NuCentral client 2",
            "",
            "Commands:",
        ]
        for command in self.getCommands():
            for line in readDocumentation(command):
                banner.append("- %s" % line)
        banner.append("- pyhelp(object): Python help for the specified object")
        return banner

    def printUsage(self):
        for line in self.usage():
            print line

    def help(self, *arguments):
        """
        help('component'): usage of a component
        help('component', 'service'): usage of a component
        help(proxy): proxy usage
        """
        if not (1 <= len(arguments) <= 2):
            self.printUsage()
            return
        if isinstance(arguments[0], (ComponentProxy, ServiceProxy)):
            if len(arguments) == 2:
                self.printUsage()
                return
            lines = arguments[0]._help()
        else:
            lines = self('CORE', 'help', *arguments)
        if lines is None:
            return
        for line in lines:
            print line
        print

    def status(self, *arguments):
        """status(): display server status and session informations"""
        print "NuCentral status: %s" % self('CORE', 'getStatus')
        info = []
        user = self('session', 'get')['user']
        if 'login' in user:
            info.append('login=%r' % user['login'])
        if 'groups' in user:
            info.append('groups=%s' % user['groups'])
        info.append('cookie=%r' % self.server.cookie)
        print "Session: %s" % ', '.join(info)

    def getCommands(self):
        return (self.call, self.authenticate,
            self.components, self.services, self.proxy, self.status, self.help, self.exit)

    def createProxies(self, locals):
        proxies = []
        for component in self.components():
            if component in locals:
                # Don't create proxy if a command has the same name
                continue
            proxy = ComponentProxy(self, component)
            locals[component] = proxy
            proxies.append(proxy)
        return proxies

class HistoryConsole(InteractiveConsole):
    def __init__(self, locals=None):
        InteractiveConsole.__init__(self, locals, "<console>")
        filename = expanduser("~/.nucentral_client.history")
        self.completer = rlcompleter.Completer(locals)
        self.charset = getTerminalCharset()
        self.init_history(filename)

    def init_history(self, histfile):
        readline.set_completer(self.completer.complete)
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except IOError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        readline.write_history_file(histfile)

    def interact(self, banner=None):
        old_hook = sys.displayhook
        sys.displayhook = self.displayHook
        try:
            return InteractiveConsole.interact(self, banner)
        finally:
            sys.displayhook = old_hook

    def showtraceback(self):
        info = exc_info()
        errtype, err, tb = info
        writeError(err, "Error", traceback=info)

    def displayHook(self, value):
        if value is None:
            return
        pprint(value)

    def raw_input(self, prompt=""):
        text = raw_input(prompt)
        return unicode(text, self.charset)

def printProxies(proxies):
    print
    print "Proxies:"
    for proxy in proxies:
        print oneLinerHelp("- %s" % proxy, proxy.__doc__)

class Application:
    def __init__(self):
        self.client = Client()
        self.interpreter = self.createInterpreter()
        self.keep_alive = KeepAlive(self.client.server)

    def createInterpreter(self):
        readline.parse_and_bind("tab: complete")
        commands = self.client.getCommands()
        locals = dict((command.__name__, command) for command in commands)
        locals['pyhelp'] = help
        proxies = self.client.createProxies(locals)
        interpreter = HistoryConsole(locals)
        self.client.printUsage()
        printProxies(proxies)
        return interpreter

    def main(self):
        try:
            self.keep_alive.start()
            self.interpreter.interact('')
        finally:
            self.keep_alive.stop()
            self.client.quit()

def main():
    app = Application()
    app.main()

if __name__ == "__main__":
    main()

