#coding: utf-8
"""
$Id$
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from __future__ import with_statement

from base64 import b64decode
from base64 import decode
from datetime import datetime
from os import unlink
from os.path import exists
from os.path import join
from tempfile import TemporaryFile
from time import time
from twisted.internet.defer import inlineCallbacks, returnValue

from nucentral.backend import Component
from nucentral.backend.error import AclError
from nucentral.common.error import exceptionAsUnicode
from nucentral.core.context import Context

from .exporter import Exporter
from .importer import Importer

SERVICE = 'runtimeFiles'

_ARCHIVE_FILE_NAME = ''
_TIMESTAMP_FILE = 'restoration.stamp'

class RestorationManager(Component):
    NAME = "nurestore"
    VERSION = '1'
    API_VERSION = 1
    REQUIRES = ('config',)

    ACLS = {
        'config': frozenset(('loadImportedConf',)),
        'acl': frozenset(('runtimeFilesModified',)),
        'nuauth': frozenset(('runtimeFilesModified',)),
        'nudpi': frozenset(('runtimeFilesModified',)),
        'nupki': frozenset(('runtimeFilesModified',)),
        'nurestore': frozenset(('import',)),
        'versionning': frozenset(('commit_all',)),
        '*': frozenset((SERVICE,)),
    }

    ROLES = {
        'nuconf_read': frozenset(),
        'nurestore_read': frozenset(('export', '@nuconf_read')),
        'nuconf_write': frozenset(),
        'nurestore_write': frozenset((
            '@nurestore_read',
            'restore',
            '@nuconf_write'
            )),
        }

    def init(self, core):
        self.core = core
        Component.init(self, core)
        self.registered_files = set()
        self.deleted_files = set()
        self.archive_saved_file = join(
            core.config.get('CORE', 'vardir'),
            _ARCHIVE_FILE_NAME
            )
        vardir = self.core.config.get('CORE', 'vardir')
        self._timestamp_file = join(vardir, _TIMESTAMP_FILE)
        self._stamp_restore('idle')

    """
    EXPORT AND DUPLICATE
    synchronize: for HA, runtimeFiles is called with ha component context
    export: when exporting configuration, runtimesFiles is called with nurestore component context
    """
    def service_duplicate(self, context, components=None):
        """Same as export service, but caller context is passed to runtimesFiles
        If components is an iterable, configuration for each component indicated
        is exported.
        If components is None, entire configuration is exported.
        """
        return self.collectFiles(context, components)

    def service_export(self, context, components=None):
        """
        If components is an iterable, configuration for each component indicated
        is exported.
        If components is None, entire configuration is exported.

        return :
            - a tuple: (tar, list of components)
            - None (nothing to export)
        """
        nurestore_context = Context.fromComponent(self)
        return self.collectFiles(nurestore_context, components)

    @inlineCallbacks
    def collectFiles(self, context, components):
        """
        makes an Exporter object, adds all runtime files to it, ignoring errors,
        then, as a last callback, makes the tar

        components:
            - None for all
            - list of components

        return :
            - a tuple with :
                - encoded tar file which contains configuration files
                - list of components which
            - None if nothing has been exported
        """
        exporter = Exporter(self)

        components = list(self._iterComponents(components))
        if not components:
            self.debug("export : nothing to export")
            returnValue(None)

        # Algorithm
        #
        # Foreach component:
        #   try:
        #       component.runtimeFiles()
        #   except:
        #       log error
        # return errors

        # TODO if component.runtimeFiles() failed, component must be removed
        # of components

        errors = []
        for component in components:
            try:
                files = yield self.core.callService(context, component, SERVICE)
                exporter.registerFiles(files, component=component)
            except Exception, err:
                self.writeError(err, "Error in %s.%s()" % (component, SERVICE))
                errors.append(exceptionAsUnicode(err))
        content, exported_components = exporter.export_and_encode(errors, components)
        returnValue((content, exported_components))

    def _iterComponents(self, components):
        """
        returns the components that have the service SERVICE implemented
        """
        context = Context.fromComponent(self)

        if components is None:
            components = self.core.getComponentList(context)

        for component_name in components:
            try:
                services = self.core.getServiceList(context, component_name)
            except AclError:
                continue
            except Exception, err:
                self.writeError(err, "Error on getting %s service list" % component_name)
                continue

            if SERVICE in services:
                self.debug("'%s.%s()' found" % (unicode(component_name), unicode(SERVICE)))
                yield component_name

    def service_import(self, context, input):
        """
        decode and import data
        """
        return self._import(input)

    def service_isRestoring(self, context):
        return self.service_state()[0] != 'idle'

    def service_state(self, context):
        return self._state

    def _stamp_restore(self, state):
        self._state = state, unicode(datetime.fromtimestamp(time()))

    @inlineCallbacks
    def service_restore(self, context, input):
        """
        receives your archive, decodes it and asks invividual components to
        take in account new data
        """
        self._stamp_restore('restoring')
        self._import(input)

        # nuface:
        #   nothing to do on the backend side.
        #   The user will have to manually load some nuface rules
        # nulog:
        #   depends on the config module to reload its new config
        # nuconf:
        #   depends on the config module to reload its new config

        comp_context = Context.fromComponent(self)

        yield self.core.callService(
            comp_context,
            'acl',
            'runtimeFilesModified'
            )

        yield self.core.callService(
            comp_context,
            'nupki',
            'runtimeFilesModified'
        )

        yield self.core.callService(
            comp_context,
            'nudpi',
            'runtimeFilesModified'
        )

        # multisite:
        #    ????

        # config:
        #   call config.loadRestored()
        #XXX Keep last
        yield self.core.callService(
            comp_context,
            'config',
            'loadImportedConf'
        )

        self._stamp_restore('restored')
        returnValue("ok")

    def _import(self, input):
        with TemporaryFile(suffix='.tar') as output:
            output.write(b64decode(input))
            output.seek(0)

            vardir = self.core.config.get('CORE', 'vardir')
            with Importer(self, vardir, output) as importer:
                importer.import_all()
            context = Context.fromComponent(self)
            if self.core.hasComponent(context, 'versionning'):
                self.core.callService(
                    context,
                    'versionning',
                    'commit_all',
                    'nurestore overriding files'
                    )

        return "ok"

    def service_importArchive(self, context, input):
        with open(self.archive_saved_file, 'w') as output:
            decode(input, output)
    def service_hasArchive(self, context):
        return exists(self.archive_saved_file)

    def service_delArchive(self, context):
        try:
            unlink(self.archive_saved_file)
        except OSError:
            pass
        return not exists(self.archive_saved_file)

    def service_listAvailComponents(self, context, components):
        """return list of component from components which can export content"""
        return list(self._iterComponents(components))
