#coding: utf-8
"""
$Id$
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from __future__ import with_statement

from os import fdopen
from os import listdir
from os import unlink
from os.path import exists
from os.path import isabs
from os.path import isdir
from os.path import join
from os.path import realpath
from shutil import rmtree
from tempfile import mkdtemp
from tempfile import mkstemp
from base64 import b64encode
import errno
import tarfile

from nucentral.common.download import encodeFileContent
from nucentral.common.logger import LoggerChild

from .constants import ARCHIVE_PREFIX
from .constants import DELETED_FILENAME
from .constants import MANIFEST_FILENAME
from .constants import COMPRESS
from .util import check_file
from .errors import ForbiddenFilename

def archive(tar, file_name):
    """
    check that the file to be archived really exists
    """
    #Let names begin with a /
    arcname = ARCHIVE_PREFIX + file_name

    #Here we can get an IOError, and let it raise because we just
    #tested 'exists', so we assume
    #err.errno != errno.ENOENT
    if not exists(file_name):
        return

    info = tar.gettarinfo(file_name, arcname=arcname)
    #alter attributes
    info.uname = ''
    info.gname = ''

    if not isdir(file_name):
        with open(file_name, 'r') as filedesc:
            tar.addfile(info, filedesc)
        return

    tar.addfile(info)
    #recurse
    print "going to recurse in %s" % file_name
    for contained_file_name in listdir(file_name):
        archive(tar, join(file_name, contained_file_name))

class Exporter(LoggerChild):
    def __init__(self, logger):
        """
        synopsys:
        multiple times, registerFiles(
            deleted=iterable of paths,
            added=iterable of paths
            )
        then export() -> filename
        or export_and_encode() -> content
        """
        LoggerChild.__init__(self, logger)

        self.to_archive = set()
        self.to_delete = set()
        self.tar_file = None
        self.deleted_filename = None
        self.manifest_filename = None

    def export_and_encode(self, result, components):
        """
        create tar file and return encoded content
        """
        try:
            self.export()
            self.debug("-> tar_file %s" % self.tar_file)
            with open(self.tar_file, 'rb') as fd:
                content = fd.read()
            return encodeFileContent(content), components
        finally:
            if exists(self.tar_file):
                unlink(self.tar_file)

    def registerFiles(self, files, component):
        self.debug('registering files for %s : %s' % (component, files))

        added = files.get('added')
        deleted = files.get('deleted')

        if added is not None:
            for entry in (added):
                self.debug("\tadded: '%s'" % unicode(entry))
            self.add_filenames(added)

        if deleted is not None:
            for file in deleted:
                self.debug("\tdeleted: '%s'" % unicode(file))
            self.delete_filenames(deleted)

    def delete_filenames(self, deleted):
#        if not all(isabs(name) for name in (deleted)):
#            raise InvalidUsage("Only absolute paths supported")
        for filename in deleted:
            if isabs(filename):
                item = realpath(filename)
                self.to_delete.add(filename)
                # self.debug if filename doesnt exist locally
            else:
                self.error("Ignore deleting path '%s (not an absolute path)" % filename)

    def add_filenames(self, added):
        """
        exceptions are never sent to client ...
        """
#        for entry in (added):
#                raise InvalidUsage("Entry is not a tuple (absolute path, simple type) %s" % entry)
#        if not all(isabs(file_name) for file_name, file_type in added):
#            raise InvalidUsage("Only absolute paths supported")
        for entry in added:
            if len(entry) == 2:
                file_name, file_type = entry
                if isabs(file_name):
                    file_name = realpath(file_name)
                    try:
                        check_file(file_name) # FIXME exception useless here
                    except ForbiddenFilename:
                        self.critical("Filename '%s' not whitelisted ; cannot export it." % file_name)
                        continue
                    if exists(file_name):
                        self.debug("Add to archive : '%s (%s)'" % (file_name, file_type))
                        self.to_archive.add((file_name, file_type))
                    else:
                        self.debug("Can not add to archive inexistant file : '%s (%s)'" % (file_name, file_type))
                else:
                    self.error("Ignore adding path '%s (%s)' (not an absolute path)" % (file_name, file_type))
            else:
                self.error("Entry is not a tuple (absolute path, simple type) %s" % entry)

    def export(self):
        self.temp_dir = mkdtemp()
        try:
            self.deleted_filename = join(self.temp_dir, DELETED_FILENAME)
            self.make_deleted()

            self.manifest_filename = join(self.temp_dir, MANIFEST_FILENAME)
            self.make_manifest()

            self.tar_all()
        finally:
            self.clean()

    def make_deleted(self):
        with open(self.deleted_filename, 'wb') as fd:
            for name in self.to_delete:
                fd.write(name + '\n')

    def make_manifest(self):
        # why encode filename ?
        # --> because base64 makes the file easier to parse: no more weird spaces aso.
        #FIXME: catch the real update_level
        update_level = "0"
        with open(self.manifest_filename, 'wb') as fd:
            fd.write("::update_level:: %s\n" % update_level)
            for file_name, file_type in self.to_archive:
                file_name = b64encode(file_name)
                fd.write("%s %s\n" % (file_name, file_type))

    def tar_all(self):
        suffix = '.tar'
        mode = 'w'
        if COMPRESS:
            suffix += '.%s' % COMPRESS
            mode += ':%s' % COMPRESS
        handle, self.tar_file = mkstemp(suffix=suffix, text=False)
        file = fdopen(handle, 'wb')
        tarfile.TarFile.debug = 3
        tar = tarfile.open(name=None, mode=mode, fileobj=file)
        for file_name, file_type in self.to_archive:
            archive(tar, file_name)

        if self.deleted_filename:
            tar.add(self.deleted_filename, arcname=DELETED_FILENAME)
        if self.manifest_filename:
            tar.add(self.manifest_filename, arcname=MANIFEST_FILENAME)

        tar.close()
        file.close()

    def clean(self):
        try:
            rmtree(self.temp_dir)
        except OSError, err:
            if err.errno != errno.ENOENT:
                raise
        self.deleted_filename = None
        self.manifest_filename = None

