#coding: utf-8
"""
$Id$
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from __future__ import with_statement

from datetime import datetime
from os.path import join as path_join
from os import unlink
from os.path import isdir
from os.path import exists
from os.path import realpath
from shutil import move
from shutil import rmtree
from tempfile import mkdtemp
from time import time
import errno
import tarfile

from nucentral.common.logger import LoggerChild

from .constants import ARCHIVE_PREFIX
from .constants import DELETED_FILENAME
from .constants import MANIFEST_FILENAME
from .util import check_file

class Importer(LoggerChild):
    def __init__(self, logger, vardir, tar_fileobj):
        LoggerChild.__init__(self, logger)
        self.vardir = vardir

        self.tar = tarfile.open(fileobj = tar_fileobj, mode = 'r')

        # DELETED_FILENAME and MANIFEST_FILENAME are extracted here
        self.work_dir = mkdtemp()

        self.deleted_filename = path_join(self.work_dir, DELETED_FILENAME)
        self.manifest_filename = path_join(self.work_dir, MANIFEST_FILENAME) # unused

        for member in (DELETED_FILENAME, MANIFEST_FILENAME):
            self.tar.extract(member, path=self.work_dir)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        """
        clean
        """
        self.tar.close()
        rmtree(self.work_dir, False, self.displayError)

    def displayError(self, function, path, excinfo):
        self.error("Error while deleting '%s' : %s (%s)" %
            (unicode(path), unicode(excinfo), unicode(function)))

    def import_all(self):
        """
        list files to delete
        list files to add
        """

        #tar all deleted files
        self.tar_deleted()
        #unlink them
        self.del_deleted()

        #untar files in ARCHIVE_PREFIX -> /
        self.debug("Extracting files from archive...")
        for tar_entry in self.archive_files():
            self.tar.extract(tar_entry.name)

    def archive_files(self):
        """
        For every file in the archive, if its paths begins with ARCHIVE_PREFIX
        then it is an archived file, I remove the ARCHIVE_PREFIX to the name
        and yield the resulting member.

        Non matching files are supposedly only DELETED_FILENAME and MANIFEST_FILENAME
        """
        for member in self.tar:
            if member.name.startswith(ARCHIVE_PREFIX):
                member.name = member.name.replace(ARCHIVE_PREFIX, "", 1)
                check_file(member.name)
                yield member
            else:
                assert member.name in (DELETED_FILENAME, MANIFEST_FILENAME), "unknow filename ... '%s'" % member.name

    def tar_deleted(self):
        """
        keep only the last in /var/lib/nucentral/conf_backup.tar.bz2
        """
        timestamp = unicode(datetime.fromtimestamp(time()))
        mode = 'w:bz2'
        suffix = '.tar.bz2'

        tar_name = "conf_backup-" + timestamp + suffix
        tar_name = path_join(self.vardir, tar_name)
        conf_last_backup = path_join(self.vardir, 'conf_last_backup' + suffix)
        tar = tarfile.open(tar_name, mode)

        try:
            with open(self.deleted_filename, 'rU') as input:
                self.debug(
                    'Backing up files that will be deleted in %s ' % conf_last_backup
                    )
                deleted = []
                for line in input:
                    line = line.rstrip(input.newlines)
                    line = realpath(line)
                    check_file(line)
                    if exists(line):
                        deleted.append(line)

                deleted.sort()
                for filename in deleted:
                    self.debug(" - %s" % filename)

                for filename in deleted:
                    try:
                        #all names guaranted to begin with a /
                        tar.add(filename)
                    except IOError, err:
                        if err.errno != errno.ENOENT:
                            raise
        finally:
            tar.close()

        move(tar_name, conf_last_backup)

    def del_deleted(self):
        with open(self.deleted_filename, 'rU') as input:
            for line in input:
                line = line.rstrip(input.newlines)
                if exists(line):
                    if isdir(line):
                        self.debug('deleting %s' % unicode(line))
                        rmtree(line)
                    else:
                        self.debug('deleting %s' % unicode(line))
                        unlink(line)

