#!/usr/bin/python
import sys
import re
from optparse import OptionParser

# start/exit nucentral
START_REGEX_STR = 'Starting (NuCentral|EdenWall server)'
EXIT_REGEX_STR = 'Exit$'
START_REGEX = re.compile(START_REGEX_STR)
EXIT_REGEX = re.compile(EXIT_REGEX_STR)
ERROR_REGEX_STR = 'Error|ERROR|Traceback'
ERROR_REGEX = re.compile(ERROR_REGEX_STR)
CONFIG_REGEX_STR = '\[config\] (Apply: |Apply error |New configuration succesfully|Rollback$|Rollback done)'
CONFIG_REGEX = re.compile(CONFIG_REGEX_STR)
USER_REGEX_STR = 'connected from|connection closed$'
USER_REGEX = re.compile(USER_REGEX_STR)
HA_REGEX_STR = 'callRemote'
UNHANDLED_REGEX_STR = 'Unhandled error'
UNHANDLED_REGEX = re.compile(UNHANDLED_REGEX_STR)

_base = '\x1b['

DEFAULT_STYLE = '\x1b[0m'
COLOR_RED = '\x1b[1;31m'
COLOR_BLUE = '\x1b[1;34m'
COLOR_GREEN = '\x1b[1;32m'
COLOR_YELLOW = '\x1b[1;33m'
COLOR_GRAY = '\x1b[37m'
COLOR_PURPLE = '\x1b[35m'
COLOR_YELLOW_ON_RED = '\x1b[1;33;41m'

def createColorFormat(color):
    return (color + '%s' + DEFAULT_STYLE).__mod__

NOOP = str
FORMAT_ERROR = createColorFormat(COLOR_PURPLE)
FORMAT_CONFIG = NOOP # createColorFormat(COLOR_BLUE)
FORMAT_USER = createColorFormat(COLOR_GRAY)
FORMAT_MATCH = createColorFormat(COLOR_YELLOW)
FORMAT_START = createColorFormat('\x1b[1;36;44m')
formatUnhandled = createColorFormat(COLOR_YELLOW_ON_RED)

class AnalyzeLog:
    def __init__(self):
        self.parse_options()
        regex_str = [START_REGEX_STR, EXIT_REGEX_STR, UNHANDLED_REGEX_STR]
        def formatStart(line):
            if self.reformat:
                line = line[:50]
            line = FORMAT_START(line)
            return '\n'+line
        self.formats = [
            (START_REGEX, formatStart),
            (UNHANDLED_REGEX, formatUnhandled),
        ]
        if self.options.error:
            regex_str.append(ERROR_REGEX_STR)
            self.formats.append((ERROR_REGEX, FORMAT_ERROR))
        if self.options.config:
            regex_str.append(CONFIG_REGEX_STR)
            self.formats.append((CONFIG_REGEX, FORMAT_CONFIG))
        if self.options.user:
            regex_str.append(USER_REGEX_STR)
            def formatUser(line):
                if 'connected from' in line:
                    line = line.split('from')[0]
                return FORMAT_USER(line)
            format = formatUser if self.reformat else FORMAT_USER
            self.formats.append((USER_REGEX, format))
        if self.options.ha:
            regex_str.append(HA_REGEX_STR)
        match = self.options.match
        if match:
            regex_str.extend(match)
            self.formats.insert(0, (re.compile('|'.join(match)), FORMAT_MATCH))
        self.regex = re.compile('|'.join(regex_str))
        if self.options.ignore:
            self.ignore_regex = re.compile('|'.join(self.options.ignore))
        else:
            self.ignore_regex = None

    def parse_options(self):
        parser = OptionParser(usage="%prog nucentral.log")
        parser.add_option("-c", "--config",
            help="Hide config",
            action="store_false", default=True)
        parser.add_option("-e", "--error",
            help="Hide errors",
            action="store_false", default=True)
        parser.add_option("-u", "--user",
            help="Hide users",
            action="store_false", default=True)
        parser.add_option("--ha",
            help="Show HA",
            action="store_true", default=False)
        parser.add_option("--match",
            help="Match the specified regex",
            type="str", action="append")
        parser.add_option("--ignore",
            help="Ignore lines matching the specified regex",
            type="str", action="append")
        parser.add_option("--raw",
            help="Don't rewrite logs",
            action="store_true", default=False)
        parser.add_option("-n", "--line-number",
            help="Display line number",
            action="store_true", default=False)

        self.options, arguments = parser.parse_args()
        if len(arguments) != 1:
            parser.print_help()
            exit(1)
        self.filename = arguments[0]
        self.reformat = not self.options.raw

    def processLine(self, line, number):
        if self.ignore_regex and self.ignore_regex.search(line):
            return
        if not self.regex.search(line):
            return
        line = line.rstrip()
        for regex, format_func in self.formats:
            if regex.search(line):
                line = format_func(line)
                break
        if self.options.line_number:
            for subline in line.splitlines():
                if subline:
                    print "%s:%s" % (number, subline)
                else:
                    print
        else:
            print line


    def main(self):
        line_number = 0
        for line in open(self.filename):
            line_number += 1
            self.processLine(line, line_number)

if __name__ == "__main__":
    AnalyzeLog().main()

