#!/usr/bin/env python
from __future__ import with_statement

from sys import argv, exit
from os import mkdir, remove, stat
from os.path import abspath, exists, expanduser, isfile, join

CHOPDIR = expanduser("~/.logchop")
USAGE = "Usage: logchop.py [start|stop|read|pop|reset|help|demo] LOGFILE"

class LogChop:
    """
        Logchop helps you to retrieve only the last lines of logs, kind of the
        same way the GNU command tail -f would. To do so, it stores marks
        (last lines) in a special file (chopfile).
        By default, the chopfile is written in ~/.logchop.

        Basic behaviour :
        When you start logchop on a logfile, it will store the last line of that
        file as a first mark. You can then do operations that will generate
        lines of log. Once you're finished, calling read() will return all the
        written after the first mark.
    """

    logfile = ""
    chopfile = ""
    log_lines = ""
    chop_lines = ""

    def __init__(self, filename, chopdir=CHOPDIR):
        """
            Builds a logchop object, with the filename of the logfile. You can
            also change the directory the chopfile will be stored in.
        """
        self.logfile = abspath(filename)

        if not exists(chopdir):
            mkdir(chopdir)

        relchopfile = "_".join(self.logfile.split("/"))
        self.chopfile = join(chopdir, relchopfile)

        if not isfile(self.logfile):
            raise Exception("There is no such file.")

    def __del__(self):
        """
            Called whenever a LogChop object is about to be destroyed.
            That way when can be sure no chopfile will be left behind.
        """
        self.stop()

    def _read_bigfile(self, filename, offset):
        """
            Conveniance method to read a file
        """
        with open(filename) as handle:
            handle.seek(offset)
            for line in handle:
                yield line

    def _read_file(self, filename):
        """
            Conveniance method to read a file
        """
        with open(filename) as handle:
            lines = handle.readlines()
        return lines

    def _append_mark_chopfile(self):
        """
            Store line in the file that logchop uses to keep marks on the
            logfile.
        """
        offset = stat(self.logfile).st_size
        with open(self.chopfile, "a") as chop_handle:
            chop_handle.write(str(offset)+"\n")

    def _first_mark(self):
        """
            Read the chopfile to tell what the first mark is.
        """
        chop_lines = self._read_file(self.chopfile)
        # Use the first line of the chopfile as the first mark
        return long(chop_lines[0])

    def _last_mark(self):
        """
            Read the chopfile to tell what the last mark is.
        """
        chop_lines = self._read_file(self.chopfile)
        # Use the last line of the chopfile as the last mark
        return long(chop_lines[-1:][0])

    def _read_from(self, offset):
        """
            Returns all line written in the logfile after from_line.
        """
        assert isinstance(offset, long), \
            "Got %s instead of %s" % (type(offset), type(long()))

        # Not too sure about this, there may be a race condition, since
        # there is some time between the mark and the actual reading of
        # the file. I'd rather put this line BEFORE than AFTER reading
        # the logfile, since that way we can be sure we don't have too
        # much info, but it doesn't seem right.
        self._append_mark_chopfile()

        log_lines = self._read_bigfile(self.logfile, offset)
        results = []
        for line in log_lines:
            results.append(line)
        return results

    def start(self):
        """
            Starts logchop by storing the last line of the logfile, also known
            as the first marker. That line will then be used to read all new
            lines written to the logfile (see read() and pop()).
        """
        if exists(self.chopfile):
            raise Exception("!!Warning!! Logchop already has marks on this file.\n" +
                            "You should use reset if you want to reset marks.")

        # rather than writting the last line in the chopfile, write the position
        # of the file
        self._append_mark_chopfile()

        return True

    def reset(self):
        """
            Reset logchop with stop(), then start().
        """
        self.stop()
        return self.start()

    def stop(self):
        """
            Logchop will stop storing lines from the logfile. The marks file
            will be removed.
        """
        try:
            remove(self.chopfile)
        except:
            return False
        return True

    def all(self):
        """
            Returns the lines written since logchop was started on the logfile,
            with start().
            The last line will be stored in order for pop() to be able to print
            all lines written since the last use of all().
        """
        return self._read_from(self._first_mark())

    def pop(self):
        """
            Returns the lines written since logchop was last used to print
            loglines, either with all() or with pop().
            As with all(), the last line will be stored in order for pop() to be
            able to print all lines written since the last use of pop().
        """
        return self._read_from(self._last_mark())

def com_demo():
    from threading import Thread
    from time import time, sleep
    global stop_thread
    stop_thread = False
    def demo_log_writer():
        global stop_thread
        while not stop_thread:
            with open("./test.log", "a") as handle:
                handle.write(str(time())+"\n")
            sleep(2)
        print "Stopping the thread."

    with open("./test.log", "a") as handle:
        handle.write(str(time())+"\n")

    print "## Launching a thread that will write the time every two seconds in ./test.log."
    print "## Don't take my word for it, open a new terminal and try 'tail -f test.log' :)."
    thread_handle = Thread(target=demo_log_writer)
    thread_handle.start()

    logchopper = LogChop("./test.log")
    print "# Let's start logchop on the file test.log."
    raw_input("$ ./logchop.py start test.log")
    com_start(logchopper)
    print ""

    print "# Let's now wait for a while, so that test.log gets written, then hit Return."
    raw_input("$ ./logchop.py read test.log")
    com_read_pop(logchopper, "read")
    print ""

    print "# Ok, we saw all that was written since 'logchop start'. What about all that was written since 'logchop read' ?"
    raw_input("$ ./logchop.py pop test.log")
    com_read_pop(logchopper, "pop")
    print ""

    print "# And now, can we see all that was written since 'logchop pop' ?"
    raw_input("$ ./logchop.py pop test.log")
    com_read_pop(logchopper, "pop")
    print ""

    print "# Wait, is it still possible to see all that was written since 'logchop start' ?"
    raw_input("$ ./logchop.py read test.log")
    com_read_pop(logchopper, "read")
    print ""

    print "# Ok, let's start over, on the same file."
    raw_input("$ ./logchop.py reset test.log")
    com_reset(logchopper)
    print ""

    print "# Let's now wait for a while, so that test.log gets written, then hit Return."
    raw_input("$ ./logchop.py read test.log")
    com_read_pop(logchopper, "read")
    print ""

    print "# Got all you need ? Let's stop logchop."
    raw_input("$ ./logchop.py stop test.log")
    com_stop(logchopper)
    print ""

    stop_thread = True
    remove("./test.log")
    thread_handle.join()

def com_start(logchopper):
    try:
        logchopper.start()
        print "Started."
    except Exception, e:
        print e

def com_reset(logchopper):
    try:
        logchopper.reset()
        print "Reset."
    except Exception, e:
        print e

def com_stop(logchopper):
    try:
        logchopper.stop()
        print "Bye."
    except Exception, e:
        print e

def com_read_pop(logchopper, command):
    if command == "read":
        results = logchopper.all()
    else:
        results = logchopper.pop()
    if results:
        for line in results:
            print line.rstrip()
    else:
        print "No new lines."

if __name__ == "__main__":
    arglist = argv

    if len(arglist) == 2 and arglist[1] == "demo":
        com_demo()
        exit(0)

    if len(arglist) != 3:
        print USAGE
        exit(1)

    command = arglist[1]
    logfile = arglist[2]

    logchopper = LogChop(logfile)

    if command == "start":
        com_start(logchopper)

    elif command == "reset":
        com_reset(logchopper)

    elif command == "stop":
        com_stop(logchopper)

    elif command in ("read", "pop"):
        com_read_pop(logchopper, command)

    else:
        print USAGE
        exit(1)
    exit(0)
