#!/usr/bin/python

"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


from getopt import getopt
from getopt import GetoptError
from sys import argv
from sys import stdout

from nucentral.client.error import NuCentralError
from nucentral.client.nucentral_client_base import NuCentralClientBase
from nuconf.common.hostnamecfg import deserialize as deserialize_host
from nuconf.common.netcfg_rw import deserialize as deserialize_network
from nuconf.common.net_exceptions import NoMatch
from nuconf.common.net_interfaces import Ethernet

_DEFAULTS = {
    "user": "admin",
    "password": "admin"
}

def _usage():
    print """ Usage:

-h --help \t this message

****************************** LOGIN **********************************
-u --user=\tusername, defaults to "admin"
-p --password=\tpassword, defaults to "admin"

************************** DANGEROUS OPTION ***************************
-c --complete-up-to=\tcomplete ethernets up to number. this is a
                 \tdangerous option
"""

def _common_start(user, password):
    client = NuCentralClientBase(protocol="https")
    client.authenticate(user, password)

    try:
        write_role = client.call('nuconf', 'takeWriteRole')
        if not write_role:
            print "Could not take write role, another session has it."
            exit(1)
    except NuCentralError, err:
        print "There is no session lock management on this appliance. (harmless if no one else saves / applies anything meanwhile)"

    serialized_host = client.call('hostname', 'getHostnameConfig')
    hostcfg = deserialize_host(serialized_host)
    hostname = hostcfg.hostname

    serialized_network = client.call('network', 'getNetconfig')
    netcfg = deserialize_network(serialized_network)
    return client, serialized_network, netcfg, hostname

def _print_ethernets(netcfg):
    print "*"*80
    print "Known ethernet interfaces on this system (%s):" % hostname
    ethernets = list(netcfg.iterEthernets())
    ethernets.sort()
    ethernet_index = 0
    for ethernet in ethernets:
        print "[%s] %s\t%s" % (ethernet_index, ethernet.hard_label, "aka \"%s\"" % ethernet.fullName() if ethernet.user_label != ethernet.hard_label else "")
        ethernet_index += 1
    print "That is %s ethernet interfaces." % len(ethernets)

def _print_commands():
    print """
Available commands:

help \t\t\t|\tthis message
list \t\t\t|\tlist ethernet interfaces
quit \t\t\t|\tquits
apply \t\t\t|\tapply the saved configuration
mketh <system_name> \t|\tfor instance 'mketh eth8'
valid \t\t\t|\tvalidates configuration
send \t\t\t|\tvalidates configuration and sends it
restart \t\t|\trestarts this wizard

"""

def _validate(netcfg):
    ok, msg = netcfg.isValidWithMsg()
    if not ok:
        print "NOT VALID"
    print msg
    return ok, msg

def _apply(client):
    print "Applying saved configuration ...",
    stdout.flush()
    client.call('config', 'apply')
    print "Ok"

def _send(client, netcfg, message):
    ok, msg = _validate(netcfg)
    if not ok:
        print "Not sending invalid config"
        return
    serialized = netcfg.serialize()
    print "Saving changes to configuration ...",
    stdout.flush()
    client.call('network', 'setNetconfig', serialized, message)
    print "Ok"

def _add_ethernet(netcfg, system_name):
    try:
        double = netcfg.getInterfaceBySystemName(system_name)
    except NoMatch:
        pass
    else:
        print "Ethernet interface with name \"%s\" already exists in configuration, I will not create another one." % system_name
        return False
    print "- Creating ethernet interface: %s...""" % system_name,
    ethernet = Ethernet(system_name=system_name)
    print " ok"
    print "- Adding %s to config..." % system_name,
    netcfg.ethernets.add(ethernet)
    print " ok"

    return True

def _read_command():
    try:
        command = raw_input().strip()
    except EOFError:
        print ""
        exit(0)
    except KeyboardInterrupt:
        print ""
        command = ""
    return command

def _sanity_check(netcfg):
    ethernets = list(netcfg.iterEthernets())
    ethernets.sort()
    for index, ethernet in enumerate(ethernets):
        expected = "eth%s" % index
        if not ethernet.hard_label == expected:
            print "Sanity check found that ethernet #%s was called %" % (expected, ethernet.hard_label)
            return False
    return True

try:
    optlist, args = getopt(argv[1:], "hu:p:c:", ("help", "user=", "password=", "complete-up-to="))
except GetoptError, err:
    print unicode(err)
    _usage()
    exit(1)

user = _DEFAULTS["user"]
password = _DEFAULTS["password"]
auto = 0

for option, value in optlist:
    if option in ("--user", "-u"):
        user = value
    elif option in ("--password", "-p"):
        password = value
    elif option in ("--help", "-h"):
        _usage()
        exit(0)
    elif option in ("--complete-up-to", "-c"):
        try:
            auto = int(value.strip())
            if auto < 1:
                print "Option %s - Expected an int >= 1, got \"%s\"" % (option, value)
                _usage()
                exit(1)
        except:
            print "Option %s - Expected an int >= 1, got \"%s\"" % (option, value)
            _usage()
            exit(1)
    else:
        assert False, "unhandled option"

client, serialized_network, netcfg, hostname = _common_start(user, password)
added = []


if auto > 0 and not _sanity_check(netcfg):
    print "Interfaces hard labels unexpected, cannot proceed automatically, don't use -c or --complete-up-to"
    exit(1)

if auto > 0:
    interfaces_nb = len(list(netcfg.iterEthernets()))

    missing = auto - interfaces_nb
    if missing < 1:
        print "You want to complete ethernets number up to %s but there are already %s configured ethernets" % (auto, interfaces_nb)
        exit(0)

    #next_index: if we have 4 interfaces eth0, eth1, eth2, eth3, the next one is eth4
    next_index = interfaces_nb
    for i in xrange(missing):
        system_name = "eth%s" % next_index
        _add_ethernet(netcfg, system_name)
        next_index += 1
        added.append(system_name)
    print "Added ethernets: %s." % ", ".join(added)
    message = "script \"%s\" added ethernets: %s." % (argv[0], ", ".join(added))
    _send(client, netcfg, message)
    _apply(client)
    exit(0)

_print_ethernets(netcfg)
_print_commands()
quit = False
while not quit:
    print ">>> ",
    command = _read_command()
    if command == "quit":
        quit = True
        continue
    if command == "restart":
        netcfg = deserialize_network(serialized_network)
        _print_ethernets(netcfg)
        continue
    if command == "help":
        _print_commands()
        continue
    if command == "list":
        _print_ethernets(netcfg)
        continue
    if command == "valid":
        _validate(netcfg)
        continue
    if command == "apply":
        _apply(client)
        continue
    if command == "send":
        if not added:
            print "No ethernet was added, sending anyway"
        else:
            print "Added ethernets: %s." % ", ".join(added)
        message = ""
        print "commit message (empty to cancel): ",
        message = _read_command()
        if message:
            _send(client, netcfg, message)
        continue
    command = command.split()
    if len(command) == 0:
        continue
    if command[0] != "mketh":
        print "unknown command: \"%s\"; type \"help\" for a list of available commands" % command[0]
        continue
    if len(command) > 2:
        print "garbage after interface name? (read %s)" % command[2:]
        continue
    system_name = command[1]
    if _add_ethernet(netcfg, system_name):
        added.append(system_name)
        print "As a supplementary check, we ask you to please ensure that \"%s\" is a valid system interface name, if not, restart" % system_name



