#!/usr/bin/python
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""


import sys, os
from time import sleep
from base64 import b64encode
from nucentral.client.nucentral_client_base import NuCentralClientBase
from nucentral.client.error import NuCentralError, SessionError

REPOSITORY_URL = 'http://tetram.inl.fr/~ft/media/ew4/update/'
UPGRADES_SORTED = [ '4.0.0', '4.0.1', '4.0.12', '4.0.12p2', '4.0.13', '4.0.14', 'dev', 'trunk' ]
UPGRADES_BY_VERSION = {
    '4.0.0' : [],
    '4.0.1' : [10, 11, 12, 15],
    '4.0.12' : [16, 17, 18, 19],
    '4.0.12p2' : [22, 23],
    '4.0.13' : [25, 26, 27, 28, 30, 31, 32, 33, 35, 36],
    '4.0.14' : [37, 39, 40, 41, 42, 43],
    'dev' : [666,],
    'trunk' : [44, 45, 46],
}

class UpgradeClient():
    def __init__(self, host, version):
        self.host = host
        self.version = version
        self.client = NuCentralClientBase(host=self.host)

    def checkVersion(self):
        if self.version not in UPGRADES_SORTED:
            raise Exception('Invalid version \'%s\', please choose one of : %s' % (self.version, ', '.join(UPGRADES_SORTED)))

    def downloadUpgrade(self, no):
        upgrade_name = 'edenwall4_upgrade_%i.tar' % no
        url = REPOSITORY_URL + upgrade_name

        if os.path.exists(upgrade_name):
            print "Upgrade no %i already downloaded" % no
            return upgrade_name

        if os.system('wget ' + url)  != 0 \
        or not os.path.exists(upgrade_name):
            raise Exception('Failed to download upgrade %i from %s' % (no, url))

        return upgrade_name

    def applyUpgrade(self, no, filename):
        # Check the update is not yet applied
        history = self.client.call('update', 'history')
        if no in [upgrade[1] for upgrade in history]:
            print "Update", no, "already aplied"
            return

        print "Applying upgrade", filename
        fd = open(filename, 'rb')
        buf = fd.read()

        try:
            self.client.call('update', 'sendUpgradeArchive', filename, b64encode(buf))
        except NuCentralError, e:
            if e.type == 'NuConfError':
                print "Got error : %s\nContinuing" % str(e)
                raise
        self.client.call('update', 'applyAll')

        # Wait for the upgrade to be applied:
        applied = False
        while not applied:
            sleep(2)
            history = self.client.call('update', 'history')
            for upgrade in history:
                applied = (no == upgrade[1])
                if applied:
                   assert upgrade[5] == 1, "Not successful status for upgrade '%s': '%s'" % (no, upgrade[5])

        print "Upgrade done"

        need_restart = self.client.call('update', 'needRestart')
        print "'%s' applied, restart: ('%s')" % (no, need_restart)
        if not need_restart:
            print "restart: nothing to do"
        elif need_restart == 'nucentral':
            self.restartNucentral()
        else:
            raise NotImplemented('restart system not implemented')

    def restartNucentral(self):
        print "Restarting nucentral"
        self.client.call('tools', 'restartNucentral')

        # Wait for nucentral to be ready
        while True:
            print "Waiting for nucentral..."
            try:
                self.client.call('CORE', 'getStatus')
            except SessionError:
                self.authenticate()

            except NuCentralError, e:
                if e.type != 'protocol':
                    print str(e)
                    print "Retrying"
                    sleep(1)
                    continue
                raise
            break

    def applyUpgrades(self):
        try:
            self.checkVersion()

            # Connect
            self.authenticate()

            # Apply upgrade one by one, starting at version 4.0.0 until asked version
            version_index = UPGRADES_SORTED.index(self.version)


            for edw_version in UPGRADES_SORTED[:version_index + 1]:
                print "Upgrading to version", edw_version
                for upgrade_no in UPGRADES_BY_VERSION[edw_version]:
                    filename = self.downloadUpgrade(upgrade_no)
                    self.applyUpgrade(upgrade_no, filename)
        except Exception, err:
            print "Upgrade failed:", err
            return 1

        print "Now at version", self.version
        return 0

    def authenticate(self):
        self.client.authenticate('admin', 'admin')
        try:
            self.client.call('nuconf', 'takeWriteRole')
        except Exception, err:
            print "Could not take nuconf write role:", err


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "Usage: ./send_upgrades.py host edenwall-version\n"
        exit(1)

    upgrade_cli = UpgradeClient(sys.argv[1], sys.argv[2])
    sys.exit(upgrade_cli.applyUpgrades())
