# Open as many TCP connections as possible: send TCP SYN and immediatly close
# the connection (TCP RST). Use multiple threads to send more connections.
# It doesn't keep more than NTHREADS connections open at the same time.
#
# Use multiple hosts or ports, or your connection tracking will be quickly full
# and you will get errors because there is no more source port available.
#
# On Linux, you can also configure the following kernel option to avoid this
# problem:
#    /proc/sys/net/ipv4/netfilter/ip_conntrack_tcp_timeout_time_wait

import socket, time, datetime, threading, random, sys
from optparse import OptionParser

DEFAULT_HOST = '172.21.0.10'
DEFAULT_PORT = 80
DEFAULT_NTHREADS = 1

class Connect(threading.Thread):
    def __init__(self, manager):
        threading.Thread.__init__(self)
        self.stop = False
        self.port_error = 0
        self.c = 0
        self._counter = 0
        self.total = 0
        self.hosts = manager.options.hosts
        self.ports = manager.ports
        if manager.options.rate:
            self.rate = float(manager.options.rate) / manager.options.nthreads
        else:
            self.rate = None

    def connect(self):
        try:
            host = random.choice(self.hosts)
            port = random.choice(self.ports)
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                s.connect((host, port))
            except socket.error, err:
                if err.errno == 10048:
                    # Only one usage of each socket address
                    # (protocol/network address/port) is normally permitted
                    self.port_error += 1
                    s.close()
                    return True
                elif err.errno == 10060:
                    # connection attempt failed because the connected
                    # party did not properly respond after a period of time,
                    # or established connection failed because connected
                    # host has failed to respond
                    sys.stdout.write("-timeout-\n")
                    s.close()
                    return True
                else:
                    raise
            s.close()
        except Exception, err:
            sys.stdout.write("thread error: %s\n" % err)
            return False
        else:
            self.c += 1
            self._counter += 1
            self.total += 1
            return True

    def run(self):
        start = time.time()
        self._counter = 0
        while not self.stop:
            if not self.connect():
                break
            if not self.rate:
                continue
            diff = time.time() - start
            pause = (1.0 / self.rate) * self._counter - diff
            if pause < 0:
                continue
            time.sleep(pause)
        self.stop = True

class Manager:
    def __init__(self):
        self.start = time.time()
        self.last = self.start
        self.options = self.parseOptions()

    def parseOptions(self):
        parser = OptionParser(usage="%prog [options]")
        parser.add_option("-H", "--host", dest="hosts",
            help="Host (default: %s)" % DEFAULT_HOST,
            action="append", type="str")
        parser.add_option("-p", "--port",
            help="Port number (eg. 80) or range (eg. 9000:9150), default: %s" % DEFAULT_PORT,
            type="str", default=None)
        parser.add_option("-n", "--threads", dest="nthreads",
            help="Number of threads, default: %s" % DEFAULT_NTHREADS,
            type="int", default=DEFAULT_NTHREADS)
        parser.add_option("-r", "--rate",
            help="Connection rate (connection per second), default: unlimited",
            type="int")
        options, arguments = parser.parse_args()
        if arguments:
            parser.print_help()
            exit(1)
        if options.port:
            try:
                if ':' in options.port:
                    start, end = options.port.split(':', 1)
                    start = int(start)
                    end = int(end)
                    self.ports = xrange(start, end+1)
                else:
                    port = int(options.port)
                    self.ports = (port,)
            except ValueError:
                print("Invalid port number or range: %s" % options.port)
                exit(1)
        else:
            self.ports = (DEFAULT_PORT,)
        if not options.hosts:
            options.hosts = (DEFAULT_HOST,)
        return options

    def stats(self):
        now = time.time()
        diff = now - self.last
        total_diff = now - self.start
        c = 0
        t = 0
        port_error = 0
        for thread in self.threads:
            c += thread.c
            t += thread.total
            port_error += thread.port_error
            thread.c = 0
        if diff:
            cur = '%.1f / sec' % (c / diff)
        else:
            cur = '---'
        if total_diff:
            avg = '%.1f / sec' % (t / total_diff)
        else:
            avg = '---'
        duration = datetime.timedelta(seconds=total_diff)
        duration = str(duration).split(".", 1)[0]
        print("cur=%s, avg=%s, total=%s in %s, %s port errors" % (
              cur,
              avg,
              t, duration,
              port_error))
        self.last = time.time()

    def main(self):
        print("Start %s threads" % self.options.nthreads)
        self.threads = [Connect(self) for n in xrange(self.options.nthreads)]
        for thread in self.threads:
            thread.start()

        try:
            while 1:
                time.sleep(3.0)
                self.stats()
                if any(thread.stop for thread in self.threads):
                    print("-error-")
                    break
        except KeyboardInterrupt:
            print("stop")
        print("Stop %s threads" % self.options.nthreads)
        for thread in self.threads:
            thread.stop = True
        for thread in self.threads:
            thread.join()
        self.stats()

Manager().main()
