
"""
Copyright (C) 2009-2011 EdenWall Technologies

This file is part of NuFirewall. 
 
 NuFirewall is free software: you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation, version 3 of the License. 
 
 NuFirewall is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 
 
 You should have received a copy of the GNU General Public License 
 along with NuFirewall.  If not, see <http://www.gnu.org/licenses/>
"""

from __future__ import with_statement
from logging import (DEBUG, INFO,
    FileHandler, StreamHandler, Formatter)
from optparse import OptionParser
from random import choice
from sys import stdout
from pwd import getpwnam
from grp import getgrnam
from os import chown

from nucentral.client import NuCentralClientBase, NuCentralError
from nucentral.client.nucentral_client_base import DEFAULT_HOST
from nucentral.common.logger import createColoredFormatter, Logger

from .service import Service, parsePrototype
from .file_watch import FileWatch
from .create_arg import PROBE_FILENAME, PROBE_FILE_CONTENT

NUCENTRAL_LOG = '/var/log/nucentral.log'
LOG_FILENAME = 'fuzzer.log'

class Fuzzer(Logger):
    def __init__(self):
        Logger.__init__(self)
        self.target_services = []   # List of Service
        self.nb_call = 0
        self.nucentral_log = None   # FileWatch

    def parseOptions(self):
        parser = OptionParser(usage="%prog [options] [component1 component2 ...]")
        parser.add_option("--blacklist",
            help="Component blacklist separated by comma (eg. 'nuface,nulog')",
            type="str", default="")
        parser.add_option("--host",
            help="Host name or IP address (default: %s)" % DEFAULT_HOST,
            type="str")
        parser.add_option("-v", "--verbose",
            help="Verbose mode",
            action="store_true", default=False)
        parser.add_option("--debug",
            help="Debug mode",
            action="store_true", default=False)
        options, arguments = parser.parse_args()
        if options.debug:
            options.verbose = True
        return options, arguments

    def scanComponents(self, components):
        #return [Service('CORE', 'help', 1, 2)]
        services_with_arg = []
        try:
            if not components:
                self.warning("Scan components")
                components = self.client.call('CORE', 'getComponentList')
            blacklist = [name.strip() for name in self.options.blacklist.split(",")]
            if blacklist:
                components = set(components)
                blacklist = set(blacklist)
                for component in components & blacklist:
                    self.info("Ignore component %s" % component)
                components -= blacklist
            for index, component in enumerate(components):
                self.info("Get component %s services (%s/%s)..."
                    % (component, 1 + index, len(components)))
                services = self.client.call('CORE', 'getServiceList', component)
                with_args = []
                for name in services:
                    doc = self.client.call('CORE', 'help', component, name)
                    prototype = doc[0].rstrip(':')
                    args = parsePrototype(prototype)
                    if not args:
                        continue
                    min_arg, max_arg = args
                    service = Service(component, name, min_arg, max_arg)
                    services_with_arg.append(service)
        except KeyboardInterrupt:
            print
            self.error("Component scanning interrupted.")
        return services_with_arg

    def setupLog(self):
        if self.options.debug:
            level = DEBUG
        else:
            level = INFO

        logger = self.logger
        logger.setLevel(DEBUG)

        handler = StreamHandler(stdout)
        formatter = createColoredFormatter(stdout, display_date=False)
        handler.setFormatter(formatter)
        handler.setLevel(level)
        logger.addHandler(handler)

        handler = FileHandler(LOG_FILENAME, mode='w')
        handler.setFormatter(Formatter('%(asctime)s: %(message)s'))
        handler.setLevel(DEBUG)
        logger.addHandler(handler)

    def init(self, components):
        self.client.authenticate('admin', 'admin')
        self.target_services = self.scanComponents(components)
        if not self.target_services:
            self.critical("No service found: exit")
            exit(1)
        self.nucentral_log = self.createNuCentralLog()
        self.createFileProbe()

    def createNuCentralLog(self):
        log = FileWatch.fromFilename(
            self.options,
            NUCENTRAL_LOG,
            start='end')
        import re
        # Context prefix "127.0.0.1:55010: " or "admin@127.0.0.1:54930: "
        CONTEXT_PREFIX = re.compile(r'^(?:%s@)?\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d{1,5}: ' % self.client.login)
        # Component prefix "[core] " or [core][service]
        COMPONENT_PREFIX = re.compile(r'^(?:\[[^]]+\])+ ')
        def stripPrefixes(line):
            # Strip timestamp "2009-09-24 02:28:45,454: "
            parts = line.split(': ', 1)
            if len(parts) == 2:
                line = parts[1]
            # Strip context prefix
            line = CONTEXT_PREFIX.sub('', line)
            # Strip component prefix
            line = COMPONENT_PREFIX.sub('', line)
            return line
        log.cleanup_func = stripPrefixes
        log.max_process_time = 5.0   # seconds
        log.ignoreRegex(r'^User calls')
        log.ignoreRegex(r'Aucun fichier ou dossier de ce type')
        log.ignoreRegex(r'exception caught here')
        log.addRegex(r'\[(?:OS|IO)Error\]', 1.0)
        # "[twisted] Unhandled error in Deferred:"
        log.addRegex(r'Unhandled', 1.0)
        log.init()
        return log

    def createFileProbe(self):
        with open(PROBE_FILENAME, 'wb') as fp:
            fp.write(PROBE_FILE_CONTENT)
        uid = getpwnam('nucentral')[3]
        gid = getgrnam('nucentral')[2]
        chown(PROBE_FILENAME, uid, gid)

    def main(self):
        self.options, components = self.parseOptions()
        self.setupLog()
        client_options = {}
        if self.options.host:
            client_options['host'] = self.options.host
        self.client = NuCentralClientBase(**client_options)
        try:
            self.init(components)
            self.client.call('nuconf', 'takeWriteRole')
            self.critical("Start fuzzing")
            try:
                self.fuzz()
            except KeyboardInterrupt:
                print
                self.error("Interrupted.")
            self.critical("Stop fuzzing")
        finally:
            self.client.logout()

    def ping(self):
        self.debug("Ping")
        self.client.call('session', 'keepAlive')

    def fuzzService(self):
        service = choice(self.target_services)
        args = service.createArguments()
        self.nb_call += 1
        try:
            args_text = ', '.join(repr(arg) for arg in args)
            self.info("#%s: Call %s.%s(%s)"
                % (self.nb_call, service.component, service.service, args_text))
            self.client.call(service.component, service.service, *args)
        except NuCentralError, err:
            if self.options.debug:
                self.writeError(err, "Service error", log_level=DEBUG)
        self.nucentral_log.live()
        if 0.5 <= self.nucentral_log.score:
            self.critical("Found enough errors in nucentral.log :-)")
            return True
        print
        return False

    def checkProbeFile(self):
        with open(PROBE_FILENAME, 'rb') as fp:
            content = fp.read()
        if content != PROBE_FILE_CONTENT:
            self.critical("Probe file erased! %s" % PROBE_FILENAME)
            return True
        else:
            return False

    def oneStep(self):
        self.nucentral_log.resetScore()
        if self.fuzzService():
            self.critical("It's time to analyze %s" % LOG_FILENAME)
            return True
        self.ping()
        if self.checkProbeFile():
            return True
        return False

    def askContinue(self):
        while True:
            answer = raw_input("Continue (Y/n)?")
            if not answer:
                return True
            answer = answer.strip().lower()
            if answer in ("y", "n"):
                return (answer == "y")
            print 'Please answer "y" or "n"'
            print

    def reset(self):
        # Reseed to the end
        self.nucentral_log.prepareFile()

    def fuzz(self):
        self.warning("Found %s services with arguments" % len(self.target_services))
        while True:
            success = self.oneStep()
            if not success:
                continue
            if self.askContinue():
                self.reset()
                continue
            break

